﻿
namespace KludgeBox
{
	using System;
	using System.Collections.Generic;

	/// <summary>
	/// A global configuration class that stores key-value pairs of different types.
	/// </summary>
	public static class GlobalConfig
	{
		// Some of the properties will be always present
		public static bool Debug { get; private set; } = true;

		private static Dictionary<string, int> _ints = new Dictionary<string, int>();
		private static Dictionary<string, long> _longs = new Dictionary<string, long>();
		private static Dictionary<string, float> _floats = new Dictionary<string, float>();
		private static Dictionary<string, double> _doubles = new Dictionary<string, double>();
		private static Dictionary<string, string> _strings = new Dictionary<string, string>();

		/// <summary>
		/// Retrieves the value associated with the specified key of the given type.
		/// If the key is not found, returns the default value for the type.
		/// </summary>
		/// <typeparam name="T">The type of the value to retrieve.</typeparam>
		/// <param name="key">The key to look up the value.</param>
		/// <param name="defaultValue">The default value to return if the key is not found.</param>
		/// <returns>The value associated with the specified key, or the default value if the key is not found.</returns>
		/// <exception cref="ArgumentException">Thrown when the type <typeparamref name="T"/> is not supported.</exception>
		public static T GetValue<T>(string key, T defaultValue = default)
		{
			if (typeof(T) == typeof(int))
			{
				if (_ints.TryGetValue(key, out int intValue))
				{
					return (T)(object)intValue;
				}
			}
			else if (typeof(T) == typeof(long))
			{
				if (_longs.TryGetValue(key, out long longValue))
				{
					return (T)(object)longValue;
				}
			}
			else if (typeof(T) == typeof(float))
			{
				if (_floats.TryGetValue(key, out float floatValue))
				{
					return (T)(object)floatValue;
				}
			}
			else if (typeof(T) == typeof(double))
			{
				if (_doubles.TryGetValue(key, out double doubleValue))
				{
					return (T)(object)doubleValue;
				}
			}
			else if (typeof(T) == typeof(string))
			{
				if (_strings.TryGetValue(key, out string stringValue))
				{
					return (T)(object)stringValue;
				}
			}

			return defaultValue;
		}

		/// <summary>
		/// Sets the value associated with the specified key of the given type.
		/// If the key does not exist, it will be added to the configuration.
		/// </summary>
		/// <typeparam name="T">The type of the value to set.</typeparam>
		/// <param name="key">The key to set the value for.</param>
		/// <param name="value">The value to be associated with the key.</param>
		/// <exception cref="ArgumentException">Thrown when the type <typeparamref name="T"/> is not supported.</exception>
		public static void SetValue<T>(string key, T value)
		{
			if (typeof(T) == typeof(int))
			{
				_ints[key] = (int)(object)value;
			}
			else if (typeof(T) == typeof(long))
			{
				_longs[key] = (long)(object)value;
			}
			else if (typeof(T) == typeof(float))
			{
				_floats[key] = (float)(object)value;
			}
			else if (typeof(T) == typeof(double))
			{
				_doubles[key] = (double)(object)value;
			}
			else if (typeof(T) == typeof(string))
			{
				_strings[key] = (string)(object)value;
			}
			else
			{
				throw new ArgumentException($"Type {typeof(T)} is not supported.");
			}
		}
		
		/// <summary>
		/// Sets Debug property to false.
		/// </summary>
		public static void DisableDebug()
		{
			Debug = false;
		}
	}
}
