﻿using System;
using System.Runtime.CompilerServices;

namespace KludgeBox
{
	public static class Maths
	{
		// Proxies for existing constants
		public const float PI = MathF.PI;
		public const float Tau = MathF.Tau;
		public const float E = MathF.E;

		public const double PI2 = Math.PI;
		public const double Tau2 = Math.Tau;
		public const double E2 = Math.E;

		public const float Sqrt2 = 1.41421354f;

        public const float Inf = Single.PositiveInfinity;
        public const float NaN = Single.NaN;

        public const double Inf2 = Double.PositiveInfinity;
        public const double NaN2 = Double.NaN;

        // Epsilons
        public const float Epsilon = 1E-06f;
        public const double Epsilon2 = 1E-14;

		// Db conversions
		private const float _linearToDbMultiplier = 8.685889f; // Constant multiplier for linear-to-dB conversion.
		private const double _linearToDbMultiplierDouble = 8.6858896380650368; // Constant multiplier for double-precision linear-to-dB conversion.
		private const float _dbToLinearMultiplier = 0.115129255f; // Constant multiplier for dB-to-linear conversion (single precision).
		private const double _dbToLinearMultiplierDouble = 0.11512925464970228; // Constant multiplier for dB-to-linear conversion (double precision).

        // Mappings for float
        [MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Sqrt(float x) => MathF.Sqrt(x);
        [MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Cbrt(float x) => MathF.Cbrt(x);
        [MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Pow(float x, float y) => MathF.Pow(x, y);

        [MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Log(float x) => MathF.Log(x);
        [MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Log2(float x) => MathF.Log2(x);
        [MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Log10(float x) => MathF.Log10(x);

        [MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Min(float x, float y) => MathF.Min(x, y);
        [MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Max(float x, float y) => MathF.Max(x, y);
        [MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Clamp(float value, float min, float max) => Max(Min(value, Max(max, min)), Min(min, max));

        [MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Floor(float x) => MathF.Floor(x);
        [MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Round(float x) => MathF.Round(x);
        [MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Ceil(float x) => MathF.Ceiling(x);

		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Sin(float x) => MathF.Sin(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Asin(float x) => MathF.Asin(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Sinh(float x) => MathF.Sinh(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Asinh(float x) => MathF.Asinh(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Cos(float x) => MathF.Cos(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Acos(float x) => MathF.Acos(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Cosh(float x) => MathF.Cosh(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Acosh(float x) => MathF.Acosh(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Tan(float x) => MathF.Tan(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Atan(float x) => MathF.Atan(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Tanh(float x) => MathF.Tanh(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Atanh(float x) => MathF.Atanh(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Atan2(float x, float y) => MathF.Atan2(x, y);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Abs(float x) => MathF.Abs(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static float Exp(float x) => MathF.Exp(x);
        [MethodImpl(MethodImplOptions.AggressiveInlining)] public static (float Sin, float Cos) SinCos(float x) => MathF.SinCos(x);

        // Mappings for double
        [MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Sqrt(double x) => Math.Sqrt(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Cbrt(double x) => Math.Cbrt(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Pow(double x, double y) => Math.Pow(x, y);

		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Log(double x) => Math.Log(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Log2(double x) => Math.Log2(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Log10(double x) => Math.Log10(x);

		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Min(double x, double y) => Math.Min(x, y);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Max(double x, double y) => Math.Max(x, y);
        [MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Clamp(double value, double min, double max) => Math.Clamp(value, Min(min, max), Max(min, max));

		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Floor(double x) => Math.Floor(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Round(double x) => Math.Round(x);
        [MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Ceil(double x) => Math.Ceiling(x);

		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Sin(double x) => Math.Sin(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Asin(double x) => Math.Asin(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Sinh(double x) => Math.Sinh(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Asinh(double x) => Math.Asinh(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Cos(double x) => Math.Cos(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Acos(double x) => Math.Acos(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Cosh(double x) => Math.Cosh(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Acosh(double x) => Math.Acosh(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Tan(double x) => Math.Tan(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Atan(double x) => Math.Atan(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Tanh(double x) => Math.Tanh(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Atanh(double x) => Math.Atanh(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Atan2(double x, double y) => Math.Atan2(x, y);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Exp(double x) => Math.Exp(x);
		[MethodImpl(MethodImplOptions.AggressiveInlining)] public static double Abs(double x) => Math.Abs(x);
        [MethodImpl(MethodImplOptions.AggressiveInlining)] public static (double Sin, double Cos) SinCos(double x) => Math.SinCos(x);

        // Mappings for integer
        [MethodImpl(MethodImplOptions.AggressiveInlining)] public static int Min(int x, int y) => Math.Min(x, y);
        [MethodImpl(MethodImplOptions.AggressiveInlining)] public static int Max(int x, int y) => Math.Max(x, y);
        [MethodImpl(MethodImplOptions.AggressiveInlining)] public static int Clamp(int value, int min, int max) => Math.Clamp(value, Min(min, max), Max(min, max));
        [MethodImpl(MethodImplOptions.AggressiveInlining)] public static int Abs(int x) => Math.Abs(x);

		// Custom properties
		/// <summary>
		/// Multiply by this to convert from radians to degrees
		/// </summary>
		public const float RadiansToDegrees = 180f / PI;
		public const float RadDeg = RadiansToDegrees;

		/// <summary>
		/// Multiply by this to convert from degrees to radians
		/// </summary>
		public const float DegreesToRadians = PI / 180;
		public const float DegRad = DegreesToRadians;
		public const double DoubleDegRad = 0.017453292519943295;
		public const double DoubleRadDeg = 57.29577951308232;

		// Custom methods

		/// <summary>
		/// Determines whether the specified value is a power of two.
		/// </summary>
		/// <param name="value">The value to check.</param>
		/// <returns><c>true</c> if the value is a power of two; otherwise, <c>false</c>.</returns>
		public static bool IsPowerOfTwo(int value) => value != 0 && (value & value - 1) == 0;

		/// <summary>
		/// Returns the next power of two that is greater than or equal to the specified value.
		/// </summary>
		/// <param name="value">The value to find the next power of two for.</param>
		/// <returns>The next power of two that is greater than or equal to the specified value.</returns>
		public static int NextPowerOfTwo(int value)
		{
			if (value == 0) return 1;
			value--;
			value |= value >> 1;
			value |= value >> 2;
			value |= value >> 4;
			value |= value >> 8;
			value |= value >> 16;
			return value + 1;
		}

		/// <summary>
		/// Determines whether the specified value is a power of two.
		/// </summary>
		/// <param name="value">The value to check.</param>
		/// <returns><c>true</c> if the value is a power of two; otherwise, <c>false</c>.</returns>
		public static bool IsPowerOfTwo(long value) => value != 0 && (value & (value - 1)) == 0;

		/// <summary>
		/// Returns the next power of two that is greater than or equal to the specified value.
		/// </summary>
		/// <param name="value">The value to find the next power of two for.</param>
		/// <returns>The next power of two that is greater than or equal to the specified value.</returns>
		public static long NextPowerOfTwo(long value)
		{
			if (value == 0) return 1;
			value--;
			value |= value >> 1;
			value |= value >> 2;
			value |= value >> 4;
			value |= value >> 8;
			value |= value >> 16;
			value |= value >> 32;
			return value + 1;
		}

		/// <summary>
		/// Moves a value towards a target value by a specified speed.
		/// </summary>
		/// <param name="from">The starting value.</param>
		/// <param name="to">The target value.</param>
		/// <param name="speed">The maximum amount to move towards the target.</param>
		/// <returns>The new value after moving towards the target.</returns>
		public static float Approach(float from, float to, float speed) => from + Clamp(to - from, -speed, speed);

		/// <summary>
		/// Moves a value towards a target value by a specified speed.
		/// </summary>
		/// <param name="from">The starting value.</param>
		/// <param name="to">The target value.</param>
		/// <param name="speed">The maximum amount to move towards the target.</param>
		/// <returns>The new value after moving towards the target.</returns>
		public static double Approach(double from, double to, double speed) => from + Clamp(to - from, -speed, speed);

		/// <summary>
		/// Linearly interpolates between two values.
		/// </summary>
		/// <param name="fromValue">The starting value.</param>
		/// <param name="toValue">The target value.</param>
		/// <param name="progress">The interpolation progress (between 0 and 1).</param>
		/// <returns>The interpolated value.</returns>
		public static float Lerp(float fromValue, float toValue, float progress) => fromValue + (toValue - fromValue) * progress;/// <summary>
		
		/// Linearly interpolates between two values.
		/// </summary>
		/// <param name="fromValue">The starting value.</param>
		/// <param name="toValue">The target value.</param>
		/// <param name="progress">The interpolation progress (between 0 and 1).</param>
		/// <returns>The interpolated value.</returns>
		public static double Lerp(double fromValue, double toValue, double progress) => fromValue + (toValue - fromValue) * progress;

		/// <summary>
		/// Spherically interpolates between two angles (in radians).
		/// </summary>
		/// <param name="fromRadians">The starting angle in radians.</param>
		/// <param name="toRadians">The target angle in radians.</param>
		/// <param name="progress">The interpolation progress (between 0 and 1).</param>
		/// <returns>The interpolated angle in radians.</returns>
		public static float SlerpRad(float fromRadians, float toRadians, float progress)
		{
			float delta = ((toRadians - fromRadians + Tau + PI) % Tau) - PI;
			return (fromRadians + delta * progress + Tau) % Tau;
		}

		/// <summary>
		/// Spherically interpolates between two angles (in degrees).
		/// </summary>
		/// <param name="fromDegrees">The starting angle in degrees.</param>
		/// <param name="toDegrees">The target angle in degrees.</param>
		/// <param name="progress">The interpolation progress (between 0 and 1).</param>
		/// <returns>The interpolated angle in degrees.</returns>
		public static float Slerp(float fromDegrees, float toDegrees, float progress)
		{
			float delta = ((toDegrees - fromDegrees + 360 + 180) % 360) - 180;
			return (fromDegrees + delta * progress + 360) % 360;
		}

		/// <summary>
		/// Maps a value from one range to another range.
		/// </summary>
		/// <param name="value">The value to map.</param>
		/// <param name="froma">The start of the original range.</param>
		/// <param name="toa">The end of the original range.</param>
		/// <param name="fromb">The start of the target range.</param>
		/// <param name="tob">The end of the target range.</param>
		/// <returns>The mapped value.</returns>
		public static float Map(float value, float froma, float toa, float fromb, float tob)
		{
			return fromb + (value - froma) * (tob - fromb) / (toa - froma);
		}

		/// <summary>
		/// Maps a value from the default range [0, 1] to a target range.
		/// </summary>
		/// <param name="value">The value to map.</param>
		/// <param name="from">The start of the target range.</param>
		/// <param name="to">The end of the target range.</param>
		/// <returns>The mapped value.</returns>
		public static float Map(float value, float from, float to)
		{
			return Map(value, 0, 1, from, to);
		}

		/// <summary>
		/// Calculates the number of digits in an integer.
		/// </summary>
		/// <param name="n">The integer value.</param>
		/// <returns>The number of digits in the integer.</returns>
		public static int Digits(int n)
		{
			return n < 100000
				? n < 100
					? n < 10 ? 1 : 2
					: n < 1000 ? 3
					: n < 10000 ? 4
					: 5
				: n < 10000000
					? n < 1000000 ? 6 : 7
					: n < 100000000 ? 8
					: n < 1000000000 ? 9 : 10;
		}

		/// <summary>
		/// Calculates the number of digits in a long integer.
		/// </summary>
		/// <param name="n">The long integer value.</param>
		/// <returns>The number of digits in the long integer.</returns>
		public static int Digits(long n)
		{
			return n == 0 ? 1 : (int)(Log10((double)n) + 1);
		}
		/// <summary>
		/// Calculates the sign of the given integer.
		/// </summary>
		/// <param name="num">The integer value.</param>
		/// <returns>
		/// 1 if the number is positive,
		/// -1 if the number is negative,
		/// 0 if the number is zero.
		/// </returns>
		public static int Sign(this int num) => Math.Sign(num);

		/// <summary>
		/// Calculates the sign of the given long integer.
		/// </summary>
		/// <param name="num">The long integer value.</param>
		/// <returns>
		/// 1 if the number is positive,
		/// -1 if the number is negative,
		/// 0 if the number is zero.
		/// </returns>
		public static long Sign(this long num) => Math.Sign(num);

		/// <summary>
		/// Calculates the sign of the given single-precision floating-point number.
		/// </summary>
		/// <param name="num">The single-precision floating-point value.</param>
		/// <returns>
		/// 1.0f if the number is positive,
		/// -1.0f if the number is negative,
		/// 0.0f if the number is zero.
		/// </returns>
		public static float Sign(this float num) => MathF.Sign(num);

		/// <summary>
		/// Calculates the sign of the given double-precision floating-point number.
		/// </summary>
		/// <param name="num">The double-precision floating-point value.</param>
		/// <returns>
		/// 1.0 if the number is positive,
		/// -1.0 if the number is negative,
		/// 0.0 if the number is zero.
		/// </returns>
		public static double Sign(this double num) => Math.Sign(num);

		/// <summary>
		/// Checks if two single-precision floating-point numbers are approximately equal.
		/// </summary>
		/// <param name="a">The first floating-point number.</param>
		/// <param name="b">The second floating-point number.</param>
		/// <returns>
		/// True if the numbers are approximately equal within a small tolerance, false otherwise.
		/// </returns>
		public static bool IsEqualApprox(this float a, float b)
		{
			if (a == b)
			{
				return true;
			}

			float tolerance = Epsilon * Abs(a);
			if (tolerance < Epsilon)
			{
				tolerance = Epsilon;
			}

			return Abs(a - b) < tolerance;
		}

		/// <summary>
		/// Checks if two double-precision floating-point numbers are approximately equal.
		/// </summary>
		/// <param name="a">The first floating-point number.</param>
		/// <param name="b">The second floating-point number.</param>
		/// <returns>
		/// True if the numbers are approximately equal within a small tolerance, false otherwise.
		/// </returns>
		public static bool IsEqualApprox(this double a, double b)
		{
			if (a == b)
			{
				return true;
			}

			double tolerance = Epsilon2 * Abs(a);
			if (tolerance < Epsilon2)
			{
				tolerance = Epsilon2;
			}

			return Abs(a - b) < tolerance;
		}


		/// <summary>
		/// Converts a dB value to its corresponding linear value (single precision).
		/// </summary>
		/// <param name="db">The dB value to convert.</param>
		/// <returns>The linear value corresponding to the given dB value.</returns>
		public static float DbToLinear(float db)
		{
			return Exp(db * _dbToLinearMultiplier);
		}

		/// <summary>
		/// Converts a dB value to its corresponding linear value (double precision).
		/// </summary>
		/// <param name="db">The dB value to convert.</param>
		/// <returns>The linear value corresponding to the given dB value.</returns>
		public static double DbToLinear(double db)
		{
			return Exp(db * _dbToLinearMultiplierDouble);
		}

		/// <summary>
		/// Checks if a double value is approximately zero within a small tolerance.
		/// </summary>
		/// <param name="s">The double value to check.</param>
		/// <returns>True if the value is approximately zero; otherwise, false.</returns>
		public static bool IsZeroApprox(double s)
		{
			return Abs(s) < Epsilon2;
		}

		/// <summary>
		/// Checks if a float value is approximately zero within a small tolerance.
		/// </summary>
		/// <param name="s">The float value to check.</param>
		/// <returns>True if the value is approximately zero; otherwise, false.</returns>
		public static bool IsZeroApprox(float s)
		{
			return Abs(s) < Epsilon;
		}

		/// <summary>
		/// Converts a linear value to its corresponding dB value (single precision).
		/// </summary>
		/// <param name="linear">The linear value to convert.</param>
		/// <returns>The dB value corresponding to the given linear value.</returns>
		public static float LinearToDb(float linear)
		{
			return Log(linear) * _linearToDbMultiplier;
		}

		/// <summary>
		/// Converts a linear value to its corresponding dB value (double precision).
		/// </summary>
		/// <param name="linear">The linear value to convert.</param>
		/// <returns>The dB value corresponding to the given linear value.</returns>
		public static double LinearToDb(double linear)
		{
			return Log(linear) * _linearToDbMultiplierDouble;
		}

		/// <summary>
		/// Returns the position of the first non-zero digit, after the decimal point.
		/// </summary>
		/// <param name="step">The step value to analyze.</param>
		/// <returns>The number of decimal places in the step value.</returns>
		/// <remarks>Note that the maximum return value is 10, which is a design decision in the implementation.</remarks>
		public static int StepDecimals(double step)
		{
			double[] array = new double[9] { 0.9999, 0.09999, 0.009999, 0.0009999, 9.999E-05, 9.999E-06, 9.999E-07, 9.999E-08, 9.999E-09 };
			double num = Abs(step);
			double num2 = num - (double)(int)num;
			for (int i = 0; i < array.Length; i++)
			{
				if (num2 >= array[i])
				{
					return i;
				}
			}

			return 0;
		}

		/// <summary>
		/// Snaps a single-precision floating-point value to the nearest multiple of the specified step value.
		/// </summary>
		/// <param name="s">The value to snap.</param>
		/// <param name="step">The step value, used for snapping.</param>
		/// <returns>The snapped value.</returns>
		public static float Snapped(float s, float step)
		{
			if (step != 0f)
			{
				return Floor(s / step + 0.5f) * step;
			}

			return s;
		}

		/// <summary>
		/// Snaps a double-precision floating-point value to the nearest multiple of the specified step value.
		/// </summary>
		/// <param name="s">The value to snap.</param>
		/// <param name="step">The step value, used for snapping.</param>
		/// <returns>The snapped value.</returns>
		public static double Snapped(double s, double step)
		{
			if (step != 0.0)
			{
				return Floor(s / step + 0.5) * step;
			}

			return s;
		}

		/// <summary>
		/// Wraps a single-precision floating-point value within a specified range.
		/// </summary>
		/// <param name="value">The value to wrap.</param>
		/// <param name="min">The minimum value of the range (inclusive).</param>
		/// <param name="max">The maximum value of the range (exclusive).</param>
		/// <returns>The wrapped value within the specified range.</returns>
		public static float Wrap(float value, float min, float max)
		{
			float num = max - min;
			if (IsZeroApprox(num))
			{
				return min;
			}

			return min + ((value - min) % num + num) % num;
		}

		/// <summary>
		/// Wraps an integer value within a specified range.
		/// </summary>
		/// <param name="value">The value to wrap.</param>
		/// <param name="min">The minimum value of the range (inclusive).</param>
		/// <param name="max">The maximum value of the range (exclusive).</param>
		/// <returns>The wrapped value within the specified range.</returns>
		public static int Wrap(int value, int min, int max)
		{
			int num = max - min;
			if (num == 0)
			{
				return min;
			}

			return min + ((value - min) % num + num) % num;
		}

		/// <summary>
		/// Wraps a double-precision floating-point value within a specified range.
		/// </summary>
		/// <param name="value">The value to wrap.</param>
		/// <param name="min">The minimum value of the range (inclusive).</param>
		/// <param name="max">The maximum value of the range (exclusive).</param>
		/// <returns>The wrapped value within the specified range.</returns>
		public static double Wrap(double value, double min, double max)
		{
			double num = max - min;
			if (IsZeroApprox(num))
			{
				return min;
			}

			return min + ((value - min) % num + num) % num;
		}

		/// <summary>
		/// Returns the number of decimal places in the specified decimal value.
		/// </summary>
		/// <param name="s">The decimal value to count decimal places for.</param>
		/// <returns>The number of decimal places in the decimal value.</returns>
		public static int DecimalCount(decimal s)
		{
			return BitConverter.GetBytes(decimal.GetBits(s)[3])[2];
		}

		/// <summary>
		/// Returns the number of decimal places in the specified double value.
		/// </summary>
		/// <param name="s">The double value to count decimal places for.</param>
		/// <returns>The number of decimal places in the double value.</returns>
		public static int DecimalCount(double s)
		{
			return DecimalCount((decimal)s);
		}

		/// <summary>
		/// Ping-pongs the input value between 0 and the specified length. 
		/// If the limit is reached, the next value returned is decreased towards 0 or increased towards the length 
		/// in a triangular wave pattern. If the length is less than zero, it is treated as a positive value.
		/// </summary>
		/// <param name="value">The input value.</param>
		/// <param name="length">The maximum length of the ping-pong interval.</param>
		/// <returns>The ping-ponged value between 0 and the specified length.</returns>
		public static double PingPong(double value, double length)
		{
			if (length == 0.0)
			{
				return 0.0;
			}

			return Abs(Fract((value - length) / (length * 2.0)) * length * 2.0 - length);

			static double Fract(double value)
			{
				return value - Floor(value);
			}
		}

		/// <summary>
		/// Ping-pongs the input value between 0 and the specified length. 
		/// If the limit is reached, the next value returned is decreased towards 0 or increased towards the length 
		/// in a triangular wave pattern. If the length is less than zero, it is treated as a positive value.
		/// </summary>
		/// <param name="value">The input value.</param>
		/// <param name="length">The maximum length of the ping-pong interval.</param>
		/// <returns>The ping-ponged value between 0 and the specified length.</returns>
		public static float PingPong(float value, float length)
		{
			if (length == 0f)
			{
				return 0f;
			}

			return Abs(Fract((value - length) / (length * 2f)) * length * 2f - length);

			static float Fract(float value)
			{
				return value - Floor(value);
			}
		}

		/// <summary>
		/// Checks if two float values are approximately equal within the specified tolerance.
		/// </summary>
		/// <param name="a">The first float value.</param>
		/// <param name="b">The second float value.</param>
		/// <param name="tolerance">The tolerance within which the values are considered equal.</param>
		/// <returns>True if the float values are approximately equal; otherwise, false.</returns>
		public static bool IsEqualApprox(float a, float b, float tolerance)
		{
			if (a == b)
			{
				return true;
			}

			return Abs(a - b) < tolerance;
		}

		/// <summary>
		/// Checks if two double values are approximately equal within the specified tolerance.
		/// </summary>
		/// <param name="a">The first double value.</param>
		/// <param name="b">The second double value.</param>
		/// <param name="tolerance">The tolerance within which the values are considered equal.</param>
		/// <returns>True if the double values are approximately equal; otherwise, false.</returns>
		public static bool IsEqualApprox(double a, double b, double tolerance)
		{
			if (a == b)
			{
				return true;
			}

			return Abs(a - b) < tolerance;
		}

        /// <summary>
        /// Cubic interpolates between two values by the factor defined in <paramref name="weight"/>
        /// with pre and post values.
        /// </summary>
        /// <param name="from">The start value for interpolation.</param>
        /// <param name="to">The destination value for interpolation.</param>
        /// <param name="pre">The value which before "from" value for interpolation.</param>
        /// <param name="post">The value which after "to" value for interpolation.</param>
        /// <param name="weight">A value on the range of 0.0 to 1.0, representing the amount of interpolation.</param>
        /// <returns>The resulting value of the interpolation.</returns>
        public static float CubicInterpolate(float from, float to, float pre, float post, float weight)
        {
            return 0.5f *
                    ((from * 2.0f) +
                            (-pre + to) * weight +
                            (2.0f * pre - 5.0f * from + 4.0f * to - post) * (weight * weight) +
                            (-pre + 3.0f * from - 3.0f * to + post) * (weight * weight * weight));
        }

        /// <summary>
        /// Cubic interpolates between two values by the factor defined in <paramref name="weight"/>
        /// with pre and post values.
        /// </summary>
        /// <param name="from">The start value for interpolation.</param>
        /// <param name="to">The destination value for interpolation.</param>
        /// <param name="pre">The value which before "from" value for interpolation.</param>
        /// <param name="post">The value which after "to" value for interpolation.</param>
        /// <param name="weight">A value on the range of 0.0 to 1.0, representing the amount of interpolation.</param>
        /// <returns>The resulting value of the interpolation.</returns>
        public static double CubicInterpolate(double from, double to, double pre, double post, double weight)
        {
            return 0.5 *
                    ((from * 2.0) +
                            (-pre + to) * weight +
                            (2.0 * pre - 5.0 * from + 4.0 * to - post) * (weight * weight) +
                            (-pre + 3.0 * from - 3.0 * to + post) * (weight * weight * weight));
        }

        /// <summary>
        /// Cubic interpolates between two rotation values with shortest path
        /// by the factor defined in <paramref name="weight"/> with pre and post values.
        /// See also <see cref="LerpAngle(float, float, float)"/>.
        /// </summary>
        /// <param name="from">The start value for interpolation.</param>
        /// <param name="to">The destination value for interpolation.</param>
        /// <param name="pre">The value which before "from" value for interpolation.</param>
        /// <param name="post">The value which after "to" value for interpolation.</param>
        /// <param name="weight">A value on the range of 0.0 to 1.0, representing the amount of interpolation.</param>
        /// <returns>The resulting value of the interpolation.</returns>
        public static float CubicInterpolateAngle(float from, float to, float pre, float post, float weight)
        {
            float fromRot = from % Tau;

            float preDiff = (pre - fromRot) % Tau;
            float preRot = fromRot + (2.0f * preDiff) % Tau - preDiff;

            float toDiff = (to - fromRot) % Tau;
            float toRot = fromRot + (2.0f * toDiff) % Tau - toDiff;

            float postDiff = (post - toRot) % Tau;
            float postRot = toRot + (2.0f * postDiff) % Tau - postDiff;

            return CubicInterpolate(fromRot, toRot, preRot, postRot, weight);
        }

        /// <summary>
        /// Cubic interpolates between two rotation values with shortest path
        /// by the factor defined in <paramref name="weight"/> with pre and post values.
        /// See also <see cref="LerpAngle(double, double, double)"/>.
        /// </summary>
        /// <param name="from">The start value for interpolation.</param>
        /// <param name="to">The destination value for interpolation.</param>
        /// <param name="pre">The value which before "from" value for interpolation.</param>
        /// <param name="post">The value which after "to" value for interpolation.</param>
        /// <param name="weight">A value on the range of 0.0 to 1.0, representing the amount of interpolation.</param>
        /// <returns>The resulting value of the interpolation.</returns>
        public static double CubicInterpolateAngle(double from, double to, double pre, double post, double weight)
        {
            double fromRot = from % Tau2;

            double preDiff = (pre - fromRot) % Tau2;
            double preRot = fromRot + (2.0 * preDiff) % Tau2 - preDiff;

            double toDiff = (to - fromRot) % Tau2;
            double toRot = fromRot + (2.0 * toDiff) % Tau2 - toDiff;

            double postDiff = (post - toRot) % Tau2;
            double postRot = toRot + (2.0 * postDiff) % Tau2 - postDiff;

            return CubicInterpolate(fromRot, toRot, preRot, postRot, weight);
        }

        /// <summary>
        /// Cubic interpolates between two values by the factor defined in <paramref name="weight"/>
        /// with pre and post values.
        /// It can perform smoother interpolation than
        /// <see cref="CubicInterpolate(float, float, float, float, float)"/>
        /// by the time values.
        /// </summary>
        /// <param name="from">The start value for interpolation.</param>
        /// <param name="to">The destination value for interpolation.</param>
        /// <param name="pre">The value which before "from" value for interpolation.</param>
        /// <param name="post">The value which after "to" value for interpolation.</param>
        /// <param name="weight">A value on the range of 0.0 to 1.0, representing the amount of interpolation.</param>
        /// <param name="toT"></param>
        /// <param name="preT"></param>
        /// <param name="postT"></param>
        /// <returns>The resulting value of the interpolation.</returns>
        public static float CubicInterpolateInTime(float from, float to, float pre, float post, float weight, float toT, float preT, float postT)
        {
            /* Barry-Goldman method */
            float t = Lerp(0.0f, toT, weight);
            float a1 = Lerp(pre, from, preT == 0 ? 0.0f : (t - preT) / -preT);
            float a2 = Lerp(from, to, toT == 0 ? 0.5f : t / toT);
            float a3 = Lerp(to, post, postT - toT == 0 ? 1.0f : (t - toT) / (postT - toT));
            float b1 = Lerp(a1, a2, toT - preT == 0 ? 0.0f : (t - preT) / (toT - preT));
            float b2 = Lerp(a2, a3, postT == 0 ? 1.0f : t / postT);
            return Lerp(b1, b2, toT == 0 ? 0.5f : t / toT);
        }

        /// <summary>
        /// Cubic interpolates between two values by the factor defined in <paramref name="weight"/>
        /// with pre and post values.
        /// It can perform smoother interpolation than
        /// <see cref="CubicInterpolate(double, double, double, double, double)"/>
        /// by the time values.
        /// </summary>
        /// <param name="from">The start value for interpolation.</param>
        /// <param name="to">The destination value for interpolation.</param>
        /// <param name="pre">The value which before "from" value for interpolation.</param>
        /// <param name="post">The value which after "to" value for interpolation.</param>
        /// <param name="weight">A value on the range of 0.0 to 1.0, representing the amount of interpolation.</param>
        /// <param name="toT"></param>
        /// <param name="preT"></param>
        /// <param name="postT"></param>
        /// <returns>The resulting value of the interpolation.</returns>
        public static double CubicInterpolateInTime(double from, double to, double pre, double post, double weight, double toT, double preT, double postT)
        {
            /* Barry-Goldman method */
            double t = Lerp(0.0, toT, weight);
            double a1 = Lerp(pre, from, preT == 0 ? 0.0 : (t - preT) / -preT);
            double a2 = Lerp(from, to, toT == 0 ? 0.5 : t / toT);
            double a3 = Lerp(to, post, postT - toT == 0 ? 1.0 : (t - toT) / (postT - toT));
            double b1 = Lerp(a1, a2, toT - preT == 0 ? 0.0 : (t - preT) / (toT - preT));
            double b2 = Lerp(a2, a3, postT == 0 ? 1.0 : t / postT);
            return Lerp(b1, b2, toT == 0 ? 0.5 : t / toT);
        }

        /// <summary>
        /// Cubic interpolates between two rotation values with shortest path
        /// by the factor defined in <paramref name="weight"/> with pre and post values.
        /// See also <see cref="LerpAngle(float, float, float)"/>.
        /// It can perform smoother interpolation than
        /// <see cref="CubicInterpolateAngle(float, float, float, float, float)"/>
        /// by the time values.
        /// </summary>
        /// <param name="from">The start value for interpolation.</param>
        /// <param name="to">The destination value for interpolation.</param>
        /// <param name="pre">The value which before "from" value for interpolation.</param>
        /// <param name="post">The value which after "to" value for interpolation.</param>
        /// <param name="weight">A value on the range of 0.0 to 1.0, representing the amount of interpolation.</param>
        /// <param name="toT"></param>
        /// <param name="preT"></param>
        /// <param name="postT"></param>
        /// <returns>The resulting value of the interpolation.</returns>
        public static float CubicInterpolateAngleInTime(float from, float to, float pre, float post, float weight, float toT, float preT, float postT)
        {
            float fromRot = from % Tau;

            float preDiff = (pre - fromRot) % Tau;
            float preRot = fromRot + (2.0f * preDiff) % Tau - preDiff;

            float toDiff = (to - fromRot) % Tau;
            float toRot = fromRot + (2.0f * toDiff) % Tau - toDiff;

            float postDiff = (post - toRot) % Tau;
            float postRot = toRot + (2.0f * postDiff) % Tau - postDiff;

            return CubicInterpolateInTime(fromRot, toRot, preRot, postRot, weight, toT, preT, postT);
        }

        /// <summary>
        /// Cubic interpolates between two rotation values with shortest path
        /// by the factor defined in <paramref name="weight"/> with pre and post values.
        /// See also <see cref="LerpAngle(double, double, double)"/>.
        /// It can perform smoother interpolation than
        /// <see cref="CubicInterpolateAngle(double, double, double, double, double)"/>
        /// by the time values.
        /// </summary>
        /// <param name="from">The start value for interpolation.</param>
        /// <param name="to">The destination value for interpolation.</param>
        /// <param name="pre">The value which before "from" value for interpolation.</param>
        /// <param name="post">The value which after "to" value for interpolation.</param>
        /// <param name="weight">A value on the range of 0.0 to 1.0, representing the amount of interpolation.</param>
        /// <param name="toT"></param>
        /// <param name="preT"></param>
        /// <param name="postT"></param>
        /// <returns>The resulting value of the interpolation.</returns>
        public static double CubicInterpolateAngleInTime(double from, double to, double pre, double post, double weight, double toT, double preT, double postT)
        {
            double fromRot = from % Tau2;

            double preDiff = (pre - fromRot) % Tau2;
            double preRot = fromRot + (2.0 * preDiff) % Tau2 - preDiff;

            double toDiff = (to - fromRot) % Tau2;
            double toRot = fromRot + (2.0 * toDiff) % Tau2 - toDiff;

            double postDiff = (post - toRot) % Tau2;
            double postRot = toRot + (2.0 * postDiff) % Tau2 - postDiff;

            return CubicInterpolateInTime(fromRot, toRot, preRot, postRot, weight, toT, preT, postT);
        }

        /// <summary>
        /// Returns the point at the given <paramref name="t"/> on a one-dimensional Bezier curve defined by
        /// the given <paramref name="control1"/>, <paramref name="control2"/>, and <paramref name="end"/> points.
        /// </summary>
        /// <param name="start">The start value for the interpolation.</param>
        /// <param name="control1">Control point that defines the bezier curve.</param>
        /// <param name="control2">Control point that defines the bezier curve.</param>
        /// <param name="end">The destination value for the interpolation.</param>
        /// <param name="t">A value on the range of 0.0 to 1.0, representing the amount of interpolation.</param>
        /// <returns>The resulting value of the interpolation.</returns>
        public static float BezierInterpolate(float start, float control1, float control2, float end, float t)
        {
            // Formula from Wikipedia article on Bezier curves
            float omt = 1.0f - t;
            float omt2 = omt * omt;
            float omt3 = omt2 * omt;
            float t2 = t * t;
            float t3 = t2 * t;

            return start * omt3 + control1 * omt2 * t * 3.0f + control2 * omt * t2 * 3.0f + end * t3;
        }

        /// <summary>
        /// Returns the point at the given <paramref name="t"/> on a one-dimensional Bezier curve defined by
        /// the given <paramref name="control1"/>, <paramref name="control2"/>, and <paramref name="end"/> points.
        /// </summary>
        /// <param name="start">The start value for the interpolation.</param>
        /// <param name="control1">Control point that defines the bezier curve.</param>
        /// <param name="control2">Control point that defines the bezier curve.</param>
        /// <param name="end">The destination value for the interpolation.</param>
        /// <param name="t">A value on the range of 0.0 to 1.0, representing the amount of interpolation.</param>
        /// <returns>The resulting value of the interpolation.</returns>
        public static double BezierInterpolate(double start, double control1, double control2, double end, double t)
        {
            // Formula from Wikipedia article on Bezier curves
            double omt = 1.0 - t;
            double omt2 = omt * omt;
            double omt3 = omt2 * omt;
            double t2 = t * t;
            double t3 = t2 * t;

            return start * omt3 + control1 * omt2 * t * 3.0 + control2 * omt * t2 * 3.0 + end * t3;
        }

        /// <summary>
        /// Returns the derivative at the given <paramref name="t"/> on a one dimensional Bezier curve defined by
        /// the given <paramref name="control1"/>, <paramref name="control2"/>, and <paramref name="end"/> points.
        /// </summary>
        /// <param name="start">The start value for the interpolation.</param>
        /// <param name="control1">Control point that defines the bezier curve.</param>
        /// <param name="control2">Control point that defines the bezier curve.</param>
        /// <param name="end">The destination value for the interpolation.</param>
        /// <param name="t">A value on the range of 0.0 to 1.0, representing the amount of interpolation.</param>
        /// <returns>The resulting value of the interpolation.</returns>
        public static float BezierDerivative(float start, float control1, float control2, float end, float t)
        {
            // Formula from Wikipedia article on Bezier curves
            float omt = 1.0f - t;
            float omt2 = omt * omt;
            float t2 = t * t;

            float d = (control1 - start) * 3.0f * omt2 + (control2 - control1) * 6.0f * omt * t + (end - control2) * 3.0f * t2;
            return d;
        }

        /// <summary>
        /// Returns the derivative at the given <paramref name="t"/> on a one dimensional Bezier curve defined by
        /// the given <paramref name="control1"/>, <paramref name="control2"/>, and <paramref name="end"/> points.
        /// </summary>
        /// <param name="start">The start value for the interpolation.</param>
        /// <param name="control1">Control point that defines the bezier curve.</param>
        /// <param name="control2">Control point that defines the bezier curve.</param>
        /// <param name="end">The destination value for the interpolation.</param>
        /// <param name="t">A value on the range of 0.0 to 1.0, representing the amount of interpolation.</param>
        /// <returns>The resulting value of the interpolation.</returns>
        public static double BezierDerivative(double start, double control1, double control2, double end, double t)
        {
            // Formula from Wikipedia article on Bezier curves
            double omt = 1.0 - t;
            double omt2 = omt * omt;
            double t2 = t * t;

            double d = (control1 - start) * 3.0 * omt2 + (control2 - control1) * 6.0 * omt * t + (end - control2) * 3.0 * t2;
            return d;
        }

        /// <summary>
        /// Returns whether <paramref name="s"/> is a finite value, i.e. it is not
        /// <see cref="NaN"/>, positive infinite, or negative infinity.
        /// </summary>
        /// <param name="s">The value to check.</param>
        /// <returns>A <see langword="bool"/> for whether or not the value is a finite value.</returns>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static bool IsFinite(float s)
        {
            return Single.IsFinite(s);
        }

        /// <summary>
        /// Returns whether <paramref name="s"/> is a finite value, i.e. it is not
        /// <see cref="NaN"/>, positive infinite, or negative infinity.
        /// </summary>
        /// <param name="s">The value to check.</param>
        /// <returns>A <see langword="bool"/> for whether or not the value is a finite value.</returns>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static bool IsFinite(double s)
        {
            return Double.IsFinite(s);
        }

        /// <summary>
        /// Returns whether <paramref name="s"/> is an infinity value (either positive infinity or negative infinity).
        /// </summary>
        /// <param name="s">The value to check.</param>
        /// <returns>A <see langword="bool"/> for whether or not the value is an infinity value.</returns>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static bool IsInf(float s)
        {
            return Single.IsInfinity(s);
        }

        /// <summary>
        /// Returns whether <paramref name="s"/> is an infinity value (either positive infinity or negative infinity).
        /// </summary>
        /// <param name="s">The value to check.</param>
        /// <returns>A <see langword="bool"/> for whether or not the value is an infinity value.</returns>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static bool IsInf(double s)
        {
            return Double.IsInfinity(s);
        }

        /// <summary>
        /// Returns whether <paramref name="s"/> is a <c>NaN</c> ("Not a Number" or invalid) value.
        /// </summary>
        /// <param name="s">The value to check.</param>
        /// <returns>A <see langword="bool"/> for whether or not the value is a <c>NaN</c> value.</returns>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static bool IsNaN(float s)
        {
            return Single.IsNaN(s);
        }

        /// <summary>
        /// Returns whether <paramref name="s"/> is a <c>NaN</c> ("Not a Number" or invalid) value.
        /// </summary>
        /// <param name="s">The value to check.</param>
        /// <returns>A <see langword="bool"/> for whether or not the value is a <c>NaN</c> value.</returns>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static bool IsNaN(double s)
        {
            return Double.IsNaN(s);
        }

        /// <summary>
        /// Performs a canonical Modulus operation, where the output is on the range [0, <paramref name="b"/>).
        /// </summary>
        /// <param name="a">The dividend, the primary input.</param>
        /// <param name="b">The divisor. The output is on the range [0, <paramref name="b"/>).</param>
        /// <returns>The resulting output.</returns>
        public static int PosMod(int a, int b)
        {
            int c = a % b;
            if ((c < 0 && b > 0) || (c > 0 && b < 0))
            {
                c += b;
            }
            return c;
        }

        /// <summary>
        /// Performs a canonical Modulus operation, where the output is on the range [0, <paramref name="b"/>).
        /// </summary>
        /// <param name="a">The dividend, the primary input.</param>
        /// <param name="b">The divisor. The output is on the range [0, <paramref name="b"/>).</param>
        /// <returns>The resulting output.</returns>
        public static float PosMod(float a, float b)
        {
            float c = a % b;
            if ((c < 0 && b > 0) || (c > 0 && b < 0))
            {
                c += b;
            }
            return c;
        }

        /// <summary>
        /// Performs a canonical Modulus operation, where the output is on the range [0, <paramref name="b"/>).
        /// </summary>
        /// <param name="a">The dividend, the primary input.</param>
        /// <param name="b">The divisor. The output is on the range [0, <paramref name="b"/>).</param>
        /// <returns>The resulting output.</returns>
        public static double PosMod(double a, double b)
        {
            double c = a % b;
            if ((c < 0 && b > 0) || (c > 0 && b < 0))
            {
                c += b;
            }
            return c;
        }

        /// <summary>
        /// Rounds <paramref name="s"/> to the nearest whole number.
        ///
        /// This is the same as <see cref="Round(float)"/>, but returns an <see langword="int"/>.
        /// </summary>
        /// <param name="s">The number to round.</param>
        /// <returns>The rounded number.</returns>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static int RoundToInt(float s)
        {
            return (int)MathF.Round(s);
        }

        /// <summary>
        /// Rounds <paramref name="s"/> to the nearest whole number.
        ///
        /// This is the same as <see cref="Round(double)"/>, but returns an <see langword="int"/>.
        /// </summary>
        /// <param name="s">The number to round.</param>
        /// <returns>The rounded number.</returns>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static int RoundToInt(double s)
        {
            return (int)Math.Round(s);
        }
    }
}
