﻿using KludgeBox.Structs;
using System;
using System.Globalization;
using System.Text;
using System.Text.RegularExpressions;

namespace KludgeBox
{
	public static class StringExtensions
	{
        private static readonly char[] _invalidFileNameCharacters = { ':', '/', '\\', '?', '*', '"', '|', '%', '<', '>' };
        
		/// <summary>
        /// Converts string to enum object
        /// </summary>
        /// <typeparam name="T">Type of enum</typeparam>
        /// <param name="value">String value to convert</param>
        /// <returns>Returns enum object</returns>
        public static T ToEnum<T>(this string value)
			where T : struct
		{
			return (T)Enum.Parse(typeof(T), value, true);
		}

		/// <summary>
		/// Removes non-alphanumeric characters from a string.
		/// </summary>
		/// <param name="str">The input string.</param>
		/// <returns>The modified string with non-alphanumeric characters removed.</returns>
		public static string Clear(this string str)
		{
			return Regex.Replace(str, "[^a-zA-Z0-9]", "");
		}

		/// <summary>
		/// Determines whether the specified string contains all the specified values.
		/// </summary>
		/// <param name="value">The string to search.</param>
		/// <param name="values">The values to search for.</param>
		/// <returns>true if all values are found in the string; otherwise, false.</returns>
		public static bool ContainsAll(this string value, params string[] values)
		{
			foreach (string one in values)
			{
				if (!value.Contains(one))
				{
					return false;
				}
			}
			return true;
		}


		/// <summary>
		/// Determines whether the specified string contains any of the specified values.
		/// </summary>
		/// <param name="value">The string to search.</param>
		/// <param name="values">The values to search for.</param>
		/// <returns>true if any value are found in the string; otherwise, false.</returns>
		public static bool ContainsAny(this string value, params string[] values)
		{
			foreach (string one in values)
			{
				if (value.Contains(one))
				{
					return true;
				}
			}
			return false;
		}


        /// <summary>
        /// Returns the index of the first occurrence of the specified string in this instance,
        /// or <c>-1</c>. Optionally, the starting search index can be specified, continuing
        /// to the end of the string.
        /// Note: If you just want to know whether a string contains a substring, use the
        /// <see cref="string.Contains(string)"/> method.
        /// </summary>
        /// <seealso cref="Find(string, char, int, bool)"/>
        /// <seealso cref="FindN(string, string, int)"/>
        /// <seealso cref="RFind(string, string, int, bool)"/>
        /// <seealso cref="RFindN(string, string, int)"/>
        /// <param name="instance">The string that will be searched.</param>
        /// <param name="what">The substring to find.</param>
        /// <param name="from">The search starting position.</param>
        /// <param name="caseSensitive">If <see langword="true"/>, the search is case sensitive.</param>
        /// <returns>The starting position of the substring, or -1 if not found.</returns>
        public static int Find(this string instance, string what, int from = 0, bool caseSensitive = true)
        {
            return instance.IndexOf(what, from,
                caseSensitive ? StringComparison.Ordinal : StringComparison.OrdinalIgnoreCase);
        }

        /// <summary>
        /// Find the first occurrence of a char. Optionally, the search starting position can be passed.
        /// </summary>
        /// <seealso cref="Find(string, string, int, bool)"/>
        /// <seealso cref="FindN(string, string, int)"/>
        /// <seealso cref="RFind(string, string, int, bool)"/>
        /// <seealso cref="RFindN(string, string, int)"/>
        /// <param name="instance">The string that will be searched.</param>
        /// <param name="what">The substring to find.</param>
        /// <param name="from">The search starting position.</param>
        /// <param name="caseSensitive">If <see langword="true"/>, the search is case sensitive.</param>
        /// <returns>The first instance of the char, or -1 if not found.</returns>
        public static int Find(this string instance, char what, int from = 0, bool caseSensitive = true)
        {
            if (caseSensitive)
                return instance.IndexOf(what, from);

            return CultureInfo.InvariantCulture.CompareInfo.IndexOf(instance, what, from, CompareOptions.OrdinalIgnoreCase);
        }

        /// <summary>
        /// Returns the index of the first case-insensitive occurrence of the specified string in this instance,
        /// or <c>-1</c>. Optionally, the starting search index can be specified, continuing
        /// to the end of the string.
        /// </summary>
        /// <seealso cref="Find(string, string, int, bool)"/>
        /// <seealso cref="Find(string, char, int, bool)"/>
        /// <seealso cref="RFind(string, string, int, bool)"/>
        /// <seealso cref="RFindN(string, string, int)"/>
        /// <param name="instance">The string that will be searched.</param>
        /// <param name="what">The substring to find.</param>
        /// <param name="from">The search starting position.</param>
        /// <returns>The starting position of the substring, or -1 if not found.</returns>
        public static int FindN(this string instance, string what, int from = 0)
        {
            return instance.IndexOf(what, from, StringComparison.OrdinalIgnoreCase);
        }


        /// <summary>
        /// Returns the index of the last occurrence of the specified string in this instance,
        /// or <c>-1</c>. Optionally, the starting search index can be specified, continuing to
        /// the beginning of the string.
        /// </summary>
        /// <seealso cref="Find(string, string, int, bool)"/>
        /// <seealso cref="Find(string, char, int, bool)"/>
        /// <seealso cref="FindN(string, string, int)"/>
        /// <seealso cref="RFindN(string, string, int)"/>
        /// <param name="instance">The string that will be searched.</param>
        /// <param name="what">The substring to search in the string.</param>
        /// <param name="from">The position at which to start searching.</param>
        /// <param name="caseSensitive">If <see langword="true"/>, the search is case sensitive.</param>
        /// <returns>The position at which the substring was found, or -1 if not found.</returns>
        public static int RFind(this string instance, string what, int from = -1, bool caseSensitive = true)
        {
            if (from == -1)
                from = instance.Length - 1;

            return instance.LastIndexOf(what, from,
                caseSensitive ? StringComparison.Ordinal : StringComparison.OrdinalIgnoreCase);
        }



        /// <summary>
        /// Returns the index of the last case-insensitive occurrence of the specified string in this instance,
        /// or <c>-1</c>. Optionally, the starting search index can be specified, continuing to
        /// the beginning of the string.
        /// </summary>
        /// <seealso cref="Find(string, string, int, bool)"/>
        /// <seealso cref="Find(string, char, int, bool)"/>
        /// <seealso cref="FindN(string, string, int)"/>
        /// <seealso cref="RFind(string, string, int, bool)"/>
        /// <param name="instance">The string that will be searched.</param>
        /// <param name="what">The substring to search in the string.</param>
        /// <param name="from">The position at which to start searching.</param>
        /// <returns>The position at which the substring was found, or -1 if not found.</returns>
        public static int RFindN(this string instance, string what, int from = -1)
        {
            if (from == -1)
                from = instance.Length - 1;

            return instance.LastIndexOf(what, from, StringComparison.OrdinalIgnoreCase);
        }

        /// <summary>
        /// Returns the extension without the leading period character (<c>.</c>)
        /// if the string is a valid file name or path. If the string does not contain
        /// an extension, returns an empty string instead.
        /// </summary>
        /// <param name="instance">The path to a file.</param>
        /// <returns>The extension of the file or an empty string.</returns>
        public static string GetExtension(this string instance)
        {
            int pos = instance.RFind(".");

            if (pos < 0)
                return instance;

            return instance.Substring(pos + 1);
        }

        /// <summary>
        ///     Formats a large number to 5 digits (2 decimal places) with a suffix. (e.g. 123,456,789,123 (123 Billion) ->
        ///     123.45B)
        /// </summary>
        /// <param name="numberToFormat"> The number to format. </param>
        /// <param name="decimalPlaces"> The number of decimal places to include - <i> defaults to <c> 2 </c> </i> </param>
        /// <returns> A <see cref="string" />. </returns>
        public static string FormatNumber(this long numberToFormat, int decimalPlaces = 2)
		{
			// Get the default string representation of the numberToFormat.
			string numberString = numberToFormat.ToString();

			foreach (NumberSuffix suffix in Enum.GetValues<NumberSuffix>())
			{
				// Assign the amount of digits to base 10.
				double currentValue = 1 * Math.Pow(10, (int)suffix * 3);

				// Get the suffix value.
				string suffixValue = Enum.GetName(typeof(NumberSuffix), (int)suffix);

				// If the suffix is the placeholder, set it to an empty string.
				if ((int)suffix == 0) { suffixValue = string.Empty; }

				// Set the return value to a rounded value with the suffix.
				if (numberToFormat >= currentValue)
				{
					numberString = $"{Math.Round(numberToFormat / currentValue, decimalPlaces, MidpointRounding.ToEven)}{suffixValue}";
				}
			}

			return numberString;
		}

		/// <summary> Suffixes for numbers based on how many digits they have left of the decimal point. </summary>
		/// <remarks> The order of the suffixes matters! </remarks>
		private enum NumberSuffix
		{
			/// <summary> A placeholder if the value is under 1 thousand </summary>
			P,
			/// <summary> Thousand </summary>
			K,
			/// <summary> Million </summary>
			M,
			/// <summary> Billion </summary>
			B,
			/// <summary> Trillion </summary>
			T,
			/// <summary> Quadrillion </summary>
			Q
		}


		public static string FirstNumber(this double number)
		{
			string formatted;

			if (number >= 1 || number == 0)
			{
				formatted = ((int)number).ToString();
			}
			else if (number < 1 && number > 0)
			{
				formatted = number.ToString("F2");
			}
			else
			{
				formatted = number.ToString();
			}

			return formatted;
		}

		/// <summary>
		/// Determines whether the specified input string matches the specified pattern using regular expressions.
		/// </summary>
		/// <param name="input">
		///     The input string to be tested for a match with the pattern.
		/// </param>
		/// <param name="pattern">
		///     The pattern to be used for matching. Wildcard characters such as '*' are allowed.
		/// </param>
		/// <returns>
		///     <c>true</c> if the input string matches the pattern; otherwise, <c>false</c>.
		/// </returns>
		public static bool IsMatch(this string input, string pattern)
		{
			string escapedPattern = Regex.Escape(pattern);

			escapedPattern = escapedPattern.Replace("\\*", ".*");

			return Regex.IsMatch(input, $"^{escapedPattern}$");
		}

        /// <summary>
        /// Decodes a hexadecimal string.
        /// </summary>
        /// <param name="instance">The hexadecimal string.</param>
        /// <returns>The byte array representation of this string.</returns>
        public static byte[] HexDecode(this string instance)
        {
            if (instance.Length % 2 != 0)
            {
                throw new ArgumentException("Hexadecimal string of uneven length.", nameof(instance));
            }
            int len = instance.Length / 2;
            byte[] ret = new byte[len];
            for (int i = 0; i < len; i++)
            {
                ret[i] = (byte)int.Parse(instance.AsSpan(i * 2, 2), NumberStyles.AllowHexSpecifier);
            }
            return ret;
        }

        /// <summary>
        /// Returns a hexadecimal representation of this byte as a string.
        /// </summary>
        /// <param name="b">The byte to encode.</param>
        /// <returns>The hexadecimal representation of this byte.</returns>
        internal static string HexEncode(this byte b)
        {
            string ret = string.Empty;

            for (int i = 0; i < 2; i++)
            {
                char c;
                int lv = b & 0xF;

                if (lv < 10)
                {
                    c = (char)('0' + lv);
                }
                else
                {
                    c = (char)('a' + lv - 10);
                }

                b >>= 4;
                ret = c + ret;
            }

            return ret;
        }

        /// <summary>
        /// Returns a hexadecimal representation of this byte array as a string.
        /// </summary>
        /// <param name="bytes">The byte array to encode.</param>
        /// <returns>The hexadecimal representation of this byte array.</returns>
        public static string HexEncode(this byte[] bytes)
        {
            string ret = string.Empty;

            foreach (byte b in bytes)
            {
                ret += b.HexEncode();
            }

            return ret;
        }

        /// <summary>
        /// Converts a string containing a hexadecimal number into an integer.
        /// Hexadecimal strings can either be prefixed with <c>0x</c> or not,
        /// and they can also start with a <c>-</c> before the optional prefix.
        /// </summary>
        /// <param name="instance">The string to convert.</param>
        /// <returns>The converted string.</returns>
        public static int HexToInt(this string instance)
        {
            if (instance.Length == 0)
            {
                return 0;
            }

            int sign = 1;

            if (instance[0] == '-')
            {
                sign = -1;
                instance = instance.Substring(1);
            }

            if (instance.StartsWith("0x"))
            {
                instance = instance.Substring(2);
            }

            return sign * int.Parse(instance, NumberStyles.HexNumber);
        }

        /// <summary>
        /// Returns <see langword="true"/> if this string is free from characters that
        /// aren't allowed in file names.
        /// </summary>
        /// <param name="instance">The string to check.</param>
        /// <returns>If the string contains a valid file name.</returns>
        public static bool IsValidFileName(this string instance)
        {
            var stripped = instance.Trim();
            if (instance != stripped)
                return false;

            if (string.IsNullOrEmpty(stripped))
                return false;

            return instance.IndexOfAny(_invalidFileNameCharacters) == -1;
        }

        /// <summary>
        /// Returns <see langword="true"/> if this string is a valid identifier.
        /// A valid identifier may contain only letters, digits and underscores (<c>_</c>)
        /// and the first character may not be a digit.
        /// </summary>
        /// <param name="instance">The string to check.</param>
        /// <returns>If the string contains a valid identifier.</returns>
        public static bool IsValidIdentifier(this string instance)
        {
            int len = instance.Length;

            if (len == 0)
                return false;

            if (instance[0] >= '0' && instance[0] <= '9')
                return false; // Identifiers cannot start with numbers.

            for (int i = 0; i < len; i++)
            {
                bool validChar = instance[i] == '_' ||
                    (instance[i] >= 'a' && instance[i] <= 'z') ||
                    (instance[i] >= 'A' && instance[i] <= 'Z') ||
                    (instance[i] >= '0' && instance[i] <= '9');

                if (!validChar)
                    return false;
            }

            return true;
        }

        /// <summary>
        /// Returns <see langword="true"/> if this string contains only a well-formatted
        /// IPv4 or IPv6 address. This method considers reserved IP addresses such as
        /// <c>0.0.0.0</c> as valid.
        /// </summary>
        /// <param name="instance">The string to check.</param>
        /// <returns>If the string contains a valid IP address.</returns>
        public static bool IsValidIPAddress(this string instance)
        {
            if (instance.Contains(':'))
            {
                string[] ip = instance.Split(':');

                for (int i = 0; i < ip.Length; i++)
                {
                    string n = ip[i];
                    if (n.Length == 0)
                        continue;

                    if (n.IsValidHexNumber(withPrefix: false))
                    {
                        long nint = n.HexToInt();
                        if (nint < 0 || nint > 0xffff)
                            return false;

                        continue;
                    }

                    if (!n.IsValidIPAddress())
                        return false;
                }
            }
            else
            {
                string[] ip = instance.Split('.');

                if (ip.Length != 4)
                    return false;

                for (int i = 0; i < ip.Length; i++)
                {
                    string n = ip[i];
                    if (!n.IsValidInt())
                        return false;

                    int val = n.ToInt();
                    if (val < 0 || val > 255)
                        return false;
                }
            }

            return true;
        }

        /// <summary>
        /// Returns <see langword="true"/> if this string contains a valid <see langword="float"/>.
        /// This is inclusive of integers, and also supports exponents.
        /// </summary>
        /// <param name="instance">The string to check.</param>
        /// <returns>If the string contains a valid floating point number.</returns>
        public static bool IsValidFloat(this string instance)
        {
            return float.TryParse(instance, out _);
        }

        /// <summary>
        /// Returns <see langword="true"/> if this string contains a valid hexadecimal number.
        /// If <paramref name="withPrefix"/> is <see langword="true"/>, then a validity of the
        /// hexadecimal number is determined by <c>0x</c> prefix, for instance: <c>0xDEADC0DE</c>.
        /// </summary>
        /// <param name="instance">The string to check.</param>
        /// <param name="withPrefix">If the string must contain the <c>0x</c> prefix to be valid.</param>
        /// <returns>If the string contains a valid hexadecimal number.</returns>
        public static bool IsValidHexNumber(this string instance, bool withPrefix = false)
        {
            if (string.IsNullOrEmpty(instance))
                return false;

            int from = 0;
            if (instance.Length != 1 && instance[0] == '+' || instance[0] == '-')
            {
                from++;
            }

            if (withPrefix)
            {
                if (instance.Length < 3)
                    return false;
                if (instance[from] != '0' || instance[from + 1] != 'x')
                    return false;
                from += 2;
            }

            for (int i = from; i < instance.Length; i++)
            {
                char c = instance[i];
                if (IsHexDigit(c))
                    continue;

                return false;
            }

            return true;

            static bool IsHexDigit(char c)
            {
                return char.IsDigit(c) || (c >= 'a' && c <= 'f') || (c >= 'A' && c <= 'F');
            }
        }

        /// <summary>
        /// Returns <see langword="true"/> if this string contains a valid color in hexadecimal
        /// HTML notation. Other HTML notations such as named colors or <c>hsl()</c> aren't
        /// considered valid by this method and will return <see langword="false"/>.
        /// </summary>
        /// <param name="instance">The string to check.</param>
        /// <returns>If the string contains a valid HTML color.</returns>
        public static bool IsValidHtmlColor(this string instance)
        {
            return KColor.HtmlIsValid(instance);
        }

        /// <summary>
        /// Returns <see langword="true"/> if this string contains a valid <see langword="int"/>.
        /// </summary>
        /// <param name="instance">The string to check.</param>
        /// <returns>If the string contains a valid integer.</returns>
        public static bool IsValidInt(this string instance)
        {
            return int.TryParse(instance, out _);
        }

        /// <summary>
        /// Converts a string, containing a decimal number, into a <see langword="float" />.
        /// </summary>
        /// <seealso cref="ToInt(string)"/>
        /// <param name="instance">The string to convert.</param>
        /// <returns>The number representation of the string.</returns>
        public static float ToFloat(this string instance)
        {
            return float.Parse(instance);
        }

        /// <summary>
        /// Converts a string, containing an integer number, into an <see langword="int" />.
        /// </summary>
        /// <seealso cref="ToFloat(string)"/>
        /// <param name="instance">The string to convert.</param>
        /// <returns>The number representation of the string.</returns>
        public static int ToInt(this string instance)
        {
            return int.Parse(instance);
        }

        /// <summary>
        /// Removes a given string from the start if it starts with it or leaves the string unchanged.
        /// </summary>
        /// <param name="instance">The string to remove the prefix from.</param>
        /// <param name="prefix">The string to remove from the start.</param>
        /// <returns>A copy of the string with the prefix string removed from the start.</returns>
        public static string TrimPrefix(this string instance, string prefix)
        {
            if (instance.StartsWith(prefix))
                return instance.Substring(prefix.Length);

            return instance;
        }

        /// <summary>
        /// Removes a given string from the end if it ends with it or leaves the string unchanged.
        /// </summary>
        /// <param name="instance">The string to remove the suffix from.</param>
        /// <param name="suffix">The string to remove from the end.</param>
        /// <returns>A copy of the string with the suffix string removed from the end.</returns>
        public static string TrimSuffix(this string instance, string suffix)
        {
            if (instance.EndsWith(suffix))
                return instance.Substring(0, instance.Length - suffix.Length);

            return instance;
        }

        /// <summary>
        /// Returns a copy of the string with special characters escaped using the C language standard.
        /// </summary>
        /// <param name="instance">The string to escape.</param>
        /// <returns>The escaped string.</returns>
        public static string CEscape(this string instance)
        {
            var sb = new StringBuilder(instance);

            sb.Replace("\\", "\\\\");
            sb.Replace("\a", "\\a");
            sb.Replace("\b", "\\b");
            sb.Replace("\f", "\\f");
            sb.Replace("\n", "\\n");
            sb.Replace("\r", "\\r");
            sb.Replace("\t", "\\t");
            sb.Replace("\v", "\\v");
            sb.Replace("\'", "\\'");
            sb.Replace("\"", "\\\"");

            return sb.ToString();
        }

        /// <summary>
        /// Returns a copy of the string with escaped characters replaced by their meanings
        /// according to the C language standard.
        /// </summary>
        /// <param name="instance">The string to unescape.</param>
        /// <returns>The unescaped string.</returns>
        public static string CUnescape(this string instance)
        {
            var sb = new StringBuilder(instance);

            sb.Replace("\\a", "\a");
            sb.Replace("\\b", "\b");
            sb.Replace("\\f", "\f");
            sb.Replace("\\n", "\n");
            sb.Replace("\\r", "\r");
            sb.Replace("\\t", "\t");
            sb.Replace("\\v", "\v");
            sb.Replace("\\'", "\'");
            sb.Replace("\\\"", "\"");
            sb.Replace("\\\\", "\\");

            return sb.ToString();
        }
    }
}
