﻿using KludgeBox.Structs;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace KludgeBox
{
    public static class StructExtensions
    {
        #region Vectors
        /// <summary>
        /// Returns a Vector2 instance with specified x and y values.
        /// </summary>
        public static Vector2F Vec2(float x, float y) { return new Vector2F(x, y); }

        /// <summary>
        /// Returns a Vector2 instance with all values set to the specified value.
        /// </summary>
        public static Vector2F Vec2(float n) { return new Vector2F(n, n); }

        /// <summary>
        /// Returns a Vector2 instance with all values set to zero.
        /// </summary>
        public static Vector2F Vec2() { return Vector2F.Zero; }


        /// <summary>
        /// Returns a Vector2I instance with specified x and y values.
        /// </summary>
        public static Vector2I Vec2I(int x, int y) { return new Vector2I(x, y); }

        /// <summary>
        /// Returns a Vector2I instance with all values set to the specified value.
        /// </summary>
        public static Vector2I Vec2I(int n) { return new Vector2I(n, n); }

        /// <summary>
        /// Returns a Vector2I instance with all values set to zero.
        /// </summary>
        public static Vector2I Vec2I() { return Vector2I.Zero; }



        /// <summary>
        /// Extends the functionality of the Vector2 class by spinning it around a random angle.
        /// </summary>
        /// <param name="vector">The Vector2 instance to spin.</param>
        /// <returns>A new Vector2 instance representing the original vector rotated by a random angle.</returns>
        public static Vector2F Spin(this Vector2F vector)
        {
            float angle = Rand.Float * 2 * Maths.PI;
            vector = vector.Rotated(angle);

            return vector;
        }
        #endregion

        #region Rect Tools
        /// <summary>
        /// Gets the X-coordinate of the top-left corner of the specified Rect2.
        /// </summary>
        /// <param name="rect">The Rect2 object.</param>
        /// <returns>The X-coordinate of the top-left corner.</returns>
        public static float X(this Rect2F rect) => rect.Position.X;

        /// <summary>
        /// Gets the Y-coordinate of the top-left corner of the specified Rect2.
        /// </summary>
        /// <param name="rect">The Rect2 object.</param>
        /// <returns>The Y-coordinate of the top-left corner.</returns>
        public static float Y(this Rect2F rect) => rect.Position.Y;

        /// <summary>
        /// Gets the width of the specified Rect2.
        /// </summary>
        /// <param name="rect">The Rect2 object.</param>
        /// <returns>The width of the Rect2.</returns>
        public static float Width(this Rect2F rect) => rect.Size.X;

        /// <summary>
        /// Gets the height of the specified Rect2.
        /// </summary>
        /// <param name="rect">The Rect2 object.</param>
        /// <returns>The height of the Rect2.</returns>
        public static float Height(this Rect2F rect) => rect.Size.Y;

        public static float GetCircumradius(this Rect2F rect)
        {
            return Maths.Sqrt(rect.Width() * rect.Width() + rect.Height() * rect.Height()) / 2;
        }

        public static Circle2F GetCircumcircle(this Rect2F rect)
        {
            var radius = rect.GetCircumradius();
            return new Circle2F(rect.GetCenter(), radius);
        }

        /// <summary>
        /// Returns a new Rect2 representing the left half of the specified Rect2.
        /// </summary>
        /// <param name="rect">The Rect2 object.</param>
        /// <returns>A new Rect2 representing the left half of the original Rect2.</returns>
        public static Rect2F LeftHalf(this Rect2F rect)
        {
            return new Rect2F(rect.Position, Vec2(rect.Width() / 2f, rect.Height()));
        }

        /// <summary>
        /// Returns a new Rect2 representing a portion of the left side of the specified Rect2, based on the percentage provided.
        /// </summary>
        /// <param name="rect">The Rect2 object.</param>
        /// <param name="percentage">The percentage of the Rect2's width to include in the new Rect2. Value should be between 0 and 1.</param>
        /// <returns>A new Rect2 representing a portion of the left side of the original Rect2.</returns>
        public static Rect2F LeftPart(this Rect2F rect, float percentage)
        {
            return new Rect2F(rect.Position, Vec2(rect.Width() * percentage, rect.Height()));
        }

        /// <summary>
        /// Returns a new Rect2 representing a portion of the left side of the specified Rect2, based on the width provided in pixels.
        /// </summary>
        /// <param name="rect">The Rect2 object.</param>
        /// <param name="width">The width of the new Rect2 in pixels.</param>
        /// <returns>A new Rect2 representing a portion of the left side of the original Rect2.</returns>
        public static Rect2F LeftPartPixels(this Rect2F rect, float width)
        {
            return new Rect2F(rect.Position, Vec2(width, rect.Height()));
        }

        /// <summary>
        /// Returns a new <see cref="Rect2F"/> representing the right half of the original rectangle.
        /// </summary>
        /// <param name="rect">The original rectangle.</param>
        /// <returns>A new <see cref="Rect2F"/> representing the right half of the original rectangle.</returns>
        public static Rect2F RightHalf(this Rect2F rect)
        {
            return new Rect2F(
                Vec2(rect.X() + rect.Width() / 2f, rect.Y()),
                Vec2(rect.Width() / 2f, rect.Height())
            );
        }

        /// <summary>
        /// Returns a new <see cref="Rect2F"/> representing a portion of the original rectangle from the right side, based on the given percentage.
        /// </summary>
        /// <param name="rect">The original rectangle.</param>
        /// <param name="percentage">The percentage of the original rectangle's width to include in the new rectangle (0.0 - 1.0).</param>
        /// <returns>A new <see cref="Rect2F"/> representing a portion of the original rectangle from the right side.</returns>
        public static Rect2F RightPart(this Rect2F rect, float percentage)
        {
            return new Rect2F(
                Vec2(rect.X() + rect.Width() * (1f - percentage), rect.Y()),
                Vec2(rect.Width() * percentage, rect.Height())
            );
        }

        /// <summary>
        /// Returns a new <see cref="Rect2F"/> representing a portion of the original rectangle from the right side, based on the given width in pixels.
        /// </summary>
        /// <param name="rect">The original rectangle.</param>
        /// <param name="width">The width in pixels of the new rectangle.</param>
        /// <returns>A new <see cref="Rect2F"/> representing a portion of the original rectangle from the right side.</returns>
        public static Rect2F RightPartPixels(this Rect2F rect, float width)
        {
            return new Rect2F(
                Vec2(rect.X() + rect.Width() - width, rect.Y()),
                Vec2(width, rect.Height())
            );
        }

        /// <summary>
        /// Returns a new <see cref="Rect2F"/> representing the top half of the original rectangle.
        /// </summary>
        /// <param name="rect">The original rectangle.</param>
        /// <returns>A new <see cref="Rect2F"/> representing the top half of the original rectangle.</returns>
        public static Rect2F TopHalf(this Rect2F rect)
        {
            return new Rect2F(
                rect.Position,
                Vec2(rect.Width(), rect.Height() / 2f)
            );
        }

        /// <summary>
        /// Returns a new <see cref="Rect2F"/> representing a portion of the original rectangle from the top side, based on the given percentage.
        /// </summary>
        /// <param name="rect">The original rectangle.</param>
        /// <param name="percentage">The percentage of the original rectangle's height to include in the new rectangle (0.0 - 1.0).</param>
        /// <returns>A new <see cref="Rect2F"/> representing a portion of the original rectangle from the top side.</returns>
        public static Rect2F TopPart(this Rect2F rect, float percentage)
        {
            return new Rect2F(
                rect.Position,
                Vec2(rect.Width(), rect.Height() * percentage)
            );
        }

        /// <summary>
        /// Returns a new <see cref="Rect2F"/> representing a portion of the original rectangle from the top side, based on the given height in pixels.
        /// </summary>
        /// <param name="rect">The original rectangle.</param>
        /// <param name="height">The height in pixels of the new rectangle.</param>
        /// <returns>A new <see cref="Rect2F"/> representing a portion of the original rectangle from the top side.</returns>
        public static Rect2F TopPartPixels(this Rect2F rect, float height)
        {
            return new Rect2F(
                rect.Position,
                Vec2(rect.Width(), height)
            );
        }

        /// <summary>
        /// Returns a new <see cref="Rect2F"/> representing the bottom half of the original rectangle.
        /// </summary>
        /// <param name="rect">The original rectangle.</param>
        /// <returns>A new <see cref="Rect2F"/> representing the bottom half of the original rectangle.</returns>
        public static Rect2F BottomHalf(this Rect2F rect)
        {
            return new Rect2F(
                Vec2(rect.X(), rect.Y() + rect.Height() / 2f),
                Vec2(rect.Width(), rect.Height() / 2f)
            );
        }

        /// <summary>
        /// Returns a new <see cref="Rect2F"/> representing a portion of the original rectangle from the bottom side, based on the given percentage.
        /// </summary>
        /// <param name="rect">The original rectangle.</param>
        /// <param name="percentage">The percentage of the original rectangle's height to include in the new rectangle (0.0 - 1.0).</param>
        /// <returns>A new <see cref="Rect2F"/> representing a portion of the original rectangle from the bottom side.</returns>
        public static Rect2F BottomPart(this Rect2F rect, float percentage)
        {
            return new Rect2F(
                Vec2(rect.X(), rect.Y() + rect.Height() * (1f - percentage)),
                Vec2(rect.Width(), rect.Height() * percentage)
            );
        }

        /// <summary>
        /// Returns a new <see cref="Rect2F"/> representing a portion of the original rectangle from the bottom side, based on the given height in pixels.
        /// </summary>
        /// <param name="rect">The original rectangle.</param>
        /// <param name="height">The height in pixels of the new rectangle.</param>
        /// <returns>A new <see cref="Rect2F"/> representing a portion of the original rectangle from the bottom side.</returns>
        public static Rect2F BottomPartPixels(this Rect2F rect, float height)
        {
            return new Rect2F(
                Vec2(rect.X(), rect.Y() + rect.Height() - height),
                Vec2(rect.Width(), height)
            );
        }

        #endregion

        #region Color
        /// <summary>
        /// Converts a grayscale value to a Color object.
        /// </summary>
        /// <param name="value">The grayscale value to be used for all color components.</param>
        /// <returns>A Color object representing the grayscale value.</returns>
        /// <remarks>
        /// The grayscale value will be applied to all color components (R, G, and B) of the resulting Color object.
        /// </remarks>
        public static KColor Col(float value) => new KColor(value, value, value);

        /// <summary>
        /// Extension method that converts a color to grayscale.
        /// </summary>
        /// <param name="kColor">The color to convert.</param>
        /// <returns>A grayscale color.</returns>
        public static KColor Grayscale(this KColor kColor)
        {
            float gray = kColor.GrayscaleValue();
            return new KColor(gray, gray, gray, kColor.A);
        }

        /// <summary>
        /// Calculates the grayscale value of a color.
        /// </summary>
        /// <param name="kColor">The color to calculate the grayscale value for.</param>
        /// <returns>The grayscale value of the color.</returns>
        public static float GrayscaleValue(this KColor kColor)
        {
            return 0.299f * kColor.R + 0.587f * kColor.G + 0.114f * kColor.B;
        }

        /// <summary>
        /// Creates a new color with RGB values set to 0.
        /// </summary>
        /// <returns>A color with RGB values set to 0.</returns>
        public static KColor Col() => new KColor(0, 0, 0);

        /// <summary>
        /// Creates a new color with the specified RGB and alpha values.
        /// </summary>
        /// <param name="r">The red component of the color.</param>
        /// <param name="g">The green component of the color.</param>
        /// <param name="b">The blue component of the color.</param>
        /// <param name="a">The alpha component of the color (default is 1).</param>
        /// <returns>A new color with the specified RGB and alpha values.</returns>
        public static KColor Col(float r, float g, float b, float a = 1) => new KColor(r, g, b, a);

        public static KColor Darken(this KColor kColor, float amount)
        {
            amount = Math.Clamp(amount, 0, 1);
            return kColor with
            {
                R = kColor.R * (1 - amount),
                G = kColor.G * (1 - amount),
                B = kColor.B * (1 - amount)
            };
        }

        /// <summary>
        /// Lightens the specified color by the given amount.
        /// </summary>
        /// <param name="kColor">The color to lighten.</param>
        /// <param name="amount">The amount by which to lighten the color. Should be a value between 0 and 1.</param>
        /// <returns>The lightened color.</returns>
        public static KColor Lighten(this KColor kColor, float amount)
        {
            amount = Math.Clamp(amount, 0, 1);
            return kColor with
            {
                R = kColor.R + (1 - kColor.R) * amount,
                G = kColor.G + (1 - kColor.G) * amount,
                B = kColor.B + (1 - kColor.B) * amount
            };
        }
        #endregion
    }
}
