﻿using System;
using Newtonsoft.Json;

namespace KludgeBox.Structs;

/// <summary>
/// Represents a Circle2D in 2D space defined by its center <see cref="Position"/> and <see cref="Radius"/>.
/// </summary>
public partial struct Circle2D : IEquatable<Circle2D>
{
    /// <summary>
    /// The center position of the Circle2D in 2D space.
    /// </summary>
    public Vector2D Position = new();

    /// <summary>
    /// The radius of the Circle2D.
    /// </summary>
    public double Radius = 0f;

    /// <summary>
    /// Gets or sets the X coordinate of the Circle2D's center position.
    /// </summary>
    [JsonIgnore]
    public double X
    {
        get => Position.X;
        set => Position = Position with { X = value };
    }

    /// <summary>
    /// Gets or sets the Y coordinate of the Circle2D's center position.
    /// </summary>
    [JsonIgnore]
    public double Y
    {
        get => Position.Y;
        set => Position = Position with { Y = value };
    }

    /// <summary>
    /// Gets or sets the diameter of the Circle2D.
    /// </summary>
    [JsonIgnore]
    public double Diameter
    {
        get => Radius * 2;
        set => Radius = value / 2;
    }

    /// <summary>
    /// Calculates and returns the area of the Circle2D.
    /// </summary>
    [JsonIgnore]
    public double Area => Maths.PI * Maths.Pow(Radius, 2);

    /// <summary>
    /// Creates a new instance of the <see cref="Circle2D"/> struct with the specified position and radius.
    /// </summary>
    /// <param name="position">The center position of the Circle2D.</param>
    /// <param name="radius">The radius of the Circle2D.</param>
    public Circle2D(Vector2D position, double radius)
    {
        Position = position;
        Radius = radius;
    }

    public Circle2D()
    {
    }

    /// <summary>
    /// Creates a new instance of the <see cref="Circle2D"/> struct with the specified radius and default position (0, 0).
    /// </summary>
    /// <param name="radius">The radius of the Circle2D.</param>
    public Circle2D(double radius)
    {
        Radius = radius;
    }

    /// <summary>
    /// Gets the bounding rectangle (outer box) that encloses the Circle2D.
    /// </summary>
    /// <returns>A <see cref="Rect2D"/> representing the bounding rectangle of the Circle2D.</returns>
    public Rect2D GetBoundingBox()
    {
        var topLeft = new Vector2D(Position.X - Radius, Position.Y - Radius);
        var size = new Vector2D(Radius * 2, Radius * 2);
        var box = new Rect2D();
        box.Size = size;
        box.Position = topLeft;
        return box;
    }

    /// <summary>
    /// Checks if this Circle2D intersects with another Circle2D.
    /// </summary>
    /// <param name="other">The other Circle2D to check for intersection.</param>
    /// <returns>True if the CircleDs intersect; otherwise, false.</returns>
    public bool IntersectsWith(Circle2D other)
    {
        var distanceSquared = (Position.X - other.Position.X) * (Position.X - other.Position.X) +
                              (Position.Y - other.Position.Y) * (Position.Y - other.Position.Y);

        var radiusSum = Radius + other.Radius;
        return distanceSquared <= radiusSum * radiusSum;
    }

    /// <summary>
    /// Checks if this Circle2D intersects with a rectangle.
    /// </summary>
    /// <param name="rect">The rectangle to check for intersection.</param>
    /// <returns>True if the Circle2D intersects with the rectangle; otherwise, false.</returns>
    public bool IntersectsWith(Rect2D rect)
    {
        var closestX = Math.Max(rect.Position.X, Math.Min(Position.X, rect.Position.X + rect.Size.X));
        var closestY = Math.Max(rect.Position.Y, Math.Min(Position.Y, rect.Position.Y + rect.Size.Y));

        var distanceSquared = (Position.X - closestX) * (Position.X - closestX) +
                              (Position.Y - closestY) * (Position.Y - closestY);

        return distanceSquared <= Radius * Radius;
    }

    public bool Equals(Circle2D other)
    {
        if (Position.Equals(other.Position)) return Radius.Equals(other.Radius);
        return false;
    }

    public bool IsEqualsApprox(Circle2D other)
    {
        if (Position.IsEqualApprox(other.Position)) return Radius.IsEqualApprox(other.Radius);

        return false;
    }

    public override string ToString()
    {
        return JsonConvert.SerializeObject(this);
    }

    public static Circle2D Parse(string text)
    {
        return JsonConvert.DeserializeObject<Circle2D>(text);
    }
}