﻿using KludgeBox.Enums;
using Newtonsoft.Json;
using System;

namespace KludgeBox.Structs;

/// <summary>
/// 2D axis-aligned bounding box. Rect2 consists of a position, a size, and
/// several utility functions. It is typically used for fast overlap tests.
/// </summary>
[Serializable]
public partial struct Rect2D : IEquatable<Rect2D>
{
    private Vector2D _position;
    private Vector2D _size;

    /// <summary>
    /// Beginning corner. Typically has values lower than <see cref="End"/>.
    /// </summary>
    /// <value>Directly uses a private field.</value>
    public Vector2D Position
    {
        readonly get => _position;
        set => _position = value;
    }

    /// <summary>
    /// Size from <see cref="Position"/> to <see cref="End"/>. Typically all components are positive.
    /// If the size is negative, you can use <see cref="Abs"/> to fix it.
    /// </summary>
    /// <value>Directly uses a private field.</value>
    public Vector2D Size
    {
        readonly get => _size;
        set => _size = value;
    }

    /// <summary>
    /// Ending corner. This is calculated as <see cref="Position"/> plus <see cref="Size"/>.
    /// Setting this value will change the size.
    /// </summary>
    /// <value>
    /// Getting is equivalent to <paramref name="value"/> = <see cref="Position"/> + <see cref="Size"/>,
    /// setting is equivalent to <see cref="Size"/> = <paramref name="value"/> - <see cref="Position"/>
    /// </value>
    [JsonIgnore]
    public Vector2D End
    {
        readonly get => _position + _size;
        set => _size = value - _position;
    }

    /// <summary>
    /// The area of this <see cref="Rect2D"/>.
    /// See also <see cref="HasArea"/>.
    /// </summary>
    [JsonIgnore]
    public readonly double Area => _size.X * _size.Y;

    /// <summary>
    /// Returns a <see cref="Rect2D"/> with equivalent position and size, modified so that
    /// the top-left corner is the origin and width and height are positive.
    /// </summary>
    /// <returns>The modified <see cref="Rect2D"/>.</returns>
    public readonly Rect2D Abs()
    {
        var end = End;
        var topLeft = new Vector2D(Maths.Min(_position.X, end.X), Maths.Min(_position.Y, end.Y));
        return new Rect2D(topLeft, _size.Abs());
    }

    /// <summary>
    /// Returns the intersection of this <see cref="Rect2D"/> and <paramref name="b"/>.
    /// If the rectangles do not intersect, an empty <see cref="Rect2D"/> is returned.
    /// </summary>
    /// <param name="b">The other <see cref="Rect2D"/>.</param>
    /// <returns>
    /// The intersection of this <see cref="Rect2D"/> and <paramref name="b"/>,
    /// or an empty <see cref="Rect2D"/> if they do not intersect.
    /// </returns>
    public readonly Rect2D Intersection(Rect2D b)
    {
        var newRect = b;

        if (!Intersects(newRect)) return new Rect2D();

        newRect._position.X = Maths.Max(b._position.X, _position.X);
        newRect._position.Y = Maths.Max(b._position.Y, _position.Y);

        var bEnd = b._position + b._size;
        var end = _position + _size;

        newRect._size.X = Maths.Min(bEnd.X, end.X) - newRect._position.X;
        newRect._size.Y = Maths.Min(bEnd.Y, end.Y) - newRect._position.Y;

        return newRect;
    }

    /// <summary>
    /// Returns <see langword="true"/> if this <see cref="Rect2D"/> is finite, by calling
    /// <see cref="Maths.IsFinite(double)"/> on each component.
    /// </summary>
    /// <returns>Whether this vector is finite or not.</returns>
    public bool IsFinite()
    {
        return _position.IsFinite() && _size.IsFinite();
    }

    /// <summary>
    /// Returns <see langword="true"/> if this <see cref="Rect2D"/> completely encloses another one.
    /// </summary>
    /// <param name="b">The other <see cref="Rect2D"/> that may be enclosed.</param>
    /// <returns>
    /// A <see langword="bool"/> for whether or not this <see cref="Rect2D"/> encloses <paramref name="b"/>.
    /// </returns>
    public readonly bool Encloses(Rect2D b)
    {
        return b._position.X >= _position.X && b._position.Y >= _position.Y &&
               b._position.X + b._size.X < _position.X + _size.X &&
               b._position.Y + b._size.Y < _position.Y + _size.Y;
    }

    /// <summary>
    /// Returns this <see cref="Rect2D"/> expanded to include a given point.
    /// </summary>
    /// <param name="to">The point to include.</param>
    /// <returns>The expanded <see cref="Rect2D"/>.</returns>
    public readonly Rect2D Expand(Vector2D to)
    {
        var expanded = this;

        var begin = expanded._position;
        var end = expanded._position + expanded._size;

        if (to.X < begin.X) begin.X = to.X;
        if (to.Y < begin.Y) begin.Y = to.Y;

        if (to.X > end.X) end.X = to.X;
        if (to.Y > end.Y) end.Y = to.Y;

        expanded._position = begin;
        expanded._size = end - begin;

        return expanded;
    }

    /// <summary>
    /// Returns the center of the <see cref="Rect2D"/>, which is equal
    /// to <see cref="Position"/> + (<see cref="Size"/> / 2).
    /// </summary>
    /// <returns>The center.</returns>
    public readonly Vector2D GetCenter()
    {
        return _position + _size * 0.5f;
    }

    /// <summary>
    /// Returns a copy of the <see cref="Rect2D"/> grown by the specified amount
    /// on all sides.
    /// </summary>
    /// <seealso cref="GrowIndividual(double, double, double, double)"/>
    /// <seealso cref="GrowSide(Side, double)"/>
    /// <param name="by">The amount to grow by.</param>
    /// <returns>The grown <see cref="Rect2D"/>.</returns>
    public readonly Rect2D Grow(double by)
    {
        var g = this;

        g._position.X -= by;
        g._position.Y -= by;
        g._size.X += by * 2;
        g._size.Y += by * 2;

        return g;
    }

    /// <summary>
    /// Returns a copy of the <see cref="Rect2D"/> grown by the specified amount
    /// on each side individually.
    /// </summary>
    /// <seealso cref="Grow(double)"/>
    /// <seealso cref="GrowSide(Side, double)"/>
    /// <param name="left">The amount to grow by on the left side.</param>
    /// <param name="top">The amount to grow by on the top side.</param>
    /// <param name="right">The amount to grow by on the right side.</param>
    /// <param name="bottom">The amount to grow by on the bottom side.</param>
    /// <returns>The grown <see cref="Rect2D"/>.</returns>
    public readonly Rect2D GrowIndividual(double left, double top, double right, double bottom)
    {
        var g = this;

        g._position.X -= left;
        g._position.Y -= top;
        g._size.X += left + right;
        g._size.Y += top + bottom;

        return g;
    }

    /// <summary>
    /// Returns a copy of the <see cref="Rect2D"/> grown by the specified amount
    /// on the specified <see cref="Side"/>.
    /// </summary>
    /// <seealso cref="Grow(double)"/>
    /// <seealso cref="GrowIndividual(double, double, double, double)"/>
    /// <param name="side">The side to grow.</param>
    /// <param name="by">The amount to grow by.</param>
    /// <returns>The grown <see cref="Rect2D"/>.</returns>
    public readonly Rect2D GrowSide(Side side, double by)
    {
        var g = this;

        g = g.GrowIndividual(Side.Left == side ? by : 0,
            Side.Top == side ? by : 0,
            Side.Right == side ? by : 0,
            Side.Bottom == side ? by : 0);

        return g;
    }

    /// <summary>
    /// Returns <see langword="true"/> if the <see cref="Rect2D"/> has
    /// area, and <see langword="false"/> if the <see cref="Rect2D"/>
    /// is linear, empty, or has a negative <see cref="Size"/>.
    /// See also <see cref="Area"/>.
    /// </summary>
    /// <returns>
    /// A <see langword="bool"/> for whether or not the <see cref="Rect2D"/> has area.
    /// </returns>
    public readonly bool HasArea()
    {
        return _size.X > 0.0f && _size.Y > 0.0f;
    }

    /// <summary>
    /// Returns <see langword="true"/> if the <see cref="Rect2D"/> contains a point,
    /// or <see langword="false"/> otherwise.
    /// </summary>
    /// <param name="point">The point to check.</param>
    /// <returns>
    /// A <see langword="bool"/> for whether or not the <see cref="Rect2D"/> contains <paramref name="point"/>.
    /// </returns>
    public readonly bool HasPoint(Vector2D point)
    {
        if (point.X < _position.X)
            return false;
        if (point.Y < _position.Y)
            return false;

        if (point.X >= _position.X + _size.X)
            return false;
        if (point.Y >= _position.Y + _size.Y)
            return false;

        return true;
    }

    /// <summary>
    /// Returns <see langword="true"/> if the <see cref="Rect2D"/> overlaps with <paramref name="b"/>
    /// (i.e. they have at least one point in common).
    ///
    /// If <paramref name="includeBorders"/> is <see langword="true"/>,
    /// they will also be considered overlapping if their borders touch,
    /// even without intersection.
    /// </summary>
    /// <param name="b">The other <see cref="Rect2D"/> to check for intersections with.</param>
    /// <param name="includeBorders">Whether or not to consider borders.</param>
    /// <returns>A <see langword="bool"/> for whether or not they are intersecting.</returns>
    public readonly bool Intersects(Rect2D b, bool includeBorders = false)
    {
        if (includeBorders)
        {
            if (_position.X > b._position.X + b._size.X) return false;
            if (_position.X + _size.X < b._position.X) return false;
            if (_position.Y > b._position.Y + b._size.Y) return false;
            if (_position.Y + _size.Y < b._position.Y) return false;
        }
        else
        {
            if (_position.X >= b._position.X + b._size.X) return false;
            if (_position.X + _size.X <= b._position.X) return false;
            if (_position.Y >= b._position.Y + b._size.Y) return false;
            if (_position.Y + _size.Y <= b._position.Y) return false;
        }

        return true;
    }

    /// <summary>
    /// Returns a larger <see cref="Rect2D"/> that contains this <see cref="Rect2D"/> and <paramref name="b"/>.
    /// </summary>
    /// <param name="b">The other <see cref="Rect2D"/>.</param>
    /// <returns>The merged <see cref="Rect2D"/>.</returns>
    public readonly Rect2D Merge(Rect2D b)
    {
        Rect2D newRect;

        newRect._position.X = Maths.Min(b._position.X, _position.X);
        newRect._position.Y = Maths.Min(b._position.Y, _position.Y);

        newRect._size.X = Maths.Max(b._position.X + b._size.X, _position.X + _size.X);
        newRect._size.Y = Maths.Max(b._position.Y + b._size.Y, _position.Y + _size.Y);

        newRect._size -= newRect._position; // Make relative again

        return newRect;
    }

    /// <summary>
    /// Constructs a <see cref="Rect2D"/> from a position and size.
    /// </summary>
    /// <param name="position">The position.</param>
    /// <param name="size">The size.</param>
    public Rect2D(Vector2D position, Vector2D size)
    {
        _position = position;
        _size = size;
    }

    /// <summary>
    /// Constructs a <see cref="Rect2D"/> from a position, width, and height.
    /// </summary>
    /// <param name="position">The position.</param>
    /// <param name="width">The width.</param>
    /// <param name="height">The height.</param>
    public Rect2D(Vector2D position, double width, double height)
    {
        _position = position;
        _size = new Vector2D(width, height);
    }

    /// <summary>
    /// Constructs a <see cref="Rect2D"/> from x, y, and size.
    /// </summary>
    /// <param name="x">The position's X coordinate.</param>
    /// <param name="y">The position's Y coordinate.</param>
    /// <param name="size">The size.</param>
    public Rect2D(double x, double y, Vector2D size)
    {
        _position = new Vector2D(x, y);
        _size = size;
    }

    /// <summary>
    /// Constructs a <see cref="Rect2D"/> from x, y, width, and height.
    /// </summary>
    /// <param name="x">The position's X coordinate.</param>
    /// <param name="y">The position's Y coordinate.</param>
    /// <param name="width">The width.</param>
    /// <param name="height">The height.</param>
    public Rect2D(double x, double y, double width, double height)
    {
        _position = new Vector2D(x, y);
        _size = new Vector2D(width, height);
    }

    /// <summary>
    /// Returns <see langword="true"/> if the
    /// <see cref="Rect2D"/>s are exactly equal.
    /// Note: Due to doubleing-point precision errors, consider using
    /// <see cref="IsEqualApprox"/> instead, which is more reliable.
    /// </summary>
    /// <param name="left">The left rect.</param>
    /// <param name="right">The right rect.</param>
    /// <returns>Whether or not the rects are exactly equal.</returns>
    public static bool operator ==(Rect2D left, Rect2D right)
    {
        return left.Equals(right);
    }

    /// <summary>
    /// Returns <see langword="true"/> if the
    /// <see cref="Rect2D"/>s are not equal.
    /// Note: Due to doubleing-point precision errors, consider using
    /// <see cref="IsEqualApprox"/> instead, which is more reliable.
    /// </summary>
    /// <param name="left">The left rect.</param>
    /// <param name="right">The right rect.</param>
    /// <returns>Whether or not the rects are not equal.</returns>
    public static bool operator !=(Rect2D left, Rect2D right)
    {
        return !left.Equals(right);
    }

    /// <summary>
    /// Returns <see langword="true"/> if this rect and <paramref name="obj"/> are equal.
    /// </summary>
    /// <param name="obj">The other object to compare.</param>
    /// <returns>Whether or not the rect and the other object are exactly equal.</returns>
    public readonly override bool Equals(object obj)
    {
        return obj is Rect2D other && Equals(other);
    }

    /// <summary>
    /// Returns <see langword="true"/> if this rect and <paramref name="other"/> are equal.
    /// </summary>
    /// <param name="other">The other rect to compare.</param>
    /// <returns>Whether or not the rects are exactly equal.</returns>
    public readonly bool Equals(Rect2D other)
    {
        return _position.Equals(other._position) && _size.Equals(other._size);
    }

    /// <summary>
    /// Returns <see langword="true"/> if this rect and <paramref name="other"/> are approximately equal,
    /// by running <see cref="Vector2D.IsEqualApprox(Vector2D)"/> on each component.
    /// </summary>
    /// <param name="other">The other rect to compare.</param>
    /// <returns>Whether or not the rects are approximately equal.</returns>
    public readonly bool IsEqualApprox(Rect2D other)
    {
        return _position.IsEqualApprox(other._position) && _size.IsEqualApprox(other.Size);
    }

    /// <summary>
    /// Serves as the hash function for <see cref="Rect2D"/>.
    /// </summary>
    /// <returns>A hash code for this rect.</returns>
    public readonly override int GetHashCode()
    {
        return _position.GetHashCode() ^ _size.GetHashCode();
    }


    public override string ToString()
    {
        return JsonConvert.SerializeObject(this);
    }

    public static Rect2D Parse(string text)
    {
        return JsonConvert.DeserializeObject<Rect2D>(text);
    }
}