﻿#define KB_RECT2F
using KludgeBox.Enums;
using Newtonsoft.Json;
using System;

namespace KludgeBox.Structs;

/// <summary>
/// 2D axis-aligned bounding box. Rect2 consists of a position, a size, and
/// several utility functions. It is typically used for fast overlap tests.
/// </summary>
[Serializable]
public partial struct Rect2F : IEquatable<Rect2F>
{
    private Vector2F _position;
    private Vector2F _size;

    /// <summary>
    /// Beginning corner. Typically has values lower than <see cref="End"/>.
    /// </summary>
    /// <value>Directly uses a private field.</value>
    public Vector2F Position
    {
        readonly get => _position;
        set => _position = value;
    }

    /// <summary>
    /// Size from <see cref="Position"/> to <see cref="End"/>. Typically all components are positive.
    /// If the size is negative, you can use <see cref="Abs"/> to fix it.
    /// </summary>
    /// <value>Directly uses a private field.</value>
    public Vector2F Size
    {
        readonly get => _size;
        set => _size = value;
    }

    /// <summary>
    /// Ending corner. This is calculated as <see cref="Position"/> plus <see cref="Size"/>.
    /// Setting this value will change the size.
    /// </summary>
    /// <value>
    /// Getting is equivalent to <paramref name="value"/> = <see cref="Position"/> + <see cref="Size"/>,
    /// setting is equivalent to <see cref="Size"/> = <paramref name="value"/> - <see cref="Position"/>
    /// </value>
    [JsonIgnore]
    public Vector2F End
    {
        readonly get => _position + _size;
        set => _size = value - _position;
    }

    /// <summary>
    /// The area of this <see cref="Rect2F"/>.
    /// See also <see cref="HasArea"/>.
    /// </summary>
    [JsonIgnore]
    public readonly float Area => _size.X * _size.Y;

    /// <summary>
    /// Returns a <see cref="Rect2F"/> with equivalent position and size, modified so that
    /// the top-left corner is the origin and width and height are positive.
    /// </summary>
    /// <returns>The modified <see cref="Rect2F"/>.</returns>
    public readonly Rect2F Abs()
    {
        var end = End;
        var topLeft = new Vector2F(Maths.Min(_position.X, end.X), Maths.Min(_position.Y, end.Y));
        return new Rect2F(topLeft, _size.Abs());
    }

    /// <summary>
    /// Returns the intersection of this <see cref="Rect2F"/> and <paramref name="b"/>.
    /// If the rectangles do not intersect, an empty <see cref="Rect2F"/> is returned.
    /// </summary>
    /// <param name="b">The other <see cref="Rect2F"/>.</param>
    /// <returns>
    /// The intersection of this <see cref="Rect2F"/> and <paramref name="b"/>,
    /// or an empty <see cref="Rect2F"/> if they do not intersect.
    /// </returns>
    public readonly Rect2F Intersection(Rect2F b)
    {
        var newRect = b;

        if (!Intersects(newRect)) return new Rect2F();

        newRect._position.X = Maths.Max(b._position.X, _position.X);
        newRect._position.Y = Maths.Max(b._position.Y, _position.Y);

        var bEnd = b._position + b._size;
        var end = _position + _size;

        newRect._size.X = Maths.Min(bEnd.X, end.X) - newRect._position.X;
        newRect._size.Y = Maths.Min(bEnd.Y, end.Y) - newRect._position.Y;

        return newRect;
    }

    /// <summary>
    /// Returns <see langword="true"/> if this <see cref="Rect2F"/> is finite, by calling
    /// <see cref="Maths.IsFinite(float)"/> on each component.
    /// </summary>
    /// <returns>Whether this vector is finite or not.</returns>
    public bool IsFinite()
    {
        return _position.IsFinite() && _size.IsFinite();
    }

    /// <summary>
    /// Returns <see langword="true"/> if this <see cref="Rect2F"/> completely encloses another one.
    /// </summary>
    /// <param name="b">The other <see cref="Rect2F"/> that may be enclosed.</param>
    /// <returns>
    /// A <see langword="bool"/> for whether or not this <see cref="Rect2F"/> encloses <paramref name="b"/>.
    /// </returns>
    public readonly bool Encloses(Rect2F b)
    {
        return b._position.X >= _position.X && b._position.Y >= _position.Y &&
               b._position.X + b._size.X < _position.X + _size.X &&
               b._position.Y + b._size.Y < _position.Y + _size.Y;
    }

    /// <summary>
    /// Returns this <see cref="Rect2F"/> expanded to include a given point.
    /// </summary>
    /// <param name="to">The point to include.</param>
    /// <returns>The expanded <see cref="Rect2F"/>.</returns>
    public readonly Rect2F Expand(Vector2F to)
    {
        var expanded = this;

        var begin = expanded._position;
        var end = expanded._position + expanded._size;

        if (to.X < begin.X) begin.X = to.X;
        if (to.Y < begin.Y) begin.Y = to.Y;

        if (to.X > end.X) end.X = to.X;
        if (to.Y > end.Y) end.Y = to.Y;

        expanded._position = begin;
        expanded._size = end - begin;

        return expanded;
    }

    /// <summary>
    /// Returns the center of the <see cref="Rect2F"/>, which is equal
    /// to <see cref="Position"/> + (<see cref="Size"/> / 2).
    /// </summary>
    /// <returns>The center.</returns>
    public readonly Vector2F GetCenter()
    {
        return _position + _size * 0.5f;
    }

    /// <summary>
    /// Returns a copy of the <see cref="Rect2F"/> grown by the specified amount
    /// on all sides.
    /// </summary>
    /// <seealso cref="GrowIndividual(float, float, float, float)"/>
    /// <seealso cref="GrowSide(Side, float)"/>
    /// <param name="by">The amount to grow by.</param>
    /// <returns>The grown <see cref="Rect2F"/>.</returns>
    public readonly Rect2F Grow(float by)
    {
        var g = this;

        g._position.X -= by;
        g._position.Y -= by;
        g._size.X += by * 2;
        g._size.Y += by * 2;

        return g;
    }

    /// <summary>
    /// Returns a copy of the <see cref="Rect2F"/> grown by the specified amount
    /// on each side individually.
    /// </summary>
    /// <seealso cref="Grow(float)"/>
    /// <seealso cref="GrowSide(Side, float)"/>
    /// <param name="left">The amount to grow by on the left side.</param>
    /// <param name="top">The amount to grow by on the top side.</param>
    /// <param name="right">The amount to grow by on the right side.</param>
    /// <param name="bottom">The amount to grow by on the bottom side.</param>
    /// <returns>The grown <see cref="Rect2F"/>.</returns>
    public readonly Rect2F GrowIndividual(float left, float top, float right, float bottom)
    {
        var g = this;

        g._position.X -= left;
        g._position.Y -= top;
        g._size.X += left + right;
        g._size.Y += top + bottom;

        return g;
    }

    /// <summary>
    /// Returns a copy of the <see cref="Rect2F"/> grown by the specified amount
    /// on the specified <see cref="Side"/>.
    /// </summary>
    /// <seealso cref="Grow(float)"/>
    /// <seealso cref="GrowIndividual(float, float, float, float)"/>
    /// <param name="side">The side to grow.</param>
    /// <param name="by">The amount to grow by.</param>
    /// <returns>The grown <see cref="Rect2F"/>.</returns>
    public readonly Rect2F GrowSide(Side side, float by)
    {
        var g = this;

        g = g.GrowIndividual(Side.Left == side ? by : 0,
            Side.Top == side ? by : 0,
            Side.Right == side ? by : 0,
            Side.Bottom == side ? by : 0);

        return g;
    }

    /// <summary>
    /// Returns <see langword="true"/> if the <see cref="Rect2F"/> has
    /// area, and <see langword="false"/> if the <see cref="Rect2F"/>
    /// is linear, empty, or has a negative <see cref="Size"/>.
    /// See also <see cref="Area"/>.
    /// </summary>
    /// <returns>
    /// A <see langword="bool"/> for whether or not the <see cref="Rect2F"/> has area.
    /// </returns>
    public readonly bool HasArea()
    {
        return _size.X > 0.0f && _size.Y > 0.0f;
    }

    /// <summary>
    /// Returns <see langword="true"/> if the <see cref="Rect2F"/> contains a point,
    /// or <see langword="false"/> otherwise.
    /// </summary>
    /// <param name="point">The point to check.</param>
    /// <returns>
    /// A <see langword="bool"/> for whether or not the <see cref="Rect2F"/> contains <paramref name="point"/>.
    /// </returns>
    public readonly bool HasPoint(Vector2F point)
    {
        if (point.X < _position.X)
            return false;
        if (point.Y < _position.Y)
            return false;

        if (point.X >= _position.X + _size.X)
            return false;
        if (point.Y >= _position.Y + _size.Y)
            return false;

        return true;
    }

    /// <summary>
    /// Returns <see langword="true"/> if the <see cref="Rect2F"/> overlaps with <paramref name="b"/>
    /// (i.e. they have at least one point in common).
    ///
    /// If <paramref name="includeBorders"/> is <see langword="true"/>,
    /// they will also be considered overlapping if their borders touch,
    /// even without intersection.
    /// </summary>
    /// <param name="b">The other <see cref="Rect2F"/> to check for intersections with.</param>
    /// <param name="includeBorders">Whether or not to consider borders.</param>
    /// <returns>A <see langword="bool"/> for whether or not they are intersecting.</returns>
    public readonly bool Intersects(Rect2F b, bool includeBorders = false)
    {
        if (includeBorders)
        {
            if (_position.X > b._position.X + b._size.X) return false;
            if (_position.X + _size.X < b._position.X) return false;
            if (_position.Y > b._position.Y + b._size.Y) return false;
            if (_position.Y + _size.Y < b._position.Y) return false;
        }
        else
        {
            if (_position.X >= b._position.X + b._size.X) return false;
            if (_position.X + _size.X <= b._position.X) return false;
            if (_position.Y >= b._position.Y + b._size.Y) return false;
            if (_position.Y + _size.Y <= b._position.Y) return false;
        }

        return true;
    }

    /// <summary>
    /// Returns a larger <see cref="Rect2F"/> that contains this <see cref="Rect2F"/> and <paramref name="b"/>.
    /// </summary>
    /// <param name="b">The other <see cref="Rect2F"/>.</param>
    /// <returns>The merged <see cref="Rect2F"/>.</returns>
    public readonly Rect2F Merge(Rect2F b)
    {
        Rect2F newRect;

        newRect._position.X = Maths.Min(b._position.X, _position.X);
        newRect._position.Y = Maths.Min(b._position.Y, _position.Y);

        newRect._size.X = Maths.Max(b._position.X + b._size.X, _position.X + _size.X);
        newRect._size.Y = Maths.Max(b._position.Y + b._size.Y, _position.Y + _size.Y);

        newRect._size -= newRect._position; // Make relative again

        return newRect;
    }

    /// <summary>
    /// Constructs a <see cref="Rect2F"/> from a position and size.
    /// </summary>
    /// <param name="position">The position.</param>
    /// <param name="size">The size.</param>
    public Rect2F(Vector2F position, Vector2F size)
    {
        _position = position;
        _size = size;
    }

    /// <summary>
    /// Constructs a <see cref="Rect2F"/> from a position, width, and height.
    /// </summary>
    /// <param name="position">The position.</param>
    /// <param name="width">The width.</param>
    /// <param name="height">The height.</param>
    public Rect2F(Vector2F position, float width, float height)
    {
        _position = position;
        _size = new Vector2F(width, height);
    }

    /// <summary>
    /// Constructs a <see cref="Rect2F"/> from x, y, and size.
    /// </summary>
    /// <param name="x">The position's X coordinate.</param>
    /// <param name="y">The position's Y coordinate.</param>
    /// <param name="size">The size.</param>
    public Rect2F(float x, float y, Vector2F size)
    {
        _position = new Vector2F(x, y);
        _size = size;
    }

    /// <summary>
    /// Constructs a <see cref="Rect2F"/> from x, y, width, and height.
    /// </summary>
    /// <param name="x">The position's X coordinate.</param>
    /// <param name="y">The position's Y coordinate.</param>
    /// <param name="width">The width.</param>
    /// <param name="height">The height.</param>
    public Rect2F(float x, float y, float width, float height)
    {
        _position = new Vector2F(x, y);
        _size = new Vector2F(width, height);
    }

    /// <summary>
    /// Returns <see langword="true"/> if the
    /// <see cref="Rect2F"/>s are exactly equal.
    /// Note: Due to floating-point precision errors, consider using
    /// <see cref="IsEqualApprox"/> instead, which is more reliable.
    /// </summary>
    /// <param name="left">The left rect.</param>
    /// <param name="right">The right rect.</param>
    /// <returns>Whether or not the rects are exactly equal.</returns>
    public static bool operator ==(Rect2F left, Rect2F right)
    {
        return left.Equals(right);
    }

    /// <summary>
    /// Returns <see langword="true"/> if the
    /// <see cref="Rect2F"/>s are not equal.
    /// Note: Due to floating-point precision errors, consider using
    /// <see cref="IsEqualApprox"/> instead, which is more reliable.
    /// </summary>
    /// <param name="left">The left rect.</param>
    /// <param name="right">The right rect.</param>
    /// <returns>Whether or not the rects are not equal.</returns>
    public static bool operator !=(Rect2F left, Rect2F right)
    {
        return !left.Equals(right);
    }

    /// <summary>
    /// Returns <see langword="true"/> if this rect and <paramref name="obj"/> are equal.
    /// </summary>
    /// <param name="obj">The other object to compare.</param>
    /// <returns>Whether or not the rect and the other object are exactly equal.</returns>
    public readonly override bool Equals(object obj)
    {
        return obj is Rect2F other && Equals(other);
    }

    /// <summary>
    /// Returns <see langword="true"/> if this rect and <paramref name="other"/> are equal.
    /// </summary>
    /// <param name="other">The other rect to compare.</param>
    /// <returns>Whether or not the rects are exactly equal.</returns>
    public readonly bool Equals(Rect2F other)
    {
        return _position.Equals(other._position) && _size.Equals(other._size);
    }

    /// <summary>
    /// Returns <see langword="true"/> if this rect and <paramref name="other"/> are approximately equal,
    /// by running <see cref="Vector2F.IsEqualApprox(Vector2F)"/> on each component.
    /// </summary>
    /// <param name="other">The other rect to compare.</param>
    /// <returns>Whether or not the rects are approximately equal.</returns>
    public readonly bool IsEqualApprox(Rect2F other)
    {
        return _position.IsEqualApprox(other._position) && _size.IsEqualApprox(other.Size);
    }

    /// <summary>
    /// Serves as the hash function for <see cref="Rect2F"/>.
    /// </summary>
    /// <returns>A hash code for this rect.</returns>
    public readonly override int GetHashCode()
    {
        return _position.GetHashCode() ^ _size.GetHashCode();
    }


    public override string ToString()
    {
        return JsonConvert.SerializeObject(this);
    }

    public static Rect2F Parse(string text)
    {
        return JsonConvert.DeserializeObject<Rect2F>(text);
    }
}