﻿#define KB_VEC2F
using System;
using System.Globalization;
using System.Text.RegularExpressions;

namespace KludgeBox.Structs;

[Serializable]
public partial struct Vector2F : IEquatable<Vector2F>
{
    /// <summary>
    /// Enumerated index values for the axes.
    /// Returned by <see cref="MaxAxisIndex"/> and <see cref="MinAxisIndex"/>.
    /// </summary>
    public enum Axis
    {
        /// <summary>
        /// The vector's X axis.
        /// </summary>
        X = 0,

        /// <summary>
        /// The vector's Y axis.
        /// </summary>
        Y
    }

    /// <summary>
    /// The vector's X component. Also accessible by using the index position <c>[0]</c>.
    /// </summary>
    public float X;

    /// <summary>
    /// The vector's Y component. Also accessible by using the index position <c>[1]</c>.
    /// </summary>
    public float Y;

    /// <summary>
    /// Access vector components using their index.
    /// </summary>
    /// <exception cref="ArgumentOutOfRangeException">
    /// <paramref name="index"/> is not 0 or 1.
    /// </exception>
    /// <value>
    /// <c>[0]</c> is equivalent to <see cref="X"/>,
    /// <c>[1]</c> is equivalent to <see cref="Y"/>.
    /// </value>
    public float this[int index]
    {
        readonly get
        {
            switch (index)
            {
                case 0:
                    return X;
                case 1:
                    return Y;
                default:
                    throw new ArgumentOutOfRangeException(nameof(index));
            }
        }
        set
        {
            switch (index)
            {
                case 0:
                    X = value;
                    return;
                case 1:
                    Y = value;
                    return;
                default:
                    throw new ArgumentOutOfRangeException(nameof(index));
            }
        }
    }


    /// <summary>
    /// Helper method for deconstruction into a tuple.
    /// </summary>
    public readonly void Deconstruct(out float x, out float y)
    {
        x = X;
        y = Y;
    }

    internal void Normalize()
    {
        var lengthsq = LengthSquared();

        if (lengthsq == 0)
        {
            X = Y = 0f;
        }
        else
        {
            float length = Maths.Sqrt(lengthsq);
            X /= length;
            Y /= length;
        }
    }

    /// <summary>
    /// Returns a new vector with all components in absolute values (i.e. positive).
    /// </summary>
    /// <returns>A vector with <see cref="Maths.Abs(float)"/> called on each component.</returns>
    public readonly Vector2F Abs()
    {
        return new Vector2F(Maths.Abs(X), Maths.Abs(Y));
    }

    /// <summary>
    /// Returns this vector's angle with respect to the X axis, or (1, 0) vector, in radians.
    ///
    /// Equivalent to the result of <see cref="Maths.Atan2(float, float)"/> when
    /// called with the vector's <see cref="Y"/> and <see cref="X"/> as parameters: <c>Maths.Atan2(v.Y, v.X)</c>.
    /// </summary>
    /// <returns>The angle of this vector, in radians.</returns>
    public readonly float Angle()
    {
        return Maths.Atan2(Y, X);
    }

    /// <summary>
    /// Returns the angle to the given vector, in radians.
    /// </summary>
    /// <param name="to">The other vector to compare this vector to.</param>
    /// <returns>The angle between the two vectors, in radians.</returns>
    public readonly float AngleTo(Vector2F to)
    {
        return Maths.Atan2(Cross(to), Dot(to));
    }

    /// <summary>
    /// Returns the angle between the line connecting the two points and the X axis, in radians.
    /// </summary>
    /// <param name="to">The other vector to compare this vector to.</param>
    /// <returns>The angle between the two vectors, in radians.</returns>
    public readonly float AngleToPoint(Vector2F to)
    {
        return Maths.Atan2(to.Y - Y, to.X - X);
    }

    /// <summary>
    /// Returns the aspect ratio of this vector, the ratio of <see cref="X"/> to <see cref="Y"/>.
    /// </summary>
    /// <returns>The <see cref="X"/> component divided by the <see cref="Y"/> component.</returns>
    public readonly float Aspect()
    {
        return X / Y;
    }

    /// <summary>
    /// Returns the vector "bounced off" from a plane defined by the given normal.
    /// </summary>
    /// <param name="normal">The normal vector defining the plane to bounce off. Must be normalized.</param>
    /// <returns>The bounced vector.</returns>
    public readonly Vector2F Bounce(Vector2F normal)
    {
        return -Reflect(normal);
    }

    /// <summary>
    /// Returns a new vector with all components rounded up (towards positive infinity).
    /// </summary>
    /// <returns>A vector with <see cref="Maths.Ceil(float)"/> called on each component.</returns>
    public readonly Vector2F Ceil()
    {
        return new Vector2F(Maths.Ceil(X), Maths.Ceil(Y));
    }

    /// <summary>
    /// Returns a new vector with all components clamped between the
    /// components of <paramref name="min"/> and <paramref name="max"/> using
    /// <see cref="Maths.Clamp(float, float, float)"/>.
    /// </summary>
    /// <param name="min">The vector with minimum allowed values.</param>
    /// <param name="max">The vector with maximum allowed values.</param>
    /// <returns>The vector with all components clamped.</returns>
    public readonly Vector2F Clamp(Vector2F min, Vector2F max)
    {
        return new Vector2F
        (
            Maths.Clamp(X, min.X, max.X),
            Maths.Clamp(Y, min.Y, max.Y)
        );
    }

    /// <summary>
    /// Returns the cross product of this vector and <paramref name="with"/>.
    /// </summary>
    /// <param name="with">The other vector.</param>
    /// <returns>The cross product value.</returns>
    public readonly float Cross(Vector2F with)
    {
        return X * with.Y - Y * with.X;
    }

    /// <summary>
    /// Performs a cubic interpolation between vectors <paramref name="preA"/>, this vector,
    /// <paramref name="b"/>, and <paramref name="postB"/>, by the given amount <paramref name="weight"/>.
    /// </summary>
    /// <param name="b">The destination vector.</param>
    /// <param name="preA">A vector before this vector.</param>
    /// <param name="postB">A vector after <paramref name="b"/>.</param>
    /// <param name="weight">A value on the range of 0.0 to 1.0, representing the amount of interpolation.</param>
    /// <returns>The interpolated vector.</returns>
    public readonly Vector2F CubicInterpolate(Vector2F b, Vector2F preA, Vector2F postB, float weight)
    {
        return new Vector2F
        (
            Maths.CubicInterpolate(X, b.X, preA.X, postB.X, weight),
            Maths.CubicInterpolate(Y, b.Y, preA.Y, postB.Y, weight)
        );
    }

    /// <summary>
    /// Performs a cubic interpolation between vectors <paramref name="preA"/>, this vector,
    /// <paramref name="b"/>, and <paramref name="postB"/>, by the given amount <paramref name="weight"/>.
    /// It can perform smoother interpolation than <see cref="CubicInterpolate"/>
    /// by the time values.
    /// </summary>
    /// <param name="b">The destination vector.</param>
    /// <param name="preA">A vector before this vector.</param>
    /// <param name="postB">A vector after <paramref name="b"/>.</param>
    /// <param name="weight">A value on the range of 0.0 to 1.0, representing the amount of interpolation.</param>
    /// <param name="t"></param>
    /// <param name="preAt"></param>
    /// <param name="postBt"></param>
    /// <returns>The interpolated vector.</returns>
    public readonly Vector2F CubicInterpolateInTime(Vector2F b, Vector2F preA, Vector2F postB, float weight, float t,
        float preAt, float postBt)
    {
        return new Vector2F
        (
            Maths.CubicInterpolateInTime(X, b.X, preA.X, postB.X, weight, t, preAt, postBt),
            Maths.CubicInterpolateInTime(Y, b.Y, preA.Y, postB.Y, weight, t, preAt, postBt)
        );
    }

    /// <summary>
    /// Returns the point at the given <paramref name="t"/> on a one-dimensional Bezier curve defined by this vector
    /// and the given <paramref name="control1"/>, <paramref name="control2"/>, and <paramref name="end"/> points.
    /// </summary>
    /// <param name="control1">Control point that defines the bezier curve.</param>
    /// <param name="control2">Control point that defines the bezier curve.</param>
    /// <param name="end">The destination vector.</param>
    /// <param name="t">A value on the range of 0.0 to 1.0, representing the amount of interpolation.</param>
    /// <returns>The interpolated vector.</returns>
    public readonly Vector2F BezierInterpolate(Vector2F control1, Vector2F control2, Vector2F end, float t)
    {
        return new Vector2F
        (
            Maths.BezierInterpolate(X, control1.X, control2.X, end.X, t),
            Maths.BezierInterpolate(Y, control1.Y, control2.Y, end.Y, t)
        );
    }

    /// <summary>
    /// Returns the derivative at the given <paramref name="t"/> on the Bezier curve defined by this vector
    /// and the given <paramref name="control1"/>, <paramref name="control2"/>, and <paramref name="end"/> points.
    /// </summary>
    /// <param name="control1">Control point that defines the bezier curve.</param>
    /// <param name="control2">Control point that defines the bezier curve.</param>
    /// <param name="end">The destination value for the interpolation.</param>
    /// <param name="t">A value on the range of 0.0 to 1.0, representing the amount of interpolation.</param>
    /// <returns>The resulting value of the interpolation.</returns>
    public readonly Vector2F BezierDerivative(Vector2F control1, Vector2F control2, Vector2F end, float t)
    {
        return new Vector2F(
            Maths.BezierDerivative(X, control1.X, control2.X, end.X, t),
            Maths.BezierDerivative(Y, control1.Y, control2.Y, end.Y, t)
        );
    }

    /// <summary>
    /// Returns the normalized vector pointing from this vector to <paramref name="to"/>.
    /// </summary>
    /// <param name="to">The other vector to point towards.</param>
    /// <returns>The direction from this vector to <paramref name="to"/>.</returns>
    public readonly Vector2F DirectionTo(Vector2F to)
    {
        return new Vector2F(to.X - X, to.Y - Y).Normalized();
    }

    /// <summary>
    /// Returns the squared distance between this vector and <paramref name="to"/>.
    /// This method runs faster than <see cref="DistanceTo"/>, so prefer it if
    /// you need to compare vectors or need the squared distance for some formula.
    /// </summary>
    /// <param name="to">The other vector to use.</param>
    /// <returns>The squared distance between the two vectors.</returns>
    public readonly float DistanceSquaredTo(Vector2F to)
    {
        return (X - to.X) * (X - to.X) + (Y - to.Y) * (Y - to.Y);
    }

    /// <summary>
    /// Returns the distance between this vector and <paramref name="to"/>.
    /// </summary>
    /// <param name="to">The other vector to use.</param>
    /// <returns>The distance between the two vectors.</returns>
    public readonly float DistanceTo(Vector2F to)
    {
        return Maths.Sqrt((X - to.X) * (X - to.X) + (Y - to.Y) * (Y - to.Y));
    }

    /// <summary>
    /// Returns the dot product of this vector and <paramref name="with"/>.
    /// </summary>
    /// <param name="with">The other vector to use.</param>
    /// <returns>The dot product of the two vectors.</returns>
    public readonly float Dot(Vector2F with)
    {
        return X * with.X + Y * with.Y;
    }

    /// <summary>
    /// Returns a new vector with all components rounded down (towards negative infinity).
    /// </summary>
    /// <returns>A vector with <see cref="Maths.Floor(float)"/> called on each component.</returns>
    public readonly Vector2F Floor()
    {
        return new Vector2F(Maths.Floor(X), Maths.Floor(Y));
    }

    /// <summary>
    /// Returns the inverse of this vector. This is the same as <c>new Vector2(1 / v.X, 1 / v.Y)</c>.
    /// </summary>
    /// <returns>The inverse of this vector.</returns>
    public readonly Vector2F Inverse()
    {
        return new Vector2F(1 / X, 1 / Y);
    }

    /// <summary>
    /// Returns <see langword="true"/> if this vector is finite, by calling
    /// <see cref="Maths.IsFinite(float)"/> on each component.
    /// </summary>
    /// <returns>Whether this vector is finite or not.</returns>
    public readonly bool IsFinite()
    {
        return Maths.IsFinite(X) && Maths.IsFinite(Y);
    }

    /// <summary>
    /// Returns <see langword="true"/> if the vector is normalized, and <see langword="false"/> otherwise.
    /// </summary>
    /// <returns>A <see langword="bool"/> indicating whether or not the vector is normalized.</returns>
    public readonly bool IsNormalized()
    {
        return Maths.Abs(LengthSquared() - 1.0f) < Maths.Epsilon;
    }

    /// <summary>
    /// Returns the length (magnitude) of this vector.
    /// </summary>
    /// <seealso cref="LengthSquared"/>
    /// <returns>The length of this vector.</returns>
    public readonly float Length()
    {
        return Maths.Sqrt(X * X + Y * Y);
    }

    /// <summary>
    /// Returns the squared length (squared magnitude) of this vector.
    /// This method runs faster than <see cref="Length"/>, so prefer it if
    /// you need to compare vectors or need the squared length for some formula.
    /// </summary>
    /// <returns>The squared length of this vector.</returns>
    public readonly float LengthSquared()
    {
        return X * X + Y * Y;
    }

    /// <summary>
    /// Returns the result of the linear interpolation between
    /// this vector and <paramref name="to"/> by amount <paramref name="weight"/>.
    /// </summary>
    /// <param name="to">The destination vector for interpolation.</param>
    /// <param name="weight">A value on the range of 0.0 to 1.0, representing the amount of interpolation.</param>
    /// <returns>The resulting vector of the interpolation.</returns>
    public readonly Vector2F Lerp(Vector2F to, float weight)
    {
        return new Vector2F
        (
            Maths.Lerp(X, to.X, weight),
            Maths.Lerp(Y, to.Y, weight)
        );
    }

    /// <summary>
    /// Returns the vector with a maximum length by limiting its length to <paramref name="length"/>.
    /// </summary>
    /// <param name="length">The length to limit to.</param>
    /// <returns>The vector with its length limited.</returns>
    public readonly Vector2F LimitLength(float length = 1.0f)
    {
        var v = this;
        var l = Length();

        if (l > 0 && length < l)
        {
            v /= l;
            v *= length;
        }

        return v;
    }

    /// <summary>
    /// Returns the axis of the vector's highest value. See <see cref="Axis"/>.
    /// If both components are equal, this method returns <see cref="Axis.X"/>.
    /// </summary>
    /// <returns>The index of the highest axis.</returns>
    public readonly Axis MaxAxisIndex()
    {
        return X < Y ? Axis.Y : Axis.X;
    }

    /// <summary>
    /// Returns the axis of the vector's lowest value. See <see cref="Axis"/>.
    /// If both components are equal, this method returns <see cref="Axis.Y"/>.
    /// </summary>
    /// <returns>The index of the lowest axis.</returns>
    public readonly Axis MinAxisIndex()
    {
        return X < Y ? Axis.X : Axis.Y;
    }

    /// <summary>
    /// Moves this vector toward <paramref name="to"/> by the fixed <paramref name="delta"/> amount.
    /// </summary>
    /// <param name="to">The vector to move towards.</param>
    /// <param name="delta">The amount to move towards by.</param>
    /// <returns>The resulting vector.</returns>
    public readonly Vector2F MoveToward(Vector2F to, float delta)
    {
        var v = this;
        var vd = to - v;
        var len = vd.Length();
        if (len <= delta || len < Maths.Epsilon)
            return to;

        return v + vd / len * delta;
    }

    /// <summary>
    /// Returns the vector scaled to unit length. Equivalent to <c>v / v.Length()</c>.
    /// </summary>
    /// <returns>A normalized version of the vector.</returns>
    public readonly Vector2F Normalized()
    {
        var v = this;
        v.Normalize();
        return v;
    }

    /// <summary>
    /// Returns a vector composed of the <see cref="Maths.PosMod(float, float)"/> of this vector's components
    /// and <paramref name="mod"/>.
    /// </summary>
    /// <param name="mod">A value representing the divisor of the operation.</param>
    /// <returns>
    /// A vector with each component <see cref="Maths.PosMod(float, float)"/> by <paramref name="mod"/>.
    /// </returns>
    public readonly Vector2F PosMod(float mod)
    {
        Vector2F v;
        v.X = Maths.PosMod(X, mod);
        v.Y = Maths.PosMod(Y, mod);
        return v;
    }

    /// <summary>
    /// Returns a vector composed of the <see cref="Maths.PosMod(float, float)"/> of this vector's components
    /// and <paramref name="modv"/>'s components.
    /// </summary>
    /// <param name="modv">A vector representing the divisors of the operation.</param>
    /// <returns>
    /// A vector with each component <see cref="Maths.PosMod(float, float)"/> by <paramref name="modv"/>'s components.
    /// </returns>
    public readonly Vector2F PosMod(Vector2F modv)
    {
        Vector2F v;
        v.X = Maths.PosMod(X, modv.X);
        v.Y = Maths.PosMod(Y, modv.Y);
        return v;
    }

    /// <summary>
    /// Returns this vector projected onto another vector <paramref name="onNormal"/>.
    /// </summary>
    /// <param name="onNormal">The vector to project onto.</param>
    /// <returns>The projected vector.</returns>
    public readonly Vector2F Project(Vector2F onNormal)
    {
        return onNormal * (Dot(onNormal) / onNormal.LengthSquared());
    }

    /// <summary>
    /// Returns this vector reflected from a plane defined by the given <paramref name="normal"/>.
    /// </summary>
    /// <param name="normal">The normal vector defining the plane to reflect from. Must be normalized.</param>
    /// <returns>The reflected vector.</returns>
    public readonly Vector2F Reflect(Vector2F normal)
    {
#if DEBUG
        if (!normal.IsNormalized()) throw new ArgumentException("Argument is not normalized.", nameof(normal));
#endif
        return 2 * Dot(normal) * normal - this;
    }

    /// <summary>
    /// Rotates this vector by <paramref name="angle"/> radians.
    /// </summary>
    /// <param name="angle">The angle to rotate by, in radians.</param>
    /// <returns>The rotated vector.</returns>
    public readonly Vector2F Rotated(float angle)
    {
        (float sin, float cos) = Maths.SinCos(angle);
        return new Vector2F
        (
            X * cos - Y * sin,
            X * sin + Y * cos
        );
    }

    /// <summary>
    /// Returns this vector with all components rounded to the nearest integer,
    /// with halfway cases rounded towards the nearest multiple of two.
    /// </summary>
    /// <returns>The rounded vector.</returns>
    public readonly Vector2F Round()
    {
        return new Vector2F(Maths.Round(X), Maths.Round(Y));
    }

    /// <summary>
    /// Returns a vector with each component set to one or negative one, depending
    /// on the signs of this vector's components, or zero if the component is zero,
    /// by calling <see cref="Maths.Sign(float)"/> on each component.
    /// </summary>
    /// <returns>A vector with all components as either <c>1</c>, <c>-1</c>, or <c>0</c>.</returns>
    public readonly Vector2F Sign()
    {
        Vector2F v;
        v.X = Maths.Sign(X);
        v.Y = Maths.Sign(Y);
        return v;
    }

    /// <summary>
    /// Returns the result of the spherical linear interpolation between
    /// this vector and <paramref name="to"/> by amount <paramref name="weight"/>.
    ///
    /// This method also handles interpolating the lengths if the input vectors
    /// have different lengths. For the special case of one or both input vectors
    /// having zero length, this method behaves like <see cref="Lerp(Vector2F, float)"/>.
    /// </summary>
    /// <param name="to">The destination vector for interpolation.</param>
    /// <param name="weight">A value on the range of 0.0 to 1.0, representing the amount of interpolation.</param>
    /// <returns>The resulting vector of the interpolation.</returns>
    public readonly Vector2F Slerp(Vector2F to, float weight)
    {
        var startLengthSquared = LengthSquared();
        var endLengthSquared = to.LengthSquared();
        if (startLengthSquared == 0.0 || endLengthSquared == 0.0)
            // Zero length vectors have no angle, so the best we can do is either lerp or throw an error.
            return Lerp(to, weight);
        float startLength = Maths.Sqrt(startLengthSquared);
        float resultLength = Maths.Lerp(startLength, Maths.Sqrt(endLengthSquared), weight);
        var angle = AngleTo(to);
        return Rotated(angle * weight) * (resultLength / startLength);
    }

    /// <summary>
    /// Returns this vector slid along a plane defined by the given <paramref name="normal"/>.
    /// </summary>
    /// <param name="normal">The normal vector defining the plane to slide on.</param>
    /// <returns>The slid vector.</returns>
    public readonly Vector2F Slide(Vector2F normal)
    {
        return this - normal * Dot(normal);
    }

    /// <summary>
    /// Returns this vector with each component snapped to the nearest multiple of <paramref name="step"/>.
    /// This can also be used to round to an arbitrary number of decimals.
    /// </summary>
    /// <param name="step">A vector value representing the step size to snap to.</param>
    /// <returns>The snapped vector.</returns>
    public readonly Vector2F Snapped(Vector2F step)
    {
        return new Vector2F(Maths.Snapped(X, step.X), Maths.Snapped(Y, step.Y));
    }

    /// <summary>
    /// Returns a perpendicular vector rotated 90 degrees counter-clockwise
    /// compared to the original, with the same length.
    /// </summary>
    /// <returns>The perpendicular vector.</returns>
    public readonly Vector2F Orthogonal()
    {
        return new Vector2F(Y, -X);
    }

    // Constants
    private static readonly Vector2F _zero = new(0, 0);
    private static readonly Vector2F _one = new(1, 1);
    private static readonly Vector2F _inf = new(Maths.Inf, Maths.Inf);

    private static readonly Vector2F _up = new(0, -1);
    private static readonly Vector2F _down = new(0, 1);
    private static readonly Vector2F _right = new(1, 0);
    private static readonly Vector2F _left = new(-1, 0);

    /// <summary>
    /// Zero vector, a vector with all components set to <c>0</c>.
    /// </summary>
    /// <value>Equivalent to <c>new Vector2(0, 0)</c>.</value>
    public static Vector2F Zero => _zero;

    /// <summary>
    /// One vector, a vector with all components set to <c>1</c>.
    /// </summary>
    /// <value>Equivalent to <c>new Vector2(1, 1)</c>.</value>
    public static Vector2F One => _one;

    /// <summary>
    /// Infinity vector, a vector with all components set to <see cref="Maths.Inf"/>.
    /// </summary>
    /// <value>Equivalent to <c>new Vector2(Maths.Inf, Maths.Inf)</c>.</value>
    public static Vector2F Inf => _inf;

    /// <summary>
    /// Up unit vector. Y is down in 2D, so this vector points -Y.
    /// </summary>
    /// <value>Equivalent to <c>new Vector2(0, -1)</c>.</value>
    public static Vector2F Up => _up;

    /// <summary>
    /// Down unit vector. Y is down in 2D, so this vector points +Y.
    /// </summary>
    /// <value>Equivalent to <c>new Vector2(0, 1)</c>.</value>
    public static Vector2F Down => _down;

    /// <summary>
    /// Right unit vector. Represents the direction of right.
    /// </summary>
    /// <value>Equivalent to <c>new Vector2(1, 0)</c>.</value>
    public static Vector2F Right => _right;

    /// <summary>
    /// Left unit vector. Represents the direction of left.
    /// </summary>
    /// <value>Equivalent to <c>new Vector2(-1, 0)</c>.</value>
    public static Vector2F Left => _left;

    /// <summary>
    /// Constructs a new <see cref="Vector2F"/> with the given components.
    /// </summary>
    /// <param name="x">The vector's X component.</param>
    /// <param name="y">The vector's Y component.</param>
    public Vector2F(float x, float y)
    {
        X = x;
        Y = y;
    }

    /// <summary>
    /// Creates a unit Vector2 rotated to the given angle. This is equivalent to doing
    /// <c>Vector2(Maths.Cos(angle), Maths.Sin(angle))</c> or <c>Vector2.Right.Rotated(angle)</c>.
    /// </summary>
    /// <param name="angle">Angle of the vector, in radians.</param>
    /// <returns>The resulting vector.</returns>
    public static Vector2F FromAngle(float angle)
    {
        (float sin, float cos) = Maths.SinCos(angle);
        return new Vector2F(cos, sin);
    }

    /// <summary>
    /// Adds each component of the <see cref="Vector2F"/>
    /// with the components of the given <see cref="Vector2F"/>.
    /// </summary>
    /// <param name="left">The left vector.</param>
    /// <param name="right">The right vector.</param>
    /// <returns>The added vector.</returns>
    public static Vector2F operator +(Vector2F left, Vector2F right)
    {
        left.X += right.X;
        left.Y += right.Y;
        return left;
    }

    /// <summary>
    /// Subtracts each component of the <see cref="Vector2F"/>
    /// by the components of the given <see cref="Vector2F"/>.
    /// </summary>
    /// <param name="left">The left vector.</param>
    /// <param name="right">The right vector.</param>
    /// <returns>The subtracted vector.</returns>
    public static Vector2F operator -(Vector2F left, Vector2F right)
    {
        left.X -= right.X;
        left.Y -= right.Y;
        return left;
    }

    /// <summary>
    /// Returns the negative value of the <see cref="Vector2F"/>.
    /// This is the same as writing <c>new Vector2(-v.X, -v.Y)</c>.
    /// This operation flips the direction of the vector while
    /// keeping the same magnitude.
    /// With floats, the number zero can be either positive or negative.
    /// </summary>
    /// <param name="vec">The vector to negate/flip.</param>
    /// <returns>The negated/flipped vector.</returns>
    public static Vector2F operator -(Vector2F vec)
    {
        vec.X = -vec.X;
        vec.Y = -vec.Y;
        return vec;
    }

    /// <summary>
    /// Multiplies each component of the <see cref="Vector2F"/>
    /// by the given <see cref="float"/>.
    /// </summary>
    /// <param name="vec">The vector to multiply.</param>
    /// <param name="scale">The scale to multiply by.</param>
    /// <returns>The multiplied vector.</returns>
    public static Vector2F operator *(Vector2F vec, float scale)
    {
        vec.X *= scale;
        vec.Y *= scale;
        return vec;
    }

    /// <summary>
    /// Multiplies each component of the <see cref="Vector2F"/>
    /// by the given <see cref="float"/>.
    /// </summary>
    /// <param name="scale">The scale to multiply by.</param>
    /// <param name="vec">The vector to multiply.</param>
    /// <returns>The multiplied vector.</returns>
    public static Vector2F operator *(float scale, Vector2F vec)
    {
        vec.X *= scale;
        vec.Y *= scale;
        return vec;
    }

    /// <summary>
    /// Multiplies each component of the <see cref="Vector2F"/>
    /// by the components of the given <see cref="Vector2F"/>.
    /// </summary>
    /// <param name="left">The left vector.</param>
    /// <param name="right">The right vector.</param>
    /// <returns>The multiplied vector.</returns>
    public static Vector2F operator *(Vector2F left, Vector2F right)
    {
        left.X *= right.X;
        left.Y *= right.Y;
        return left;
    }

    /// <summary>
    /// Multiplies each component of the <see cref="Vector2F"/>
    /// by the given <see cref="float"/>.
    /// </summary>
    /// <param name="vec">The dividend vector.</param>
    /// <param name="divisor">The divisor value.</param>
    /// <returns>The divided vector.</returns>
    public static Vector2F operator /(Vector2F vec, float divisor)
    {
        vec.X /= divisor;
        vec.Y /= divisor;
        return vec;
    }

    /// <summary>
    /// Divides each component of the <see cref="Vector2F"/>
    /// by the components of the given <see cref="Vector2F"/>.
    /// </summary>
    /// <param name="vec">The dividend vector.</param>
    /// <param name="divisorv">The divisor vector.</param>
    /// <returns>The divided vector.</returns>
    public static Vector2F operator /(Vector2F vec, Vector2F divisorv)
    {
        vec.X /= divisorv.X;
        vec.Y /= divisorv.Y;
        return vec;
    }

    /// <summary>
    /// Gets the remainder of each component of the <see cref="Vector2F"/>
    /// with the components of the given <see cref="float"/>.
    /// This operation uses truncated division, which is often not desired
    /// as it does not work well with negative numbers.
    /// Consider using <see cref="PosMod(float)"/> instead
    /// if you want to handle negative numbers.
    /// </summary>
    /// <example>
    /// <code>
    /// GD.Print(new Vector2(10, -20) % 7); // Prints "(3, -6)"
    /// </code>
    /// </example>
    /// <param name="vec">The dividend vector.</param>
    /// <param name="divisor">The divisor value.</param>
    /// <returns>The remainder vector.</returns>
    public static Vector2F operator %(Vector2F vec, float divisor)
    {
        vec.X %= divisor;
        vec.Y %= divisor;
        return vec;
    }

    /// <summary>
    /// Gets the remainder of each component of the <see cref="Vector2F"/>
    /// with the components of the given <see cref="Vector2F"/>.
    /// This operation uses truncated division, which is often not desired
    /// as it does not work well with negative numbers.
    /// Consider using <see cref="PosMod(Vector2F)"/> instead
    /// if you want to handle negative numbers.
    /// </summary>
    /// <example>
    /// <code>
    /// GD.Print(new Vector2(10, -20) % new Vector2(7, 8)); // Prints "(3, -4)"
    /// </code>
    /// </example>
    /// <param name="vec">The dividend vector.</param>
    /// <param name="divisorv">The divisor vector.</param>
    /// <returns>The remainder vector.</returns>
    public static Vector2F operator %(Vector2F vec, Vector2F divisorv)
    {
        vec.X %= divisorv.X;
        vec.Y %= divisorv.Y;
        return vec;
    }

    /// <summary>
    /// Returns <see langword="true"/> if the vectors are exactly equal.
    /// Note: Due to floating-point precision errors, consider using
    /// <see cref="IsEqualApprox"/> instead, which is more reliable.
    /// </summary>
    /// <param name="left">The left vector.</param>
    /// <param name="right">The right vector.</param>
    /// <returns>Whether or not the vectors are exactly equal.</returns>
    public static bool operator ==(Vector2F left, Vector2F right)
    {
        return left.Equals(right);
    }

    /// <summary>
    /// Returns <see langword="true"/> if the vectors are not equal.
    /// Note: Due to floating-point precision errors, consider using
    /// <see cref="IsEqualApprox"/> instead, which is more reliable.
    /// </summary>
    /// <param name="left">The left vector.</param>
    /// <param name="right">The right vector.</param>
    /// <returns>Whether or not the vectors are not equal.</returns>
    public static bool operator !=(Vector2F left, Vector2F right)
    {
        return !left.Equals(right);
    }

    /// <summary>
    /// Compares two <see cref="Vector2F"/> vectors by first checking if
    /// the X value of the <paramref name="left"/> vector is less than
    /// the X value of the <paramref name="right"/> vector.
    /// If the X values are exactly equal, then it repeats this check
    /// with the Y values of the two vectors.
    /// This operator is useful for sorting vectors.
    /// </summary>
    /// <param name="left">The left vector.</param>
    /// <param name="right">The right vector.</param>
    /// <returns>Whether or not the left is less than the right.</returns>
    public static bool operator <(Vector2F left, Vector2F right)
    {
        if (left.X == right.X) return left.Y < right.Y;
        return left.X < right.X;
    }

    /// <summary>
    /// Compares two <see cref="Vector2F"/> vectors by first checking if
    /// the X value of the <paramref name="left"/> vector is greater than
    /// the X value of the <paramref name="right"/> vector.
    /// If the X values are exactly equal, then it repeats this check
    /// with the Y values of the two vectors.
    /// This operator is useful for sorting vectors.
    /// </summary>
    /// <param name="left">The left vector.</param>
    /// <param name="right">The right vector.</param>
    /// <returns>Whether or not the left is greater than the right.</returns>
    public static bool operator >(Vector2F left, Vector2F right)
    {
        if (left.X == right.X) return left.Y > right.Y;
        return left.X > right.X;
    }

    /// <summary>
    /// Compares two <see cref="Vector2F"/> vectors by first checking if
    /// the X value of the <paramref name="left"/> vector is less than
    /// or equal to the X value of the <paramref name="right"/> vector.
    /// If the X values are exactly equal, then it repeats this check
    /// with the Y values of the two vectors.
    /// This operator is useful for sorting vectors.
    /// </summary>
    /// <param name="left">The left vector.</param>
    /// <param name="right">The right vector.</param>
    /// <returns>Whether or not the left is less than or equal to the right.</returns>
    public static bool operator <=(Vector2F left, Vector2F right)
    {
        if (left.X == right.X) return left.Y <= right.Y;
        return left.X < right.X;
    }

    /// <summary>
    /// Compares two <see cref="Vector2F"/> vectors by first checking if
    /// the X value of the <paramref name="left"/> vector is greater than
    /// or equal to the X value of the <paramref name="right"/> vector.
    /// If the X values are exactly equal, then it repeats this check
    /// with the Y values of the two vectors.
    /// This operator is useful for sorting vectors.
    /// </summary>
    /// <param name="left">The left vector.</param>
    /// <param name="right">The right vector.</param>
    /// <returns>Whether or not the left is greater than or equal to the right.</returns>
    public static bool operator >=(Vector2F left, Vector2F right)
    {
        if (left.X == right.X) return left.Y >= right.Y;
        return left.X > right.X;
    }

    /// <summary>
    /// Returns <see langword="true"/> if the vector is exactly equal
    /// to the given object (<paramref name="obj"/>).
    /// Note: Due to floating-point precision errors, consider using
    /// <see cref="IsEqualApprox"/> instead, which is more reliable.
    /// </summary>
    /// <param name="obj">The object to compare with.</param>
    /// <returns>Whether or not the vector and the object are equal.</returns>
    public readonly override bool Equals(object obj)
    {
        return obj is Vector2F other && Equals(other);
    }

    /// <summary>
    /// Returns <see langword="true"/> if the vectors are exactly equal.
    /// Note: Due to floating-point precision errors, consider using
    /// <see cref="IsEqualApprox"/> instead, which is more reliable.
    /// </summary>
    /// <param name="other">The other vector.</param>
    /// <returns>Whether or not the vectors are exactly equal.</returns>
    public readonly bool Equals(Vector2F other)
    {
        return X == other.X && Y == other.Y;
    }

    /// <summary>
    /// Returns <see langword="true"/> if this vector and <paramref name="other"/> are approximately equal,
    /// by running <see cref="Maths.IsEqualApprox(float, float)"/> on each component.
    /// </summary>
    /// <param name="other">The other vector to compare.</param>
    /// <returns>Whether or not the vectors are approximately equal.</returns>
    public readonly bool IsEqualApprox(Vector2F other)
    {
        return Maths.IsEqualApprox(X, other.X) && Maths.IsEqualApprox(Y, other.Y);
    }

    /// <summary>
    /// Returns <see langword="true"/> if this vector's values are approximately zero,
    /// by running <see cref="Maths.IsZeroApprox(float)"/> on each component.
    /// This method is faster than using <see cref="IsEqualApprox"/> with one value
    /// as a zero vector.
    /// </summary>
    /// <returns>Whether or not the vector is approximately zero.</returns>
    public readonly bool IsZeroApprox()
    {
        return Maths.IsZeroApprox(X) && Maths.IsZeroApprox(Y);
    }

    /// <summary>
    /// Serves as the hash function for <see cref="Vector2F"/>.
    /// </summary>
    /// <returns>A hash code for this vector.</returns>
    public readonly override int GetHashCode()
    {
        return Y.GetHashCode() ^ X.GetHashCode();
    }

    /// <summary>
    /// Converts this <see cref="Vector2F"/> to a string.
    /// </summary>
    /// <returns>A string representation of this vector.</returns>
    public readonly override string ToString()
    {
        return $"({X.ToString(CultureInfo.InvariantCulture)}, {Y.ToString(CultureInfo.InvariantCulture)})";
    }


    /// <summary>
    /// Regular expression for parsing a string into a Vector2.
    /// </summary>
    public static Regex ParsingRegex { get; private set; } =
        new(@"^\s*\(\s*(-?\d+(?:\.\d+)?)\s*,\s*(-?\d+(?:\.\d+)?)\s*\)\s*$");

    /// <summary>
    /// Parses a string representation of a Vector2.
    /// </summary>
    /// <param name="text">The string to parse.</param>
    /// <returns>A parsed Vector2 object, or a default Vector2 if parsing fails.</returns>
    public static Vector2F Parse(string text)
    {
        // Attempt to match the input string against the parsing regex.
        var match = ParsingRegex.Match(text);

        // If the regex match is not successful, return a default Vector2.
        if (!match.Success)
            return default;

        // Extract the captured groups from the regex match.
        var groups = match.Groups;
        var xStr = groups[1].Value;
        var yStr = groups[2].Value;

        // Parse the string values to float and create a new Vector2.
        return new Vector2F
        {
            X = float.Parse(xStr, CultureInfo.InvariantCulture), Y = float.Parse(yStr, CultureInfo.InvariantCulture)
        };
    }
}