﻿using Godot;
using System.Collections.Generic;

namespace KludgeBox.Godot
{
	/// <summary>
	/// A static class that provides caching functionality for resources.
	/// </summary>
	public static class Cache
	{
		private static Dictionary<string, Resource> _resourcesCache = new();

		/// <summary>
		/// Gets the resource associated with the given resource path.
		/// If the resource is already cached, it will be returned from the cache.
		/// Otherwise, it will be loaded and added to the cache.
		/// </summary>
		/// <param name="path">The path of the resource to retrieve.</param>
		/// <returns>The requested resource if found, otherwise null.</returns>
		public static Resource Get(string path)
		{
			Resource resource;
			if (_resourcesCache.TryGetValue(path, out resource))
				if (GodotObject.IsInstanceValid(resource))
					return resource;

			// If the resource is not found in the cache or is invalid, load it and add it to the cache.
			resource = GD.Load<Resource>(path);
			_resourcesCache[path] = resource;

			return resource;
		}

		/// <summary>
		/// Gets the resource associated with the given resource path and casts it to the specified type.
		/// If the resource is already cached, it will be returned from the cache.
		/// Otherwise, it will be loaded and added to the cache.
		/// </summary>
		/// <typeparam name="TValue">The type to cast the resource to.</typeparam>
		/// <param name="path">The path of the resource to retrieve.</param>
		/// <returns>
		/// The requested resource if found and can be cast to the specified type, 
		/// otherwise null if the resource is not found or cannot be cast to the specified type.
		/// </returns>
		public static TValue Get<TValue>(string path) where TValue : Resource
		{
			Resource resource;
			if (_resourcesCache.TryGetValue(path, out resource))
			{
				TValue value = resource as TValue;
				if (GodotObject.IsInstanceValid(value))
					return value;
			}
				
			// If the resource is not found in the cache or is invalid, load it and add it to the cache.
			resource = GD.Load<TValue>(path);
			_resourcesCache[path] = resource;

			return resource as TValue;
		}
	}
}
