﻿using System;
using System.Linq;
using Godot;
using Godot.Collections;
using KludgeBox.Collections;
using KludgeBox.Godot.Nodes;
using KludgeBox.Structs;
using Color = Godot.Color;

namespace KludgeBox.Godot;

/// <summary>
	/// Static wrapper around existing methods to reduce code volume.
	/// </summary>
	public static class GodotExtensions
	{
		#region Vectors
		/// <summary>
		/// Extends the functionality of the Vector2 class by spinning it around a random angle.
		/// </summary>
		/// <param name="vector">The Vector2 instance to spin.</param>
		/// <returns>A new Vector2 instance representing the original vector rotated by a random angle.</returns>
		public static Vector2 Spin(this Vector2 vector)
		{
			float angle = Rand.Float * 2 * Maths.PI;
			vector = vector.Rotated(angle);

			return vector;
		}
		#endregion

		#region Rect Tools
		/// <summary>
		/// Gets the X-coordinate of the top-left corner of the specified Rect2.
		/// </summary>
		/// <param name="rect">The Rect2 object.</param>
		/// <returns>The X-coordinate of the top-left corner.</returns>
		public static float X(this Rect2 rect) => rect.Position.X;

		/// <summary>
		/// Gets the Y-coordinate of the top-left corner of the specified Rect2.
		/// </summary>
		/// <param name="rect">The Rect2 object.</param>
		/// <returns>The Y-coordinate of the top-left corner.</returns>
		public static float Y(this Rect2 rect) => rect.Position.Y;

		/// <summary>
		/// Gets the width of the specified Rect2.
		/// </summary>
		/// <param name="rect">The Rect2 object.</param>
		/// <returns>The width of the Rect2.</returns>
		public static float Width(this Rect2 rect) => rect.Size.X;

		/// <summary>
		/// Gets the height of the specified Rect2.
		/// </summary>
		/// <param name="rect">The Rect2 object.</param>
		/// <returns>The height of the Rect2.</returns>
		public static float Height(this Rect2 rect) => rect.Size.Y;

		public static float GetCircumradius(this Rect2 rect)
		{
			return Maths.Sqrt(rect.Width() * rect.Width() + rect.Height() * rect.Height()) / 2;
		}

		public static Circle2F GetCircumcircle(this Rect2 rect)
		{
			var radius = rect.GetCircumradius();
			var center = rect.GetCenter();
			return new Circle2F(new Vector2F(center.X, center.Y), radius);
		}
		
		/// <summary>
		/// Returns a new Rect2 representing the left half of the specified Rect2.
		/// </summary>
		/// <param name="rect">The Rect2 object.</param>
		/// <returns>A new Rect2 representing the left half of the original Rect2.</returns>
		public static Rect2 LeftHalf(this Rect2 rect)
		{
			return new Rect2(rect.Position, new Vector2(rect.Width() / 2f, rect.Height()));
		}

		/// <summary>
		/// Returns a new Rect2 representing a portion of the left side of the specified Rect2, based on the percentage provided.
		/// </summary>
		/// <param name="rect">The Rect2 object.</param>
		/// <param name="percentage">The percentage of the Rect2's width to include in the new Rect2. Value should be between 0 and 1.</param>
		/// <returns>A new Rect2 representing a portion of the left side of the original Rect2.</returns>
		public static Rect2 LeftPart(this Rect2 rect, float percentage)
		{
			return new Rect2(rect.Position, new Vector2(rect.Width() * percentage, rect.Height()));
		}

		/// <summary>
		/// Returns a new Rect2 representing a portion of the left side of the specified Rect2, based on the width provided in pixels.
		/// </summary>
		/// <param name="rect">The Rect2 object.</param>
		/// <param name="width">The width of the new Rect2 in pixels.</param>
		/// <returns>A new Rect2 representing a portion of the left side of the original Rect2.</returns>
		public static Rect2 LeftPartPixels(this Rect2 rect, float width)
		{
			return new Rect2(rect.Position, new Vector2(width, rect.Height()));
		}

		/// <summary>
		/// Returns a new <see cref="Rect2"/> representing the right half of the original rectangle.
		/// </summary>
		/// <param name="rect">The original rectangle.</param>
		/// <returns>A new <see cref="Rect2"/> representing the right half of the original rectangle.</returns>
		public static Rect2 RightHalf(this Rect2 rect)
		{
			return new Rect2(
				new Vector2(rect.X() + rect.Width() / 2f, rect.Y()),
				new Vector2(rect.Width() / 2f, rect.Height())
			);
		}

		/// <summary>
		/// Returns a new <see cref="Rect2"/> representing a portion of the original rectangle from the right side, based on the given percentage.
		/// </summary>
		/// <param name="rect">The original rectangle.</param>
		/// <param name="percentage">The percentage of the original rectangle's width to include in the new rectangle (0.0 - 1.0).</param>
		/// <returns>A new <see cref="Rect2"/> representing a portion of the original rectangle from the right side.</returns>
		public static Rect2 RightPart(this Rect2 rect, float percentage)
		{
			return new Rect2(
				new Vector2(rect.X() + rect.Width() * (1f - percentage), rect.Y()),
				new Vector2(rect.Width() * percentage, rect.Height())
			);
		}

		/// <summary>
		/// Returns a new <see cref="Rect2"/> representing a portion of the original rectangle from the right side, based on the given width in pixels.
		/// </summary>
		/// <param name="rect">The original rectangle.</param>
		/// <param name="width">The width in pixels of the new rectangle.</param>
		/// <returns>A new <see cref="Rect2"/> representing a portion of the original rectangle from the right side.</returns>
		public static Rect2 RightPartPixels(this Rect2 rect, float width)
		{
			return new Rect2(
				new Vector2(rect.X() + rect.Width() - width, rect.Y()),
				new Vector2(width, rect.Height())
			);
		}

		/// <summary>
		/// Returns a new <see cref="Rect2"/> representing the top half of the original rectangle.
		/// </summary>
		/// <param name="rect">The original rectangle.</param>
		/// <returns>A new <see cref="Rect2"/> representing the top half of the original rectangle.</returns>
		public static Rect2 TopHalf(this Rect2 rect)
		{
			return new Rect2(
				rect.Position,
				new Vector2(rect.Width(), rect.Height() / 2f)
			);
		}

		/// <summary>
		/// Returns a new <see cref="Rect2"/> representing a portion of the original rectangle from the top side, based on the given percentage.
		/// </summary>
		/// <param name="rect">The original rectangle.</param>
		/// <param name="percentage">The percentage of the original rectangle's height to include in the new rectangle (0.0 - 1.0).</param>
		/// <returns>A new <see cref="Rect2"/> representing a portion of the original rectangle from the top side.</returns>
		public static Rect2 TopPart(this Rect2 rect, float percentage)
		{
			return new Rect2(
				rect.Position,
				new Vector2(rect.Width(), rect.Height() * percentage)
			);
		}

		/// <summary>
		/// Returns a new <see cref="Rect2"/> representing a portion of the original rectangle from the top side, based on the given height in pixels.
		/// </summary>
		/// <param name="rect">The original rectangle.</param>
		/// <param name="height">The height in pixels of the new rectangle.</param>
		/// <returns>A new <see cref="Rect2"/> representing a portion of the original rectangle from the top side.</returns>
		public static Rect2 TopPartPixels(this Rect2 rect, float height)
		{
			return new Rect2(
				rect.Position,
				new Vector2(rect.Width(), height)
			);
		}

		/// <summary>
		/// Returns a new <see cref="Rect2"/> representing the bottom half of the original rectangle.
		/// </summary>
		/// <param name="rect">The original rectangle.</param>
		/// <returns>A new <see cref="Rect2"/> representing the bottom half of the original rectangle.</returns>
		public static Rect2 BottomHalf(this Rect2 rect)
		{
			return new Rect2(
				new Vector2(rect.X(), rect.Y() + rect.Height() / 2f),
				new Vector2(rect.Width(), rect.Height() / 2f)
			);
		}

		/// <summary>
		/// Returns a new <see cref="Rect2"/> representing a portion of the original rectangle from the bottom side, based on the given percentage.
		/// </summary>
		/// <param name="rect">The original rectangle.</param>
		/// <param name="percentage">The percentage of the original rectangle's height to include in the new rectangle (0.0 - 1.0).</param>
		/// <returns>A new <see cref="Rect2"/> representing a portion of the original rectangle from the bottom side.</returns>
		public static Rect2 BottomPart(this Rect2 rect, float percentage)
		{
			return new Rect2(
				new Vector2(rect.X(), rect.Y() + rect.Height() * (1f - percentage)),
				new Vector2(rect.Width(), rect.Height() * percentage)
			);
		}

		/// <summary>
		/// Returns a new <see cref="Rect2"/> representing a portion of the original rectangle from the bottom side, based on the given height in pixels.
		/// </summary>
		/// <param name="rect">The original rectangle.</param>
		/// <param name="height">The height in pixels of the new rectangle.</param>
		/// <returns>A new <see cref="Rect2"/> representing a portion of the original rectangle from the bottom side.</returns>
		public static Rect2 BottomPartPixels(this Rect2 rect, float height)
		{
			return new Rect2(
				new Vector2(rect.X(), rect.Y() + rect.Height() - height),
				new Vector2(rect.Width(), height)
			);
		}

		#endregion

		#region Nodes
		/// <summary>
		/// Shortcut for GodotObject.IsInstanceValid()
		/// </summary>
		/// <param name="gdObj"></param>
		/// <returns></returns>
		public static bool IsValid(this GodotObject gdObj)
		{
			return GodotObject.IsInstanceValid(gdObj);
		}
		
		/// <summary>
		/// Extension method that creates and returns a Destructor node to handle delayed removal of a given node.
		/// If the delay is less than or equal to 0, the node is immediately removed using QueueFree().
		/// </summary>
		/// <param name="node">The node to be destructed.</param>
		/// <param name="delay">The delay (in seconds) before the node is removed. Default is 0.</param>
		/// <returns>The created Destructor node, or null if the delay is 0 or negative.</returns>
		public static Destructor Destruct(this Node node, double delay = 0)
		{
			if (delay <= 0)
			{
				node.QueueFree();
				return null;
			}
			var destructor = new Destructor(delay);
			node.AddChild(destructor);
			return destructor;
		}

		/// <summary>
		/// Gets the absolute size of the sprite, taking into account the texture and global scale.
		/// </summary>
		/// <param name="sprite">The Sprite2D object.</param>
		/// <returns>The absolute size of the sprite.</returns>
		public static Vector2 GetAbsoluteSize(this Sprite2D sprite)
		{
			var texture = sprite.Texture;
			if (!GodotObject.IsInstanceValid(texture))
				return new Vector2();
			var size = sprite.Texture.GetSize();
			var scale = sprite.GlobalScale;

			return new Vector2(size.X * scale.X, size.Y * scale.Y);
		}

		/// <summary>
		/// Sets the absolute scale of the sprite based on the desired size.
		/// </summary>
		/// <param name="sprite">The Sprite2D object.</param>
		/// <param name="size">The desired absolute size of the sprite.</param>
		public static void SetAbsoluteScale(this Sprite2D sprite, Vector2 size)
		{
			var textureSize = sprite.Texture.GetSize();
			sprite.Scale = new Vector2(size.X / textureSize.X, size.Y / textureSize.Y);
		}

		/// <summary>
		/// Calculates the distance between two Node2D objects.
		/// </summary>
		/// <param name="node">The first Node2D object.</param>
		/// <param name="other">The second Node2D object.</param>
		/// <returns>The distance between the two Node2D objects.</returns>
		public static float DistanceTo(this Node2D node, Node2D other)
		{
			return node.Position.DistanceTo(other.Position);
		}

		/// <summary>
		/// Calculates the distance between a Node2D object and a Vector2 position.
		/// </summary>
		/// <param name="node">The Node2D object.</param>
		/// <param name="pos">The Vector2 position.</param>
		/// <returns>The distance between the Node2D object and the Vector2 position.</returns>
		public static float DistanceTo(this Node2D node, Vector2 pos)
		{
			return node.Position.DistanceTo(pos);
		}

		/// <summary>
		/// Calculates the distance between a Vector2 position and a Node2D object.
		/// </summary>
		/// <param name="pos">The Vector2 position.</param>
		/// <param name="other">The Node2D object.</param>
		/// <returns>The distance between the Vector2 position and the Node2D object.</returns>
		public static float DistanceTo(this Vector2 pos, Node2D other)
		{
			return pos.DistanceTo(other.Position);
		}

		/// <summary>
		/// Creates a Dummy2D object dropped at the position of the Node2D object.
		/// </summary>
		/// <param name="node">The Node2D object.</param>
		/// <param name="keepAlive">If true, the Dummy2D object will not be despawned when empty.</param>
		/// <returns>The created Dummy2D object or null if it couldn't be created.</returns>
		public static Dummy2D DropDummy(this Node2D node, bool keepAlive = false)
		{
			var parent = node.GetParent();
			if (!parent.IsValid()) return null;

			var dummy = new Dummy2D();
			dummy.Despawn = !keepAlive;
			parent.AddChild(dummy);

			dummy.Position = node.Position;
			dummy.Rotation = node.Rotation;

			return dummy;
		}

		/// <summary>
		/// Extension method that reparents the given node to a new parent node.
		/// If the current parent node is valid, the node is reparented to the new parent node using Reparent().
		/// If the current parent node is invalid but the new parent node is valid, the node is added as a child to the new parent.
		/// </summary>
		/// <param name="node">The node to be reparented.</param>
		/// <param name="newParent">The new parent node to reparent to.</param>
		public static void ParentTo(this Node node, Node newParent)
		{
			var currentParent = node.GetParent();

			if (GodotObject.IsInstanceValid(currentParent))
			{
				node.Reparent(newParent, true);
				return;
			}

			if (GodotObject.IsInstanceValid(newParent))
			{
				newParent.AddChild(node);
				node.Owner = newParent;
			}
		}

		/// <summary>
		/// Extension method that moves the node to the foreground within its parent's children.
		/// </summary>
		/// <param name="node">The node to be moved to the foreground.</param>
		public static void ToForeground(this Node node)
		{
			var parent = node.GetParent();

			if (GodotObject.IsInstanceValid(parent))
			{
				parent.MoveChild(node, -1);
			}
		}

		/// <summary>
		/// Extension method that moves the node to the background within its parent's children.
		/// </summary>
		/// <param name="node">The node to be moved to the background.</param>
		public static void ToBackground(this Node node)
		{
			var parent = node.GetParent();

			if (GodotObject.IsInstanceValid(parent))
			{
				parent.MoveChild(node, 0);
			}
		}

		/// <summary>
		/// Extension method that returns the visible rectangle (in global coordinates) of the Camera2D node's viewport.
		/// The visible rectangle is determined based on the camera's position, zoom level, and viewport size.
		/// </summary>
		/// <param name="camera">The Camera2D node for which to calculate the visible rectangle.</param>
		/// <returns>The visible rectangle in global coordinates.</returns>
		public static Rect2 GetVisibleRect(this Camera2D camera)
		{
			var viewport = camera.GetViewportRect();
			var cameraSize = viewport.Size / camera.Zoom;
			return new Rect2
			{
				Position = camera.Position - cameraSize / 2,
				Size = cameraSize
			};
		}

		/// <summary>
		/// Extension method that calculates and returns the radius of the visible area of the Camera2D node's viewport.
		/// The radius is determined based on the visible rectangle's circumradius.
		/// </summary>
		/// <param name="camera">The Camera2D node for which to calculate the visible area radius.</param>
		/// <returns>The radius of the visible area in global coordinates.</returns>
		public static float GetRadius(this Camera2D camera)
		{
			return camera.GetVisibleRect().GetCircumradius();
		}

		/// <summary>
		///		Returns true if there is at least one child of required type.
		/// </summary>
		/// <remarks>
		///		Note that under the hood it uses <see cref="GetChild"/> to find the child.
		/// </remarks>
		/// <typeparam name="T"></typeparam>
		/// <param name="node"></param>
		/// <returns></returns>
		public static bool HasChildOfType<T>(this Node node) where T : class
		{
			return node.GetChild<T>() is not null;
		}

		/// <summary>
		///		Returns first children of required type or null.
		/// </summary>
		/// <typeparam name="T"></typeparam>
		/// <param name="node"></param>
		/// <returns></returns>
		public static T GetChild<T>(this Node node) where T : class
		{
			if (node.GetChildCount() < 1)
				return default;

			Array<Node> children = node.GetChildren();
			return children.OfType<T>().FirstOrDefault();
		}

		/// <summary>
		/// Tries to recursively retrieve the parent node of the specified type from a given child node.
		/// </summary>
		/// <typeparam name="T">The type of the parent node to retrieve.</typeparam>
		/// <param name="child">The child node from which to start searching.</param>
		/// <returns>The parent node of the specified type if found, otherwise null.</returns>
		public static T TryGetParentOfType<T>(this Node child) where T : class
		{
			var parent = child.GetParent();
			if (parent is null)
				return default;

			if (parent is T)
				return parent as T;

			return parent.TryGetParentOfType<T>();
		}
		#endregion

		#region Color
		/// <summary>
		/// Extension method that converts a color to grayscale.
		/// </summary>
		/// <param name="color">The color to convert.</param>
		/// <returns>A grayscale color.</returns>
		public static Color Grayscale(this Color color)
		{
			float gray = color.GrayscaleValue();
			return new Color(gray, gray, gray, color.A);
		}

		/// <summary>
		/// Calculates the grayscale value of a color.
		/// </summary>
		/// <param name="color">The color to calculate the grayscale value for.</param>
		/// <returns>The grayscale value of the color.</returns>
		public static float GrayscaleValue(this Color color)
		{
			return 0.299f * color.R + 0.587f * color.G + 0.114f * color.B;
		}


		/// <summary>
		/// Converts a Vector3 object representing RGB values to a Color object.
		/// </summary>
		/// <param name="rgb">The Vector3 object representing the RGB values (X: Red, Y: Green, Z: Blue).</param>
		/// <returns>A Color object created from the RGB values.</returns>
		/// <remarks>
		/// The X component of the Vector3 object will be used as the Red value of the resulting Color object.
		/// The Y component of the Vector3 object will be used as the Green value of the resulting Color object.
		/// The Z component of the Vector3 object will be used as the Blue value of the resulting Color object.
		/// </remarks>
		public static Color ToColor(Vector3 rgb) => new Color(rgb.X, rgb.Y, rgb.Z);

		/// <summary>
		/// Darkens the specified color by the given amount.
		/// </summary>
		/// <param name="color">The color to darken.</param>
		/// <param name="amount">The amount by which to darken the color. Should be a value between 0 and 1.</param>
		/// <returns>The darkened color.</returns>
		public static Color Darken(this Color color, float amount)
		{
			amount = Math.Clamp(amount, 0, 1);
			return color with
			{
				R = color.R * (1 - amount),
				G = color.G * (1 - amount),
				B = color.B * (1 - amount)
			};
		}

		/// <summary>
		/// Lightens the specified color by the given amount.
		/// </summary>
		/// <param name="color">The color to lighten.</param>
		/// <param name="amount">The amount by which to lighten the color. Should be a value between 0 and 1.</param>
		/// <returns>The lightened color.</returns>
		public static Color Lighten(this Color color, float amount)
		{
			amount = Math.Clamp(amount, 0, 1);
			return color with
			{
				R = color.R + (1 - color.R) * amount,
				G = color.G + (1 - color.G) * amount,
				B = color.B + (1 - color.B) * amount
			};
		}
		#endregion
	}