﻿using System.Collections.Generic;
using System.Linq;
using Godot;
using KludgeBox.GameFeatures;
using static KludgeBox.Godot.GodotExtensions;

namespace KludgeBox.Godot.Nodes;

[GlobalClass]
public partial class Floor : Node2D
{
	/// <summary>
	/// Used for terrain generation on-the-fly
	/// </summary>
	[Export]
	public Camera2D Camera { get; set; }

	[Export]
	public string TexturePath { get; set; } = "res://Assets/Textures/floor1.png";

	public Texture2D Texture => _texture;

	private Cooldown _checksCooldown = new Cooldown(0.5, CooldownMode.Single);
	private Texture2D _texture;

	// Why TF I'm doing this?
	private List<FloorTile> _tiles = new List<FloorTile>();
	private Dictionary<Vector2I, FloorTile> _grid = new();

	public override void _Ready()
	{
		_checksCooldown.OnReady += CheckTiles;
		_texture = GD.Load<Texture2D>(TexturePath);
		CheckTiles();
	}


	public override void _Process(double delta)
	{
		_checksCooldown.Update(delta);
	}



	private void CheckTiles()
	{
		_checksCooldown.Use();
		var floorRadius = Camera.GetRadius() * 4;

		var width = Texture.GetWidth();
		var height = Texture.GetHeight();

		var gridSize = Maths.Ceil(floorRadius / Maths.Max(width, height));
		var gridRadius = (int)gridSize / 2;

		var currentTile = WorldToGrid(Camera.GlobalPosition, Texture.GetSize());

		for (int w = -gridRadius; w <= gridRadius; w++)
		{
			for (int h = -gridRadius; h <= gridRadius; h++)
			{
				FloorTile tile;
				var gridPos = currentTile + new Vector2I(w,h);
				if (!_grid.TryGetValue(gridPos, out tile))
				{
					tile = new FloorTile(this, gridPos);
					_grid[gridPos] = tile;
					AddChild(tile);
				}
			}
		}

		var markedForRemoval = _tiles.Where(t=>IsTooFar(t, floorRadius));
		foreach (var tile in markedForRemoval)
		{
			tile.QueueFree();
			_tiles.Remove(tile);
		}

		foreach ((var pos, var tile) in _grid)
		{
			if (markedForRemoval.Contains(tile))
				_grid.Remove(pos);
		}
	}


	public bool IsTooFar(FloorTile tile, double distance)
	{
		return tile.Position.DistanceTo(Camera.Position) > distance;
	}


	public Vector2 GridToWorld(Vector2I gridPos, Vector2 tileSize)
	{
		return new Vector2(
			gridPos.X * tileSize.X,
			gridPos.Y * tileSize.Y);
	}


	public Vector2I WorldToGrid(Vector2 pos, Vector2 tileSize)
	{
		return new Vector2I(
			(int)(pos.X/tileSize.X),
			(int)(pos.Y/tileSize.Y));
	}
}