﻿using System;
using System.Collections.Generic;
using KludgeBox.Structs;
using Newtonsoft.Json.Linq;

namespace KludgeBox.SaveLoad
{
	/// <summary>
	/// Class responsible for parsing values from the strings.
	/// </summary>
	public static class Parser
	{
		/// <summary>
		/// The class is responsible for containing and registration of the different type parsers.
		/// Since C# generates separate static class for every TValue it's can be easily used for quick access.
		/// </summary>
		/// <typeparam name="TValue"></typeparam>
		public static class Parsers<TValue>
		{
			public static Func<string, TValue> parser;

			/// <summary>
			/// Registers the parsing function for TValue type
			/// </summary>
			/// <param name="func"></param>
			public static void Register(Func<string, TValue> func)
			{
				parser = func;
				_parsers[typeof(TValue)] = (string str) => func(str);
			}
		}
		
		


		private static Dictionary<Type,Func<string, object>> _parsers = new();


		static Parser()
		{
			Parsers<bool>.Register(ParseBool);
			Parsers<int>.Register(ParseInt);
			Parsers<float>.Register(ParseFloat);
			Parsers<double>.Register(ParseDouble);
			Parsers<string>.Register(ParseString);
			Parsers<Vector2F>.Register(ParseVec2F);
			Parsers<Vector2D>.Register(ParseVec2D);
			Parsers<Vector2I>.Register(ParseVec2I);
			Parsers<Rect2F>.Register(ParseRect2F);
			Parsers<Rect2D>.Register(ParseRect2D);
			Parsers<Rect2I>.Register(ParseRect2I);
			Parsers<KColor>.Register(ParseColor);
			Parsers<Circle2F>.Register(ParseCircle);
			Parsers<Circle2D>.Register(ParseCircleD);
		}

		public static TValue Parse<TValue>(string str)
		{
			return (TValue)_parsers[typeof(TValue)](str);
		}

		public static object Parse(string str, Type type)
		{
			return _parsers[type](str);
		}

		public static bool CanParse(Type type)
		{
			return _parsers.ContainsKey(type);
		}

		public static bool IsNull(string propertyName)
		{
			if (!SaveLoad.Loader.IsWorking)
				return true;

			var property = SaveLoad.Loader.CurrentObject[propertyName];
			if (property is null)
				return true;

			return IsNull(property);
		}

		public static bool IsNull(JToken token)
		{
			var type = token.Type;
			if (type is JTokenType.Object)
			{
				var obj = (JObject) token;
				if (obj.ContainsKey("isNull"))
					return true;
				return false;
			}

			if (type is JTokenType.Property)
			{
				return IsNull(((JProperty)token).Value);
			}

			return false;
		}

		public static bool ParseBool(string str)
		{
			return Boolean.Parse(str);
		}

		public static int ParseInt(string str)
		{
			return Int32.Parse(str);
		}

		public static float ParseFloat(string str)
		{
			return Single.Parse(str);
		}

		public static double ParseDouble(string str)
		{
			return Double.Parse(str);
		}

		public static string ParseString(string str)
		{
			return str;
		}

		public static Vector2F ParseVec2F(string str)
		{
			return Vector2F.Parse(str);
		}
		
		public static Vector2D ParseVec2D(string str)
		{
			return Vector2D.Parse(str);
		}

		public static Vector2I ParseVec2I(string str)
		{
			return Vector2I.Parse(str);
		}

		public static Rect2F ParseRect2F(string str)
		{
			return Rect2F.Parse(str);
		}
		
		public static Rect2D ParseRect2D(string str)
		{
			return Rect2D.Parse(str);
		}

		public static Rect2I ParseRect2I(string str)
		{
			return Rect2I.Parse(str);
		}

		public static KColor ParseColor(string str)
		{
			return KColor.Parse(str);
		}

		public static Circle2F ParseCircle(string str)
		{
			return Circle2F.Parse(str);
		}

		public static Circle2D ParseCircleD(string str)
		{
			return Circle2D.Parse(str);
		}
	}
}
