﻿using KludgeBox.VFS;
using KludgeBox.VFS.FileSystems;

namespace KludgeBox.Tests.VFS.FileSystems;

[TestFixture]
public class ComplexFileSystemTests
{
    private const string TestFile = "test.txt";
    private const string PhysicalRoot = "tests/";
    
    private const string Dir1 = "/dir1/";
    private const string Dir2 = "/dir2/";
    private const string Dir3 = Dir2 + "dir3/";
    private const string Dir4 = Dir1 + "dir4/";
    
    private const string File1 = Dir1 + "file1.txt";
    private const string File2 = Dir2 + "file2.txt";
    private const string File3 = Dir3 + "file2.txt";
    private const string File4 = Dir4 + "file1.txt";

    private const string FileContents1 = "This is a test file.";
    private const string FileContents2 = "This is another test file.";
    private const string FileContents3 = $"{FileContents1}\n{FileContents2}";

    private static FileSystem[] RwFileSystems => new FileSystem[]
    {
        new MemoryFileSystem(true),
        new MemoryFileSystem(false),
        new MinimalFileSystem(),
        new PhysicalFileSystem(Path.GetFullPath(PhysicalRoot))
    };

    [SetUp]
    public void SetUp()
    {
        if(Directory.Exists(PhysicalRoot))
            Directory.Delete(PhysicalRoot, true);
    }
    
    [Timeout(1000)]
    [TestCaseSource(nameof(RwFileSystems))]
    public void FileSystemTests(FileSystem fileSystem)
    {
        // Create and write test
        fileSystem.CreateDirectory(Dir1);
        Assert.That(fileSystem.IsDirectory(Dir1), Is.True);
        fileSystem.CreateFile(File1);
        Assert.That(fileSystem.IsFile(File1), Is.True); // File1 should exist.
        
        var emptyContents = fileSystem.ReadAllText(File1);
        Assert.That(emptyContents, Is.EqualTo(String.Empty)); // File1 should be empty.
        
        fileSystem.WriteAllText(File1, FileContents1);
        Assert.That(fileSystem.ReadAllText(File1), Is.EqualTo(FileContents1)); // File1 should have the correct contents.

        // Move test
        fileSystem.CreateDirectory(Dir2);
        fileSystem.MoveFile(File1, File2);
        Assert.That(fileSystem.IsFile(File2), Is.True); // new file exists.
        Assert.That(fileSystem.IsFile(File1), Is.False); // old file does not exist.
        Assert.That(() => fileSystem.GetFiles(Dir1).First(f => f.Path == File1), Throws.Exception.TypeOf<InvalidOperationException>()); // old file also does not exist in it's parent directory.
        Assert.That(() => fileSystem.GetFiles(Dir2).First(f => f.Path == File2), Throws.Nothing); // new file also does exist in it's parent directory.
        Assert.That(fileSystem.ReadAllText(File2), Is.EqualTo(FileContents1)); // new file should have the correct contents.
        
        // Copy test
        fileSystem.CopyFile(File2, File1);
        Assert.That(fileSystem.IsFile(File1), Is.True); // File1 should exist.
        Assert.That(fileSystem.IsFile(File2), Is.True); // File2 should exist.
        Assert.That(fileSystem.ReadAllText(File1), Is.EqualTo(fileSystem.ReadAllText(File2))); // File1 should have the the same contents as File2);
        Assert.That(() => fileSystem.CopyDirectory(Dir1, Dir2), Throws.Exception.TypeOf<IOException>()); // Attempt to copy directory to existing directory throws IOException.
        
        fileSystem.CopyDirectory(Dir2, Dir3);
        fileSystem.CopyDirectory(Dir1, Dir4);
        
        Assert.That(fileSystem.IsDirectory(Dir3), Is.True); // Dir3 should exist.
        Assert.That(fileSystem.IsDirectory(Dir4), Is.True); // Dir4 should exist.
        
        Assert.That(fileSystem.IsFile(File3), Is.True); // File3 should exist.
        Assert.That(fileSystem.IsFile(File4), Is.True); // File4 should exist.
        Assert.That(fileSystem.ReadAllText(File3), Is.EqualTo(FileContents1)); // File3 should have the correct contents.
        Assert.That(fileSystem.ReadAllText(File4), Is.EqualTo(FileContents1)); // File4 should have the correct contents.
        
        // Append test
        fileSystem.AppendText(File3, "\n"+FileContents2);
        fileSystem.AppendText(File4, "\n"+FileContents2);
        Assert.That(fileSystem.ReadAllText(File3), Is.EqualTo(FileContents3));
        Assert.That(fileSystem.ReadAllText(File4), Is.EqualTo(FileContents3));
        
    }
}