﻿using System.Text;
using KludgeBox.VFS;
using KludgeBox.VFS.FileSystems;
using static KludgeBox.VFS.PathHelper;

namespace KludgeBox.Tests.VFS.FileSystems;

public class GeneralFileSystemTests
{
    private const string TestFile = "test.txt";
    private const string PhysicalRoot = "tests/";
    
    private const string Dir1 = "/dir1/";
    private const string Dir2 = "/dir2/";
    private const string Dir3 = Dir1 + "dir3/";
    private const string Dir4 = Dir2+ "dir4/";
    
    private const string File1 = Dir1 + "file1.txt";
    private const string File2 = Dir2 + "file2.txt";
    private const string File3 = Dir3 + "file2.txt";
    private const string File4 = Dir4 + "file1.txt";

    private const string FileContents1 = "This is a test file.";
    private const string FileContents2 = "This is another test file.";
    private const string FileContents3 = $"{FileContents1}\n{FileContents2}";

    private static FileSystem[] RwFileSystems
    {
        get
        {
            var physFs = new PhysicalFileSystem(Path.GetFullPath(PhysicalRoot)); // physical FS
            var memMultipart = new MemoryFileSystem(true);// multipart MemFS
            var memMonolith = new MemoryFileSystem(false);// default MemFS
            var minFs = new MinimalFileSystem(); // minimal FS
            
            var innerMem = new MemoryFileSystem(); // FS for proxy
            FileSystem proxyFs = null;

            try
            {
                var proxyRoot = "/path1/path2/path3/";
                innerMem.CreateDirectory(proxyRoot);
                proxyFs = new ProxyFileSystem(innerMem, proxyRoot);
            }
            catch (Exception e)
            {
                Console.WriteLine(e);
                try
                {
                    proxyFs = new ProxyFileSystem(innerMem, "/");
                }
                catch (Exception e2)
                {
                    Console.WriteLine(e2);
                    proxyFs = innerMem;
                }
            }
            
            return new FileSystem[]
            {
                memMultipart,
                memMonolith,
                minFs,
                physFs, 
                proxyFs
            };
        }
    }

    private static FsFile AddFile(FileSystem fs)
    {
        fs.CreateDirectory(Dir1);
        return fs.CreateFile(File1);
    }
    
    private static FsFile AddAnotherFile(FileSystem fs)
    {
        fs.CreateDirectory(Dir2);
        return fs.CreateFile(File2);
    }
    
    [SetUp]
    public void SetUp()
    {
        if(Directory.Exists(PhysicalRoot))
            Directory.Delete(PhysicalRoot, true);
        Directory.CreateDirectory(PhysicalRoot);
    }

    [TearDown]
    public void TearDown()
    {
        if(Directory.Exists(PhysicalRoot))
            Directory.Delete(PhysicalRoot, true);
    }

    #region Tests
    [TestCaseSource(nameof(RwFileSystems))]
    public void CreateDirectory_CreateRecursively(FileSystem fs)
    {
        fs.CreateDirectory(Dir4);
        fs.CreateDirectory(Dir3);
        
        Assert.That(fs.IsDirectory(Dir1), Is.True);
        Assert.That(fs.IsDirectory(Dir2), Is.True);
        Assert.That(fs.IsDirectory(Dir3), Is.True);
        Assert.That(fs.IsDirectory(Dir4), Is.True);
        
        Assert.That(fs.GetDirectories(Dir1).Length, Is.EqualTo(1));
        Assert.That(fs.GetDirectories(Dir2).Length, Is.EqualTo(1));
        Assert.That(fs.GetDirectories(Dir3).Length, Is.EqualTo(0));
        Assert.That(fs.GetDirectories(Dir4).Length, Is.EqualTo(0));

        var dirs = fs.GetDirectories(Dir1);
        Assert.That(fs.GetDirectories(Dir1).Contains(new FsDirectory(fs, Dir3)), Is.True);
        Assert.That(fs.GetDirectories(Dir2).Contains(new FsDirectory(fs, Dir4)), Is.True);
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void CreateDirectory_ReturnsExistingDirectory(FileSystem fs)
    {
        var dir1 = fs.CreateDirectory(Dir4);
        var dir2 = fs.CreateDirectory(Dir4);
        
        Assert.That(dir1, Is.EqualTo(dir2));
        Assert.That(dir1.Entries, Is.EqualTo(dir2.Entries));
    }


    [TestCaseSource(nameof(RwFileSystems))]
    public void CreateDirectory_TrailingSeparator(FileSystem fs)
    {
        fs.CreateDirectory(Dir1);
        fs.CreateDirectory(Dir2.TrimEndingDirectorySeparator());
        
        Assert.That(fs.IsDirectory(Dir1));
        Assert.That(fs.IsDirectory(Dir1.TrimEndingDirectorySeparator()));
        
        Assert.That(fs.IsDirectory(Dir2));
        Assert.That(fs.IsDirectory(Dir2.TrimEndingDirectorySeparator()));
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void CreateAndWriteToFile(FileSystem fs)
    {
        fs.CreateDirectory(Dir1);
        var file = fs.CreateFile(File1);
        
        Assert.That(fs.IsFile(File1), Is.True);
        Assert.That(fs.ReadAllText(File1), Is.EqualTo(""));
        
        fs.WriteAllText(File1, FileContents1);
        Assert.That(fs.ReadAllText(File1), Is.EqualTo(FileContents1));
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void CreateFile_RewritesExistingFile(FileSystem fs)
    {
        var file = AddFile(fs);
        file.WriteAllText(FileContents1);
        Assert.That(fs.ReadAllText(File1), Is.EqualTo(FileContents1));
        
        fs.CreateFile(File1);
        Assert.That(fs.ReadAllText(File1), Is.EqualTo(""));
    }

    [TestCaseSource(nameof(RwFileSystems))]
    public void CreateFile_PathWithEndingSeparator(FileSystem fs)
    {
        Assert.That(()=>fs.CreateFile(Dir1), Throws.Nothing);
        Assert.That(()=>fs.CreateFile(Dir2), Throws.Nothing);
        
        Assert.That(fs.IsDirectory(Dir1), Is.False);
        Assert.That(fs.IsFile(Dir1), Is.False);
        Assert.That(fs.IsFile(Dir1.TrimEndingDirectorySeparator()));
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void OpenWrite_CreatesFileIfItDoesNotExist(FileSystem fs)
    {
        Assert.That(fs.IsFile(File1), Is.False);
        fs.CreateDirectory(Dir1);
        var writer = fs.OpenWrite(File1);
        writer.Dispose();
        Assert.That(fs.IsFile(File1), Is.True);
        Assert.That(fs.ReadAllText(File1), Is.EqualTo(""));
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void OpenRead_ThrowsIfFileDoesNotExist(FileSystem fs)
    {
        fs.CreateDirectory(Dir1);
        Assert.That(() => fs.OpenRead(File1), Throws.Exception.TypeOf<FileNotFoundException>());
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void OpenRead(FileSystem fs)
    {
        var file = AddFile(fs);
        file.WriteAllText(FileContents1);
        using var reader = file.OpenRead();
        var memoryStream = new MemoryStream();
        reader.CopyTo(memoryStream);
        Assert.That(memoryStream.ToArray(), Is.EqualTo(Encoding.UTF8.GetBytes(FileContents1)));
    }


    [TestCaseSource(nameof(RwFileSystems))]
    public void OpenWrite_DoesNotChangesExistingFileIfNotWritten(FileSystem fs)
    {
        var file = AddFile(fs);
        file.WriteAllText(FileContents1);
        Assert.That(fs.ReadAllText(File1), Is.EqualTo(FileContents1));
        
        var writer = fs.OpenWrite(File1);
        writer.Dispose();
        Assert.That(fs.ReadAllText(File1), Is.EqualTo(FileContents1));
    }

    [TestCaseSource(nameof(RwFileSystems))]
    public void OpenWrite_OverwritesExistingFileIfWritten(FileSystem fs)
    {
        var file = AddFile(fs);
        file.WriteAllText(FileContents1);
        Assert.That(fs.ReadAllText(File1), Is.EqualTo(FileContents1));
        
        var writer = fs.OpenWrite(File1);
        writer.Write(Encoding.UTF8.GetBytes(FileContents2));
        writer.Dispose();
        Assert.That(fs.ReadAllText(File1), Is.EqualTo(FileContents2));
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void MoveFile(FileSystem fs)
    {
        AddFile(fs);
        fs.CreateDirectory(Dir2);
        fs.WriteAllText(File1, FileContents1);
        fs.MoveFile(File1, File2);
        
        Assert.That(fs.IsFile(File1), Is.False);
        Assert.That(fs.IsFile(File2), Is.True);
        Assert.That(fs.ReadAllText(File2), Is.EqualTo(FileContents1));
    }


    [TestCaseSource(nameof(RwFileSystems))]
    public void MoveFile_ThrowsIfSourceFileDoesNotExist(FileSystem fs)
    {
        fs.CreateDirectory(Dir1);
        fs.CreateDirectory(Dir2);
        Assert.That(() => fs.MoveFile(File1, File2), Throws.Exception.TypeOf<FileNotFoundException>());
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void MoveFile_ThrowsIfDestinationFileExists(FileSystem fs)
    {
        AddFile(fs);
        AddAnotherFile(fs);
        fs.WriteAllText(File1, FileContents1);
        
        Assert.That(() => fs.MoveFile(File1, File2), Throws.Exception.TypeOf<IOException>());
        Assert.That(fs.IsFile(File1), Is.True);
        Assert.That(fs.ReadAllText(File1), Is.EqualTo(FileContents1));
        Assert.That(fs.ReadAllText(File2), Is.EqualTo(""));
    }


    [TestCaseSource(nameof(RwFileSystems))]
    public void MoveFile_Overwrite(FileSystem fs)
    {
        AddFile(fs);
        AddAnotherFile(fs);
        fs.WriteAllText(File1, FileContents1);

        Assert.That(() => fs.MoveFile(File1, File2, true), Throws.Nothing);
        Assert.That(fs.IsFile(File1), Is.False);
        Assert.That(fs.IsFile(File2), Is.True);
        Assert.That(fs.ReadAllText(File2), Is.EqualTo(FileContents1));
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void CopyFile(FileSystem fs)
    {
        AddFile(fs);
        fs.CreateDirectory(Dir2);
        fs.WriteAllText(File1, FileContents1);
        fs.CopyFile(File1, File2);
        
        Assert.That(fs.IsFile(File1), Is.True);
        Assert.That(fs.IsFile(File2), Is.True);
        Assert.That(fs.ReadAllText(File2), Is.EqualTo(fs.ReadAllText(File1)));
    }


    [TestCaseSource(nameof(RwFileSystems))]
    public void CopyFile_ThrowsIfSourceFileDoesNotExist(FileSystem fs)
    {
        fs.CreateDirectory(Dir1);
        fs.CreateDirectory(Dir2);
        
        Assert.That(() => fs.CopyFile(File1, File2), Throws.Exception.TypeOf<FileNotFoundException>());
    }
    
    
    [TestCaseSource(nameof(RwFileSystems))]
    public void CopyFile_ThrowsIfDestinationFileExists(FileSystem fs)
    {
        AddFile(fs);
        AddAnotherFile(fs);
        fs.WriteAllText(File1, FileContents1);
        
        Assert.That(() => fs.CopyFile(File1, File2), Throws.Exception.TypeOf<IOException>());
        Assert.That(fs.ReadAllText(File1), Is.EqualTo(FileContents1));
        Assert.That(fs.ReadAllText(File2), Is.EqualTo(""));
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void CopyFile_Overwrite(FileSystem fs)
    {
        AddFile(fs);
        AddAnotherFile(fs);
        fs.WriteAllText(File1, FileContents1);
        
        Assert.That(() => fs.CopyFile(File1, File2, true), Throws.Nothing);
        Assert.That(fs.ReadAllText(File1), Is.EqualTo(FileContents1));
        Assert.That(fs.ReadAllText(File2), Is.EqualTo(FileContents1));
    }


    [TestCaseSource(nameof(RwFileSystems))]
    public void CopyDirectory(FileSystem fs)
    {
        AddFile(fs);
        fs.WriteAllText(File1, FileContents1);
        fs.CopyDirectory(Dir1, Dir2);
        
        Assert.That(fs.IsDirectory(Dir1), Is.True);
        Assert.That(fs.IsDirectory(Dir2), Is.True);
        Assert.That(fs.IsFile(File1), Is.True);
        Assert.That(fs.ReadAllText(Combine(Dir2, File1.GetName())), Is.EqualTo(FileContents1));
    }


    [TestCaseSource(nameof(RwFileSystems))]
    public void CopyDirectory_ThrowsIfSourceDirectoryDoesNotExist(FileSystem fs)
    {
        Assert.That(() => fs.CopyDirectory(Dir1, Dir2), Throws.Exception.TypeOf<DirectoryNotFoundException>());
        Assert.That(fs.IsDirectory(Dir1), Is.False);
        Assert.That(fs.IsDirectory(Dir2), Is.False);
    }
    

    [TestCaseSource(nameof(RwFileSystems))]
    public void CopyDirectory_ThrowsIfDestinationDirectoryExists(FileSystem fs)
    {
        fs.CreateDirectory(Dir1);
        fs.CreateDirectory(Dir2);
        
        Assert.That(() => fs.CopyDirectory(Dir1, Dir2), Throws.Exception.TypeOf<IOException>());
    }


    [TestCaseSource(nameof(RwFileSystems))]
    public void MoveDirectory(FileSystem fs)
    {
        AddFile(fs);
        fs.WriteAllText(File1, FileContents1);
        fs.MoveDirectory(Dir1, Dir2);
        
        Assert.That(fs.IsDirectory(Dir1), Is.False);
        Assert.That(fs.IsDirectory(Dir2), Is.True);
        Assert.That(fs.IsFile(File1), Is.False);
        Assert.That(fs.IsFile(Combine(Dir2, File1.GetName())), Is.True);
        Assert.That(fs.ReadAllText(Combine(Dir2, File1.GetName())), Is.EqualTo(FileContents1));
    }


    [TestCaseSource(nameof(RwFileSystems))]
    public void DeleteDirectory_Empty(FileSystem fs)
    {
        fs.CreateDirectory(Dir1);
        Assert.That(fs.IsDirectory(Dir1), Is.True);
        fs.DeleteDirectory(Dir1);
        Assert.That(fs.IsDirectory(Dir1), Is.False);
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void DeleteDirectory_NotEmpty(FileSystem fs)
    {
        AddFile(fs);
        fs.WriteAllText(File1, FileContents1);
        Assert.That(fs.IsFile(File1), Is.True);
        fs.DeleteDirectory(Dir1, true);
        Assert.That(fs.IsDirectory(Dir1), Is.False);
        Assert.That(fs.IsFile(File1), Is.False);
    }


    [TestCaseSource(nameof(RwFileSystems))]
    public void DeleteDirectory_Recursive(FileSystem fs)
    {
        fs.CreateDirectory(Dir3);
        fs.WriteAllText(File3, FileContents3);
        
        Assert.That(() => fs.DeleteDirectory(Dir1), Throws.Exception.TypeOf<IOException>());
        Assert.That(fs.IsFile(File3), Is.True);
        Assert.That(fs.IsDirectory(Dir3), Is.True);
        
        fs.DeleteDirectory(Dir1, true);
        Assert.That(fs.IsDirectory(Dir3), Is.False);
        Assert.That(fs.IsFile(File3), Is.False);
        Assert.That(fs.IsDirectory(Dir1), Is.False);
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void DeleteDirectory_ThrowsIfDeletingFile(FileSystem fs)
    {
        AddFile(fs);
        Assert.That(fs.IsFile(File1));
        Assert.That(()=>fs.DeleteDirectory(File1), Throws.Exception.TypeOf<DirectoryNotFoundException>());
        Assert.That(fs.IsFile(File1));
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void DeleteDirectory_ThrowsIfDoesNotExist(FileSystem fs)
    {
        fs.CreateDirectory(Dir1);
        Assert.That(()=>fs.DeleteDirectory(Dir1), Throws.Nothing);
        Assert.That(()=>fs.DeleteDirectory(Dir1), Throws.Exception.TypeOf<DirectoryNotFoundException>());
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void DeleteDirectory_ThrowsIfDeletingRoot(FileSystem fs)
    {
        Assert.That(fs.Root.Exists);
        Assert.That(()=> fs.Root.Delete(), Throws.Exception.TypeOf<IOException>());
        Assert.That(fs.Root.Exists);
    }


    [TestCaseSource(nameof(RwFileSystems))]
    public void DeleteDirectory_ThrowsIfMovingRoot(FileSystem fs)
    {
        Assert.That(fs.Root.Exists);
        Assert.That(()=> fs.MoveDirectory("/", Dir1), Throws.Exception.TypeOf<IOException>());
        Assert.That(fs.Root.Exists);
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void ReadAllBytes(FileSystem fs)
    {
        AddFile(fs);
        byte[] bytes = new byte[1024];
        Random.Shared.NextBytes(bytes);
        fs.WriteAllBytes(File1, bytes);
        var readBytes = fs.ReadAllBytes(File1);
        Assert.That(readBytes, Is.EqualTo(bytes));
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void ReadAllBytes_ThrowsFileNotFoundException(FileSystem fs)
    {
        AddFile(fs);
        Assert.That(()=>fs.ReadAllBytes(File1), Throws.Nothing);
        fs.DeleteFile(File1);
        Assert.That(()=>fs.ReadAllBytes(File1), Throws.Exception.TypeOf<FileNotFoundException>());
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void Root(FileSystem fs)
    {
        if(fs is IProxyFileSystem proxy)
            Assert.That(fs.Root.RealPath, Is.EqualTo(proxy.RealPath));
    }
    

    [TestCaseSource(nameof(RwFileSystems))]
    public void GetDirectory(FileSystem fs)
    {
        AddFile(fs);
        Assert.That(()=>fs.GetDirectory(Dir1), Throws.Nothing);
        Assert.That(fs.GetDirectory(Dir1).Exists);
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void GetDirectory_ThrowsIfGettingFile(FileSystem fs)
    {
        AddFile(fs);
        Assert.That(fs.IsFile(File1));
        Assert.That(()=>fs.GetDirectory(File1), Throws.Exception.TypeOf<DirectoryNotFoundException>());
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void GetDirectory_ThrowsIfDoesNotExist(FileSystem fs)
    {
        fs.CreateDirectory(Dir1);
        Assert.That(()=>fs.GetDirectory(Dir1), Throws.Nothing);
        fs.DeleteDirectory(Dir1);
        Assert.That(()=>fs.GetDirectory(Dir1), Throws.Exception.TypeOf<DirectoryNotFoundException>());
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void GetFile(FileSystem fs)
    {
        AddFile(fs);
        Assert.That(()=>fs.GetFile(File1), Throws.Nothing);
        Assert.That(fs.GetFile(File1).Exists);
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void GetFile_ThrowsIfGettingDirectory(FileSystem fs)
    {
        AddFile(fs);
        Assert.That(fs.IsDirectory(Dir1));
        Assert.That(()=>fs.GetFile(Dir1), Throws.Exception.TypeOf<FileNotFoundException>());
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void GetFile_ThrowsIfDoesNotExist(FileSystem fs)
    {
        AddFile(fs);
        Assert.That(fs.IsFile(File1));
        fs.DeleteFile(File1);
        Assert.That(()=>fs.GetFile(File1), Throws.Exception.TypeOf<FileNotFoundException>());
    }
    
    
    [TestCaseSource(nameof(RwFileSystems))]
    public void GetEntry(FileSystem fs)
    {
        AddFile(fs);
        Assert.That(()=>fs.GetEntry(File1), Throws.Nothing);
        Assert.That(fs.GetEntry(File1).Exists);
    }

    
    [TestCaseSource(nameof(RwFileSystems))]
    public void GetEntry_ThrowsIfDoesNotExist(FileSystem fs)
    {
        AddFile(fs);
        Assert.That(fs.IsFile(File1));
        fs.DeleteFile(File1);
        Assert.That(()=>fs.GetEntry(File1), Throws.Exception.TypeOf<FileNotFoundException>());
    }
    #endregion
}