﻿using KludgeBox.VFS;
using KludgeBox.VFS.FileSystems;

namespace KludgeBox.Tests.VFS.FileSystems;

[TestFixture]
public class MergedFileSystemTests
{
    private const string ProxyRoot = "/path1/path2/path3/";
    private const string PhysicalRoot = "tests/";
    private const string File = "file.txt";

    private const string PhysicalDir = "/phys/";
    private const string MemoryDir = "/mem/";
    private const string ProxyDir = "/proxy/";
    private const string SharedDir = "/shared/";

    private const string PhysicalFile = PhysicalDir + File;
    private const string MemoryFile = MemoryDir + File;
    private const string ProxyFile = ProxyDir + File;
    private const string SharedFile = SharedDir + File;

    private const string FileContents1 = "First content";
    private const string FileContents2 = "Second content";
    
    
    private MergedFileSystem _merged;
    private MemoryFileSystem _memory;
    private PhysicalFileSystem _physical;
    private ProxyFileSystem _proxyToMemory;
    private ProxyFileSystem _proxyToPhysical;
    private MemoryFileSystem _conflict;

    [SetUp]
    public void SetUp()
    {
        Cleanup();
        Directory.CreateDirectory(PhysicalRoot);
        CreateFs();
        PopulateFs();
    }

    private void Cleanup()
    {
        if(Directory.Exists(PhysicalRoot))
            Directory.Delete(PhysicalRoot, true);
    }
    
    private void CreateFs()
    {
        _physical = new (Path.GetFullPath(PhysicalRoot));
        _memory = new();
        
        _physical.CreateDirectory(ProxyRoot);
        _memory.CreateDirectory(ProxyRoot);

        _proxyToMemory = new (_memory, ProxyRoot);
        _proxyToPhysical = new (_physical, ProxyRoot);
        
        _conflict = new ();
    }

    private void PopulateFs()
    {
        _physical.CreateDirectory(PhysicalDir);
        _memory.CreateDirectory(MemoryDir);
        _proxyToMemory.CreateDirectory(ProxyDir);
        _proxyToPhysical.CreateDirectory(ProxyDir);

        _physical.CreateDirectory(SharedDir);
        _memory.CreateDirectory(SharedDir);
        _proxyToMemory.CreateDirectory(SharedDir);
        _proxyToPhysical.CreateDirectory(SharedDir);

        _physical.CreateFile(SharedFile);
        _memory.CreateFile(SharedFile);
        _proxyToMemory.CreateFile(SharedFile);
        _proxyToPhysical.CreateFile(SharedFile);

        _physical.CreateFile(PhysicalFile);
        _memory.CreateFile(MemoryFile);
        _proxyToMemory.CreateFile(ProxyFile);
        _proxyToPhysical.CreateFile(ProxyFile);
        
        _physical.WriteAllText(PhysicalFile, FileContents1);
        _memory.WriteAllText(MemoryFile, FileContents2);
        _proxyToMemory.WriteAllText(ProxyFile, FileContents1);
        _proxyToPhysical.WriteAllText(ProxyFile, FileContents2);
        
        _physical.WriteAllText(SharedFile, FileContents1);
        _memory.WriteAllText(SharedFile, FileContents2);
        _proxyToMemory.WriteAllText(SharedFile, FileContents1);
        _proxyToPhysical.WriteAllText(SharedFile, FileContents2);

        _conflict.CreateDirectory(MemoryDir);
        _conflict.CreateDirectory(MemoryFile);
        _conflict.CreateFile(SharedDir);
    }

    [TearDown]
    public void TearDown()
    {
        Cleanup();
    }


    [Test]
    public void CreateAndMerge()
    {
        _merged = new MergedFileSystem();
        _merged.AddFileSystem(_physical);
        _merged.AddFileSystem(_memory);
        _merged.AddFileSystem(_proxyToMemory);
        
        Assert.That(_merged.ReadAllText(SharedFile), Is.EqualTo(FileContents1));
        Assert.That(_merged.GetFile(SharedFile), Is.EqualTo(_proxyToMemory.GetFile(SharedFile)));
        
        _merged.AddFileSystem(_proxyToPhysical);
        Assert.That(_merged.GetFile(SharedFile), Is.EqualTo(_proxyToPhysical.GetFile(SharedFile)));
        Assert.That(_merged.GetFile(SharedFile), Is.Not.EqualTo(_proxyToMemory.GetFile(SharedFile)));
        Assert.That(_merged.ReadAllText(SharedFile), Is.EqualTo(FileContents2));

        Assert.That(_merged.IsDirectory(ProxyRoot));
        Assert.That(_merged.IsDirectory(MemoryDir));
        Assert.That(_merged.IsDirectory(PhysicalDir));
        Assert.That(_merged.IsDirectory(SharedDir));
        
        Assert.That(_merged.IsFile(SharedFile));
        Assert.That(_merged.IsFile(PhysicalFile));
        Assert.That(_merged.IsFile(ProxyFile));
        Assert.That(_merged.IsFile(MemoryFile));
        
        _merged.AddFileSystem(_conflict);
        Assert.That(_merged.IsDirectory(SharedDir));
        Assert.That(_merged.IsFile(SharedFile));
        Assert.That(_merged.IsFile(MemoryFile));
        Assert.That(_merged.ReadAllText(SharedFile), Is.EqualTo(FileContents2));
    }
}