﻿using KludgeBox.VFS;

namespace KludgeBox.Tests.VFS;

[TestFixture]
public class PathHelperTests
{
    // Unix paths
    private const string UnixDirPath = "/home/user/.hidden/";
    private const string UnixFilePath = "/home/user/.hidden/file.ext";
    
    private const string NormalizedUnixDirPath = "/home/user/.hidden/";
    private const string TrimmedUnixDirPath = "/home/user/.hidden";
    private const string NormalizedUnixFilePath = UnixFilePath;
    private const string UnixDirParent = "/home/user/";
    
    private const string RootlessUnixDir = "home/user/.hidden/";
    private const string RootlessUnixFile = "home/user/.hidden/file.ext";
    
    
    // Windows paths
    private const string WindowsDirPath = @"C:\Users\User\.thisOne\";
    private const string WindowsFilePath = @"C:\Users\User\.thisOne\file.ext";
    
    private const string NormalizedWindowsDirPath = "C:/Users/User/.thisOne/";
    private const string TrimmedWindowsDirPath = @"C:\Users\User\.thisOne";
    private const string NormalizedWindowsFilePath = "C:/Users/User/.thisOne/file.ext";
    private const string WindowsDirParent = "C:/Users/User/";

    private const string RootlessWindowsDir = "Users/User/.thisOne/";
    private const string RootlessWindowsFile = "Users/User/.thisOne/file.ext";
    
    private const string NormalizedWindowsRoot = "C:/";
    private const string WindowsRoot = "C:\\";
    private const string CursedWindowsRoot = "c:\\";
    
    
    // URI
    private const string DirUri = "scheme://directory/";
    private const string FileUri = "scheme://directory/file.ext";
    
    private const string NormalizedDirUri = "scheme://directory/";
    private const string TrimmedDirUri = "scheme://directory";
    private const string NormalizedFileUri = "scheme://directory/file.ext";
    private const string DirUriParent = UriScheme;

    private const string RootlessDirUri = "directory/";
    private const string RootlessFileUri = "directory/file.ext";
    
    private const string UriScheme = "scheme://";
    private const string CursedUriScheme = "ScHeMe://";

    
    // Special things
    private const string FileName = "file.ext";
    private const string UnixRoot = "/";
    

    [TestCase(UnixFilePath, ExpectedResult = FileName)]
    [TestCase(WindowsFilePath, ExpectedResult = FileName)]
    [TestCase(FileUri, ExpectedResult = FileName)]
    [TestCase(FileName, ExpectedResult = FileName)]
    [TestCase(WindowsDirPath, ExpectedResult = ".thisOne")]
    [TestCase(UnixDirParent, ExpectedResult = "user")]
    [TestCase(DirUri, ExpectedResult = "directory")]
    [TestCase(UnixRoot, ExpectedResult = "")]
    [TestCase(WindowsRoot, ExpectedResult = "")]
    [TestCase(UriScheme, ExpectedResult = "")]
    public string GetName(string path)
    {
        return path.GetName();
    }

    [Test]
    public void GetName_Throws_ArgumentNullException()
    {
        Assert.Throws<ArgumentNullException>(() => GetName(""));
    }

    [TestCase(UnixDirPath, ExpectedResult = NormalizedUnixDirPath)]
    [TestCase(UnixFilePath, ExpectedResult = NormalizedUnixFilePath)]
    [TestCase(WindowsDirPath, ExpectedResult = NormalizedWindowsDirPath)]
    [TestCase(WindowsFilePath, ExpectedResult = NormalizedWindowsFilePath)]
    [TestCase(FileUri, ExpectedResult = NormalizedFileUri)]
    [TestCase(DirUri, ExpectedResult = NormalizedDirUri)]
    [TestCase(UnixRoot, ExpectedResult = UnixRoot)]
    public string NormalizePath(string path)
    {
        return PathHelper.NormalizePath(path);
    }

    [TestCase(UnixDirPath, FileName, ExpectedResult = NormalizedUnixFilePath)]
    [TestCase(WindowsDirPath, FileName, ExpectedResult = NormalizedWindowsFilePath)]
    [TestCase(DirUri, FileName, ExpectedResult = NormalizedFileUri)]
    [TestCase(UnixRoot, UnixFilePath, ExpectedResult = NormalizedUnixFilePath)]
    public string Combine(string path1, string path2)
    {
        return PathHelper.Combine(path1, path2);
    }

    [Test]
    public void Combine_Throws_ArgumentNullException()
    {
        Assert.Throws<ArgumentNullException>(() => { PathHelper.Combine(null, null); });
    }

    [TestCase(UnixDirPath, ExpectedResult = TrimmedUnixDirPath)]
    [TestCase(DirUri, ExpectedResult = TrimmedDirUri)]
    [TestCase(WindowsDirPath, ExpectedResult = TrimmedWindowsDirPath)]
    [TestCase(UnixRoot,ExpectedResult = UnixRoot)]
    public string TrimEndingDirectorySeparator(string path)
    {
        return PathHelper.TrimEndingDirectorySeparator(path);
    }

    [TestCase(WindowsDirPath, ExpectedResult = false)]
    [TestCase(UnixDirPath, ExpectedResult = false)]
    [TestCase(DirUri, ExpectedResult = false)]
    [TestCase(WindowsFilePath, ExpectedResult = true)]
    [TestCase(UnixFilePath, ExpectedResult = true)]
    [TestCase(FileUri, ExpectedResult = true)]
    [TestCase(UnixRoot, ExpectedResult = false)]
    [TestCase(FileName, ExpectedResult = true)]
    [TestCase("", ExpectedResult = false)]
    [TestCase(null, ExpectedResult = false)]
    public bool HasExtension(string path)
    {
        return PathHelper.HasExtension(path);
    }
    
    
    [TestCase(WindowsFilePath, ExpectedResult = NormalizedWindowsDirPath)]
    [TestCase(UnixFilePath, ExpectedResult = UnixDirPath)]
    [TestCase(FileUri, ExpectedResult = DirUri)]
    [TestCase(WindowsDirPath, ExpectedResult = WindowsDirParent)]
    [TestCase(UnixDirPath, ExpectedResult = UnixDirParent)]
    [TestCase(DirUri, ExpectedResult = DirUriParent)]
    [TestCase(UnixRoot, ExpectedResult = UnixRoot)]
    [TestCase("relative/path/", ExpectedResult = "relative/")]
    [TestCase("relative/", ExpectedResult = "")]
    public string GetParent(string path)
    {
        return PathHelper.GetParent(path);
    }

    [TestCase(UnixRoot, ExpectedResult = true)]
    [TestCase(NormalizedWindowsRoot, ExpectedResult = true)]
    [TestCase(CursedWindowsRoot, ExpectedResult = true)]
    [TestCase(UriScheme, ExpectedResult = true)]
    [TestCase(CursedUriScheme, ExpectedResult = true)]
    [TestCase(UnixDirPath, ExpectedResult = false)]
    [TestCase(WindowsDirPath, ExpectedResult = false)]
    [TestCase(FileUri, ExpectedResult = false)]
    [TestCase("", ExpectedResult = false)]
    public bool IsRoot(string path)
    {
        return PathHelper.IsRoot(path);
    }

    [TestCase(UnixRoot, ExpectedResult = UnixRoot)]
    [TestCase(WindowsRoot, ExpectedResult = NormalizedWindowsRoot)]
    [TestCase(CursedWindowsRoot, ExpectedResult = NormalizedWindowsRoot)]
    [TestCase(UriScheme, ExpectedResult = UriScheme)]
    [TestCase(CursedUriScheme, ExpectedResult = UriScheme)]
    [TestCase(FileName, ExpectedResult = "")]
    public string GetRoot(string path)
    {
        return PathHelper.GetRoot(path);
    }

    [TestCase(UnixRoot, ExpectedResult = new[] {UnixRoot})]
    [TestCase(WindowsRoot, ExpectedResult = new[] {NormalizedWindowsRoot})]
    [TestCase(CursedWindowsRoot, ExpectedResult = new[] {NormalizedWindowsRoot})]
    [TestCase(FileName, ExpectedResult = new[]{FileName})]
    [TestCase(UnixFilePath, ExpectedResult = new[] {UnixRoot, "home", "user", ".hidden", "file.ext"})] // /home/user/.hidden/file.ext
    [TestCase(UnixDirPath, ExpectedResult = new[] {UnixRoot, "home", "user", ".hidden"})] // /home/user/.hidden/
    [TestCase(UnixDirParent, ExpectedResult = new[] {UnixRoot, "home", "user"})] // /home/user/
    [TestCase(WindowsFilePath, ExpectedResult = new[] {NormalizedWindowsRoot, "Users", "User", ".thisOne", "file.ext"})] // C:\Users\User\.thisOne\file.ext
    [TestCase(WindowsDirPath, ExpectedResult = new[] {NormalizedWindowsRoot, "Users", "User", ".thisOne"})] // C:\Users\User\.thisOne\
    [TestCase(WindowsDirParent, ExpectedResult = new[] {NormalizedWindowsRoot, "Users", "User"})] // C:\Users\User\
    [TestCase(FileUri, ExpectedResult = new[] {UriScheme, "directory", "file.ext"})] // scheme://directory/file.ext
    [TestCase(DirUri, ExpectedResult = new[] {UriScheme, "directory"})] // scheme://directory/
    [TestCase(DirUriParent, ExpectedResult = new[] {UriScheme})] // scheme://
    public string[] PathParts(string path)
    {
        return PathHelper.PathParts(path);
    }

    [TestCase(FileName, ExpectedResult = false)]
    [TestCase(WindowsFilePath, ExpectedResult = true)]
    [TestCase(CursedWindowsRoot, ExpectedResult = true)]
    [TestCase(CursedUriScheme, ExpectedResult = true)]
    [TestCase(UnixRoot, ExpectedResult = true)]
    [TestCase(NormalizedWindowsDirPath, ExpectedResult = true)]
    public bool HasRoot(string path)
    {
        return PathHelper.HasRoot(path);
    }

    [TestCase(WindowsFilePath, ExpectedResult = RootlessWindowsFile)]
    [TestCase(WindowsDirPath, ExpectedResult = RootlessWindowsDir)]
    [TestCase(NormalizedWindowsFilePath, ExpectedResult = RootlessWindowsFile)]
    [TestCase(UnixFilePath, ExpectedResult = RootlessUnixFile)]
    [TestCase(UnixDirPath, ExpectedResult = RootlessUnixDir)]
    [TestCase(FileUri, ExpectedResult = RootlessFileUri)]
    [TestCase(DirUri, ExpectedResult = RootlessDirUri)]
    [TestCase(UnixRoot, ExpectedResult = "")]
    [TestCase(WindowsRoot, ExpectedResult = "")]
    [TestCase(CursedWindowsRoot, ExpectedResult = "")]
    [TestCase(UriScheme, ExpectedResult = "")]
    [TestCase(CursedUriScheme, ExpectedResult = "")]
    [TestCase("", ExpectedResult = "")]
    [TestCase(null, ExpectedResult = null)]
    public string RemoveRoot(string path)
    {
        return PathHelper.RemoveRoot(path);
    }

    [TestCase(UnixFilePath, ExpectedResult = new[]
    {
        "/",
        "/home/",
        "/home/user/",
        "/home/user/.hidden/",
        "/home/user/.hidden/file.ext",
    })]
    [TestCase(WindowsFilePath, ExpectedResult = new[] // C:\Users\User\.thisOne\file.ext
    {
        "C:/",
        "C:/Users/",
        "C:/Users/User/",
        "C:/Users/User/.thisOne/",
        "C:/Users/User/.thisOne/file.ext"
    })]
    [TestCase(UriScheme, ExpectedResult = new[] { UriScheme })]
    [TestCase("relative/path/file.ext", ExpectedResult = new[]
    {
        "relative/",
        "relative/path/",
        "relative/path/file.ext"
    })]
    public string[] PathsChain(string path)
    {
        var chain = PathHelper.PathsChain(path);
        foreach (var cpath in chain)
            Console.WriteLine(cpath);
        return chain;
    }
}